// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: substrait/extension_rels.proto
// Protobuf C++ Version: 7.35.1

#include "substrait/extension_rels.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey
        {&::_pbi::kDescriptorMethods, &::descriptor_table_substrait_2fextension_5frels_2eproto, /* tracker*/ nullptr,},
        // ::arrow::substrait_ext::AsOfJoinRel
        {&::_pbi::kDescriptorMethods, &::descriptor_table_substrait_2fextension_5frels_2eproto, /* tracker*/ nullptr,},
        // ::arrow::substrait_ext::NamedTapRel
        {&::_pbi::kDescriptorMethods, &::descriptor_table_substrait_2fextension_5frels_2eproto, /* tracker*/ nullptr,},
        // ::arrow::substrait_ext::SegmentedAggregateRel
        {&::_pbi::kDescriptorMethods, &::descriptor_table_substrait_2fextension_5frels_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace arrow {
namespace substrait_ext {
class NamedTapRel::_Internal {
 public:
  using HasBits = decltype(::std::declval<NamedTapRel>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(NamedTapRel, _impl_._has_bits_);
};

constexpr NamedTapRel::ParseTableT_ NamedTapRel::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(NamedTapRel, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::substrait_ext::NamedTapRel>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string kind = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 1, 0,
        PROTOBUF_FIELD_OFFSET(NamedTapRel, _impl_.kind_)}},
      // string name = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 2, 0,
        PROTOBUF_FIELD_OFFSET(NamedTapRel, _impl_.name_)}},
      // repeated string columns = 3;
      {::_pbi::TcParser::FastUR1,
       {26, 0, 0,
        PROTOBUF_FIELD_OFFSET(NamedTapRel, _impl_.columns_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string kind = 1;
      {PROTOBUF_FIELD_OFFSET(NamedTapRel, _impl_.kind_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string name = 2;
      {PROTOBUF_FIELD_OFFSET(NamedTapRel, _impl_.name_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // repeated string columns = 3;
      {PROTOBUF_FIELD_OFFSET(NamedTapRel, _impl_.columns_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
    }},
    // no aux_entries
    {{
      "\37\4\4\7\0\0\0\0"
      "arrow.substrait_ext.NamedTapRel"
      "kind"
      "name"
      "columns"
    }},
  };
}


inline constexpr NamedTapRel::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        columns_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::NamedTapRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::NamedTapRel, _impl_.columns_)>()
         }
        ,
        kind_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr NamedTapRel::NamedTapRel(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL NamedTapRel::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) NamedTapRel(arena);
}
constexpr auto NamedTapRel::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(NamedTapRel), alignof(NamedTapRel));
}
constexpr auto NamedTapRel::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &NamedTapRel::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<NamedTapRel>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &NamedTapRel::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<NamedTapRel>(), &NamedTapRel::ByteSizeLong,
              &NamedTapRel::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(NamedTapRel, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[2],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_substrait_2fextension_5frels_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct NamedTapRelGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr NamedTapRelGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 NamedTapRel_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(NamedTapRel::InternalGenerateClassData_(
            _default, &NamedTapRel_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<NamedTapRel>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~NamedTapRelGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) NamedTapRel _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<NamedTapRel>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(NamedTapRelGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST NamedTapRelGlobalsTypeInternal NamedTapRel_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* NamedTapRel_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return NamedTapRel_globals_.GetClassData();
#else
  return NamedTapRel_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class SegmentedAggregateRel::_Internal {
 public:
  using HasBits = decltype(::std::declval<SegmentedAggregateRel>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(SegmentedAggregateRel, _impl_._has_bits_);
};

constexpr SegmentedAggregateRel::ParseTableT_ SegmentedAggregateRel::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(SegmentedAggregateRel, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      3,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::substrait_ext::SegmentedAggregateRel>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // repeated .substrait.Expression.FieldReference grouping_keys = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(SegmentedAggregateRel, _impl_.grouping_keys_)}},
      // repeated .substrait.Expression.FieldReference segment_keys = 2;
      {::_pbi::TcParser::FastMtR1,
       {18, 1, 1,
        PROTOBUF_FIELD_OFFSET(SegmentedAggregateRel, _impl_.segment_keys_)}},
      // repeated .substrait.AggregateRel.Measure measures = 3;
      {::_pbi::TcParser::FastMtR1,
       {26, 2, 2,
        PROTOBUF_FIELD_OFFSET(SegmentedAggregateRel, _impl_.measures_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .substrait.Expression.FieldReference grouping_keys = 1;
      {PROTOBUF_FIELD_OFFSET(SegmentedAggregateRel, _impl_.grouping_keys_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .substrait.Expression.FieldReference segment_keys = 2;
      {PROTOBUF_FIELD_OFFSET(SegmentedAggregateRel, _impl_.segment_keys_), _Internal::kHasBitsOffset + 1, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .substrait.AggregateRel.Measure measures = 3;
      {PROTOBUF_FIELD_OFFSET(SegmentedAggregateRel, _impl_.measures_), _Internal::kHasBitsOffset + 2, 2, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::substrait::Expression_FieldReference>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::substrait::Expression_FieldReference_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::substrait::Expression_FieldReference>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::substrait::Expression_FieldReference_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::substrait::AggregateRel_Measure>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::substrait::AggregateRel_Measure_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr SegmentedAggregateRel::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        grouping_keys_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::SegmentedAggregateRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.grouping_keys_)>()
         }
        ,
        segment_keys_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::SegmentedAggregateRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.segment_keys_)>()
         }
        ,
        measures_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::SegmentedAggregateRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.measures_)>()
         }
     {}

template <typename>
constexpr SegmentedAggregateRel::SegmentedAggregateRel(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL SegmentedAggregateRel::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SegmentedAggregateRel(arena);
}
constexpr auto SegmentedAggregateRel::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(SegmentedAggregateRel), alignof(SegmentedAggregateRel));
}
constexpr auto SegmentedAggregateRel::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &SegmentedAggregateRel::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SegmentedAggregateRel>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SegmentedAggregateRel::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<SegmentedAggregateRel>(), &SegmentedAggregateRel::ByteSizeLong,
              &SegmentedAggregateRel::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SegmentedAggregateRel, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[3],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_substrait_2fextension_5frels_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct SegmentedAggregateRelGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr SegmentedAggregateRelGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 SegmentedAggregateRel_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(SegmentedAggregateRel::InternalGenerateClassData_(
            _default, &SegmentedAggregateRel_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<SegmentedAggregateRel>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~SegmentedAggregateRelGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) SegmentedAggregateRel _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<SegmentedAggregateRel>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(SegmentedAggregateRelGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST SegmentedAggregateRelGlobalsTypeInternal SegmentedAggregateRel_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* SegmentedAggregateRel_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return SegmentedAggregateRel_globals_.GetClassData();
#else
  return SegmentedAggregateRel_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class AsOfJoinRel_AsOfJoinKey::_Internal {
 public:
  using HasBits = decltype(::std::declval<AsOfJoinRel_AsOfJoinKey>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(AsOfJoinRel_AsOfJoinKey, _impl_._has_bits_);
};

constexpr AsOfJoinRel_AsOfJoinKey::ParseTableT_ AsOfJoinRel_AsOfJoinKey::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(AsOfJoinRel_AsOfJoinKey, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      2,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .substrait.Expression by = 2;
      {::_pbi::TcParser::FastMtR1,
       {18, 0, 1,
        PROTOBUF_FIELD_OFFSET(AsOfJoinRel_AsOfJoinKey, _impl_.by_)}},
      // .substrait.Expression on = 1;
      {::_pbi::TcParser::FastMtS1,
       {10, 1, 0,
        PROTOBUF_FIELD_OFFSET(AsOfJoinRel_AsOfJoinKey, _impl_.on_)}},
    }}, {{
      65535, 65535
    }}, {{
      // .substrait.Expression on = 1;
      {PROTOBUF_FIELD_OFFSET(AsOfJoinRel_AsOfJoinKey, _impl_.on_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .substrait.Expression by = 2;
      {PROTOBUF_FIELD_OFFSET(AsOfJoinRel_AsOfJoinKey, _impl_.by_), _Internal::kHasBitsOffset + 0, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::substrait::Expression>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::substrait::Expression_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::substrait::Expression>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::substrait::Expression_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr AsOfJoinRel_AsOfJoinKey::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        by_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey, _impl_.by_)>()
         }
        ,
        on_{nullptr} {}

template <typename>
constexpr AsOfJoinRel_AsOfJoinKey::AsOfJoinRel_AsOfJoinKey(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL AsOfJoinRel_AsOfJoinKey::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) AsOfJoinRel_AsOfJoinKey(arena);
}
constexpr auto AsOfJoinRel_AsOfJoinKey::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(AsOfJoinRel_AsOfJoinKey), alignof(AsOfJoinRel_AsOfJoinKey));
}
constexpr auto AsOfJoinRel_AsOfJoinKey::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &AsOfJoinRel_AsOfJoinKey::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<AsOfJoinRel_AsOfJoinKey>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &AsOfJoinRel_AsOfJoinKey::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<AsOfJoinRel_AsOfJoinKey>(), &AsOfJoinRel_AsOfJoinKey::ByteSizeLong,
              &AsOfJoinRel_AsOfJoinKey::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(AsOfJoinRel_AsOfJoinKey, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_substrait_2fextension_5frels_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct AsOfJoinRel_AsOfJoinKeyGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr AsOfJoinRel_AsOfJoinKeyGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 AsOfJoinRel_AsOfJoinKey_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(AsOfJoinRel_AsOfJoinKey::InternalGenerateClassData_(
            _default, &AsOfJoinRel_AsOfJoinKey_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<AsOfJoinRel_AsOfJoinKey>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~AsOfJoinRel_AsOfJoinKeyGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) AsOfJoinRel_AsOfJoinKey _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<AsOfJoinRel_AsOfJoinKey>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(AsOfJoinRel_AsOfJoinKeyGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST AsOfJoinRel_AsOfJoinKeyGlobalsTypeInternal AsOfJoinRel_AsOfJoinKey_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* AsOfJoinRel_AsOfJoinKey_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return AsOfJoinRel_AsOfJoinKey_globals_.GetClassData();
#else
  return AsOfJoinRel_AsOfJoinKey_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class AsOfJoinRel::_Internal {
 public:
  using HasBits = decltype(::std::declval<AsOfJoinRel>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(AsOfJoinRel, _impl_._has_bits_);
};

constexpr AsOfJoinRel::ParseTableT_ AsOfJoinRel::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(AsOfJoinRel, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::substrait_ext::AsOfJoinRel>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // int64 tolerance = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(AsOfJoinRel, _impl_.tolerance_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(AsOfJoinRel, _impl_.tolerance_)}},
      // repeated .arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey keys = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(AsOfJoinRel, _impl_.keys_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey keys = 1;
      {PROTOBUF_FIELD_OFFSET(AsOfJoinRel, _impl_.keys_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // int64 tolerance = 2;
      {PROTOBUF_FIELD_OFFSET(AsOfJoinRel, _impl_.tolerance_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt64)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr AsOfJoinRel::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        keys_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::AsOfJoinRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel, _impl_.keys_)>()
         }
        ,
        tolerance_{::int64_t{0}} {}

template <typename>
constexpr AsOfJoinRel::AsOfJoinRel(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL AsOfJoinRel::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) AsOfJoinRel(arena);
}
constexpr auto AsOfJoinRel::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(AsOfJoinRel), alignof(AsOfJoinRel));
}
constexpr auto AsOfJoinRel::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &AsOfJoinRel::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<AsOfJoinRel>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &AsOfJoinRel::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<AsOfJoinRel>(), &AsOfJoinRel::ByteSizeLong,
              &AsOfJoinRel::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(AsOfJoinRel, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_substrait_2fextension_5frels_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct AsOfJoinRelGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr AsOfJoinRelGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 AsOfJoinRel_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(AsOfJoinRel::InternalGenerateClassData_(
            _default, &AsOfJoinRel_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<AsOfJoinRel>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~AsOfJoinRelGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) AsOfJoinRel _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<AsOfJoinRel>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(AsOfJoinRelGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST AsOfJoinRelGlobalsTypeInternal AsOfJoinRel_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* AsOfJoinRel_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return AsOfJoinRel_globals_.GetClassData();
#else
  return AsOfJoinRel_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace substrait_ext
}  // namespace arrow
static constexpr const ::_pb::EnumDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_enum_descriptors_substrait_2fextension_5frels_2eproto = nullptr;
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_substrait_2fextension_5frels_2eproto = nullptr;
const ::uint32_t
    TableStruct_substrait_2fextension_5frels_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey, _impl_.on_),
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey, _impl_.by_),
        1,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel, _impl_.keys_),
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel, _impl_.tolerance_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::NamedTapRel, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::NamedTapRel, _impl_.kind_),
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::NamedTapRel, _impl_.name_),
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::NamedTapRel, _impl_.columns_),
        1,
        2,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.grouping_keys_),
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.segment_keys_),
        PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.measures_),
        0,
        1,
        2,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey)},
        {7, sizeof(::arrow::substrait_ext::AsOfJoinRel)},
        {14, sizeof(::arrow::substrait_ext::NamedTapRel)},
        {23, sizeof(::arrow::substrait_ext::SegmentedAggregateRel)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey_globals_,
        &::arrow::substrait_ext::AsOfJoinRel_globals_,
        &::arrow::substrait_ext::NamedTapRel_globals_,
        &::arrow::substrait_ext::SegmentedAggregateRel_globals_,
};
const char descriptor_table_protodef_substrait_2fextension_5frels_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\036substrait/extension_rels.proto\022\023arrow."
    "substrait_ext\032\027substrait/algebra.proto\"\261"
    "\001\n\013AsOfJoinRel\022:\n\004keys\030\001 \003(\0132,.arrow.sub"
    "strait_ext.AsOfJoinRel.AsOfJoinKey\022\021\n\tto"
    "lerance\030\002 \001(\003\032S\n\013AsOfJoinKey\022!\n\002on\030\001 \001(\013"
    "2\025.substrait.Expression\022!\n\002by\030\002 \003(\0132\025.su"
    "bstrait.Expression\":\n\013NamedTapRel\022\014\n\004kin"
    "d\030\001 \001(\t\022\014\n\004name\030\002 \001(\t\022\017\n\007columns\030\003 \003(\t\"\303"
    "\001\n\025SegmentedAggregateRel\022;\n\rgrouping_key"
    "s\030\001 \003(\0132$.substrait.Expression.FieldRefe"
    "rence\022:\n\014segment_keys\030\002 \003(\0132$.substrait."
    "Expression.FieldReference\0221\n\010measures\030\003 "
    "\003(\0132\037.substrait.AggregateRel.MeasureBK\n\022"
    "io.arrow.substraitP\001Z!github.com/apache/"
    "arrow/substrait\252\002\017Arrow.Substraitb\006proto"
    "3"
};
static const ::_pbi::DescriptorTable* PROTOBUF_NONNULL const
    descriptor_table_substrait_2fextension_5frels_2eproto_deps[1] = {
        &::descriptor_table_substrait_2falgebra_2eproto,
};
static ::absl::once_flag descriptor_table_substrait_2fextension_5frels_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_substrait_2fextension_5frels_2eproto = {
    false,
    false,
    601,
    descriptor_table_protodef_substrait_2fextension_5frels_2eproto,
    "substrait/extension_rels.proto",
    &descriptor_table_substrait_2fextension_5frels_2eproto_once,
    descriptor_table_substrait_2fextension_5frels_2eproto_deps,
    1,
    4,
    schemas,
    file_message_globals,
    TableStruct_substrait_2fextension_5frels_2eproto::offsets,
    file_level_enum_descriptors_substrait_2fextension_5frels_2eproto,
    file_level_service_descriptors_substrait_2fextension_5frels_2eproto,
};
namespace arrow {
namespace substrait_ext {
// ===================================================================

void AsOfJoinRel_AsOfJoinKey::clear_on() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.on_ != nullptr) _impl_.on_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
void AsOfJoinRel_AsOfJoinKey::clear_by() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.by_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
AsOfJoinRel_AsOfJoinKey::AsOfJoinRel_AsOfJoinKey(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AsOfJoinRel_AsOfJoinKey_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey)
}
PROTOBUF_NDEBUG_INLINE AsOfJoinRel_AsOfJoinKey::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        by_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey,
              PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey, _impl_.by_)>()
          , from.by_
        }
     {}

AsOfJoinRel_AsOfJoinKey::AsOfJoinRel_AsOfJoinKey(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const AsOfJoinRel_AsOfJoinKey& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AsOfJoinRel_AsOfJoinKey_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  AsOfJoinRel_AsOfJoinKey* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.on_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.on_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey)
}
PROTOBUF_NDEBUG_INLINE AsOfJoinRel_AsOfJoinKey::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        by_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel_AsOfJoinKey, _impl_.by_)>()
         }
     {}

inline void AsOfJoinRel_AsOfJoinKey::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.on_ = {};
}
AsOfJoinRel_AsOfJoinKey::~AsOfJoinRel_AsOfJoinKey() {
  // @@protoc_insertion_point(destructor:arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey)
  SharedDtor(*this);
}
inline void AsOfJoinRel_AsOfJoinKey::SharedDtor(MessageLite& self) {
  AsOfJoinRel_AsOfJoinKey& this_ = static_cast<AsOfJoinRel_AsOfJoinKey&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.on_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull AsOfJoinRel_AsOfJoinKey_class_data_ =
        AsOfJoinRel_AsOfJoinKey::InternalGenerateClassData_(AsOfJoinRel_AsOfJoinKey_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AsOfJoinRel_AsOfJoinKey::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AsOfJoinRel_AsOfJoinKey_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(AsOfJoinRel_AsOfJoinKey_class_data_.tc_table);
  return AsOfJoinRel_AsOfJoinKey_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AsOfJoinRel_AsOfJoinKey::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AsOfJoinRel_AsOfJoinKey_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&AsOfJoinRel_AsOfJoinKey_globals_));
  return AsOfJoinRel_AsOfJoinKey_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const AsOfJoinRel_AsOfJoinKey::ParseTableT_
    AsOfJoinRel_AsOfJoinKey::_table_ =
        AsOfJoinRel_AsOfJoinKey::InternalGenerateParseTable_(AsOfJoinRel_AsOfJoinKey_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void AsOfJoinRel_AsOfJoinKey::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.by_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.on_ != nullptr);
      _impl_.on_->Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL AsOfJoinRel_AsOfJoinKey::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const AsOfJoinRel_AsOfJoinKey& this_ = static_cast<const AsOfJoinRel_AsOfJoinKey&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL AsOfJoinRel_AsOfJoinKey::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const AsOfJoinRel_AsOfJoinKey& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .substrait.Expression on = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.on_, this_._impl_.on_->GetCachedSize(), target,
        stream);
  }

  // repeated .substrait.Expression by = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_by_size());
         i < n; i++) {
      const auto& repfield = this_._internal_by().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              2, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t AsOfJoinRel_AsOfJoinKey::ByteSizeLong(const MessageLite& base) {
  const AsOfJoinRel_AsOfJoinKey& this_ = static_cast<const AsOfJoinRel_AsOfJoinKey&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t AsOfJoinRel_AsOfJoinKey::ByteSizeLong() const {
  const AsOfJoinRel_AsOfJoinKey& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // repeated .substrait.Expression by = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_by_size();
      for (const auto& msg : this_._internal_by()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // .substrait.Expression on = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.on_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void AsOfJoinRel_AsOfJoinKey::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<AsOfJoinRel_AsOfJoinKey*>(&to_msg);
  auto& from = static_cast<const AsOfJoinRel_AsOfJoinKey&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_by()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_by());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.on_ != nullptr);
      if (_this->_impl_.on_ == nullptr) {
        _this->_impl_.on_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.on_);
      } else {
        _this->_impl_.on_->MergeFrom(*from._impl_.on_);
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void AsOfJoinRel_AsOfJoinKey::CopyFrom(const AsOfJoinRel_AsOfJoinKey& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void AsOfJoinRel_AsOfJoinKey::InternalSwap(AsOfJoinRel_AsOfJoinKey* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.by_.InternalSwap(&other->_impl_.by_);
  swap(_impl_.on_, other->_impl_.on_);
}

::google::protobuf::Metadata AsOfJoinRel_AsOfJoinKey::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

AsOfJoinRel::AsOfJoinRel(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AsOfJoinRel_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.substrait_ext.AsOfJoinRel)
}
PROTOBUF_NDEBUG_INLINE AsOfJoinRel::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::substrait_ext::AsOfJoinRel& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        keys_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::arrow::substrait_ext::AsOfJoinRel,
              PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel, _impl_.keys_)>()
          , from.keys_
        }
     {}

AsOfJoinRel::AsOfJoinRel(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const AsOfJoinRel& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AsOfJoinRel_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  AsOfJoinRel* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.tolerance_ = from._impl_.tolerance_;

  // @@protoc_insertion_point(copy_constructor:arrow.substrait_ext.AsOfJoinRel)
}
PROTOBUF_NDEBUG_INLINE AsOfJoinRel::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        keys_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::AsOfJoinRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::AsOfJoinRel, _impl_.keys_)>()
         }
     {}

inline void AsOfJoinRel::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.tolerance_ = {};
}
AsOfJoinRel::~AsOfJoinRel() {
  // @@protoc_insertion_point(destructor:arrow.substrait_ext.AsOfJoinRel)
  SharedDtor(*this);
}
inline void AsOfJoinRel::SharedDtor(MessageLite& self) {
  AsOfJoinRel& this_ = static_cast<AsOfJoinRel&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull AsOfJoinRel_class_data_ =
        AsOfJoinRel::InternalGenerateClassData_(AsOfJoinRel_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AsOfJoinRel::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AsOfJoinRel_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(AsOfJoinRel_class_data_.tc_table);
  return AsOfJoinRel_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AsOfJoinRel::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AsOfJoinRel_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&AsOfJoinRel_globals_));
  return AsOfJoinRel_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const AsOfJoinRel::ParseTableT_
    AsOfJoinRel::_table_ =
        AsOfJoinRel::InternalGenerateParseTable_(AsOfJoinRel_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void AsOfJoinRel::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.substrait_ext.AsOfJoinRel)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.keys_.Clear();
  }
  _impl_.tolerance_ = ::int64_t{0};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL AsOfJoinRel::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const AsOfJoinRel& this_ = static_cast<const AsOfJoinRel&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL AsOfJoinRel::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const AsOfJoinRel& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.substrait_ext.AsOfJoinRel)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey keys = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_keys_size());
         i < n; i++) {
      const auto& repfield = this_._internal_keys().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // int64 tolerance = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_tolerance() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt64ToArrayWithField<2>(
              stream, this_._internal_tolerance(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.substrait_ext.AsOfJoinRel)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t AsOfJoinRel::ByteSizeLong(const MessageLite& base) {
  const AsOfJoinRel& this_ = static_cast<const AsOfJoinRel&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t AsOfJoinRel::ByteSizeLong() const {
  const AsOfJoinRel& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.substrait_ext.AsOfJoinRel)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // repeated .arrow.substrait_ext.AsOfJoinRel.AsOfJoinKey keys = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_keys_size();
      for (const auto& msg : this_._internal_keys()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // int64 tolerance = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_tolerance() != 0) {
        total_size += ::_pbi::WireFormatLite::Int64SizePlusOne(
            this_._internal_tolerance());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void AsOfJoinRel::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<AsOfJoinRel*>(&to_msg);
  auto& from = static_cast<const AsOfJoinRel&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.substrait_ext.AsOfJoinRel)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_keys()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_keys());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_tolerance() != 0) {
        _this->_impl_.tolerance_ = from._impl_.tolerance_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void AsOfJoinRel::CopyFrom(const AsOfJoinRel& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.substrait_ext.AsOfJoinRel)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void AsOfJoinRel::InternalSwap(AsOfJoinRel* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.keys_.InternalSwap(&other->_impl_.keys_);
  swap(_impl_.tolerance_, other->_impl_.tolerance_);
}

::google::protobuf::Metadata AsOfJoinRel::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

NamedTapRel::NamedTapRel(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, NamedTapRel_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.substrait_ext.NamedTapRel)
}
PROTOBUF_NDEBUG_INLINE NamedTapRel::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::substrait_ext::NamedTapRel& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        columns_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::arrow::substrait_ext::NamedTapRel,
              PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::NamedTapRel, _impl_.columns_)>()
          , from.columns_
        }
        ,
        kind_(arena, from.kind_),
        name_(arena, from.name_) {}

NamedTapRel::NamedTapRel(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const NamedTapRel& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, NamedTapRel_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  NamedTapRel* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.substrait_ext.NamedTapRel)
}
PROTOBUF_NDEBUG_INLINE NamedTapRel::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        columns_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::NamedTapRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::NamedTapRel, _impl_.columns_)>()
         }
        ,
        kind_(arena),
        name_(arena) {}

inline void NamedTapRel::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
NamedTapRel::~NamedTapRel() {
  // @@protoc_insertion_point(destructor:arrow.substrait_ext.NamedTapRel)
  SharedDtor(*this);
}
inline void NamedTapRel::SharedDtor(MessageLite& self) {
  NamedTapRel& this_ = static_cast<NamedTapRel&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.kind_.Destroy();
  this_._impl_.name_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull NamedTapRel_class_data_ =
        NamedTapRel::InternalGenerateClassData_(NamedTapRel_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
NamedTapRel::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&NamedTapRel_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(NamedTapRel_class_data_.tc_table);
  return NamedTapRel_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
NamedTapRel::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&NamedTapRel_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&NamedTapRel_globals_));
  return NamedTapRel_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const NamedTapRel::ParseTableT_
    NamedTapRel::_table_ =
        NamedTapRel::InternalGenerateParseTable_(NamedTapRel_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void NamedTapRel::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.substrait_ext.NamedTapRel)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.columns_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.kind_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.name_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL NamedTapRel::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const NamedTapRel& this_ = static_cast<const NamedTapRel&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL NamedTapRel::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const NamedTapRel& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.substrait_ext.NamedTapRel)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string kind = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_kind().empty()) {
      const ::std::string& _s = this_._internal_kind();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.substrait_ext.NamedTapRel.kind");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string name = 2;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_name().empty()) {
      const ::std::string& _s = this_._internal_name();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.substrait_ext.NamedTapRel.name");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // repeated string columns = 3;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_columns_size(); i < n; ++i) {
      const auto& s = this_._internal_columns().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.substrait_ext.NamedTapRel.columns");
      target = stream->WriteString(3, s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.substrait_ext.NamedTapRel)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t NamedTapRel::ByteSizeLong(const MessageLite& base) {
  const NamedTapRel& this_ = static_cast<const NamedTapRel&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t NamedTapRel::ByteSizeLong() const {
  const NamedTapRel& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.substrait_ext.NamedTapRel)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // repeated string columns = 3;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_columns().size());
      for (int i = 0, n = this_._internal_columns().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_columns().Get(i));
      }
    }
    // string kind = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_kind().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_kind());
      }
    }
    // string name = 2;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_name().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_name());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void NamedTapRel::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<NamedTapRel*>(&to_msg);
  auto& from = static_cast<const NamedTapRel&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.substrait_ext.NamedTapRel)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_columns()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_columns());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_kind().empty()) {
        _this->_internal_set_kind(from._internal_kind());
      } else {
        if (_this->_impl_.kind_.IsDefault()) {
          _this->_internal_set_kind("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_name().empty()) {
        _this->_internal_set_name(from._internal_name());
      } else {
        if (_this->_impl_.name_.IsDefault()) {
          _this->_internal_set_name("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void NamedTapRel::CopyFrom(const NamedTapRel& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.substrait_ext.NamedTapRel)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void NamedTapRel::InternalSwap(NamedTapRel* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.columns_.InternalSwap(&other->_impl_.columns_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.kind_, &other->_impl_.kind_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.name_, &other->_impl_.name_, arena);
}

::google::protobuf::Metadata NamedTapRel::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

void SegmentedAggregateRel::clear_grouping_keys() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.grouping_keys_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
void SegmentedAggregateRel::clear_segment_keys() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.segment_keys_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
void SegmentedAggregateRel::clear_measures() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.measures_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
SegmentedAggregateRel::SegmentedAggregateRel(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SegmentedAggregateRel_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.substrait_ext.SegmentedAggregateRel)
}
PROTOBUF_NDEBUG_INLINE SegmentedAggregateRel::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::substrait_ext::SegmentedAggregateRel& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        grouping_keys_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::arrow::substrait_ext::SegmentedAggregateRel,
              PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.grouping_keys_)>()
          , from.grouping_keys_
        }
        ,
        segment_keys_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::arrow::substrait_ext::SegmentedAggregateRel,
              PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.segment_keys_)>()
          , from.segment_keys_
        }
        ,
        measures_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::arrow::substrait_ext::SegmentedAggregateRel,
              PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.measures_)>()
          , from.measures_
        }
     {}

SegmentedAggregateRel::SegmentedAggregateRel(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const SegmentedAggregateRel& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SegmentedAggregateRel_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SegmentedAggregateRel* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.substrait_ext.SegmentedAggregateRel)
}
PROTOBUF_NDEBUG_INLINE SegmentedAggregateRel::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        grouping_keys_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::SegmentedAggregateRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.grouping_keys_)>()
         }
        ,
        segment_keys_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::SegmentedAggregateRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.segment_keys_)>()
         }
        ,
        measures_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::substrait_ext::SegmentedAggregateRel,
            PROTOBUF_FIELD_OFFSET(::arrow::substrait_ext::SegmentedAggregateRel, _impl_.measures_)>()
         }
     {}

inline void SegmentedAggregateRel::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
SegmentedAggregateRel::~SegmentedAggregateRel() {
  // @@protoc_insertion_point(destructor:arrow.substrait_ext.SegmentedAggregateRel)
  SharedDtor(*this);
}
inline void SegmentedAggregateRel::SharedDtor(MessageLite& self) {
  SegmentedAggregateRel& this_ = static_cast<SegmentedAggregateRel&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SegmentedAggregateRel_class_data_ =
        SegmentedAggregateRel::InternalGenerateClassData_(SegmentedAggregateRel_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SegmentedAggregateRel::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SegmentedAggregateRel_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SegmentedAggregateRel_class_data_.tc_table);
  return SegmentedAggregateRel_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SegmentedAggregateRel::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SegmentedAggregateRel_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&SegmentedAggregateRel_globals_));
  return SegmentedAggregateRel_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const SegmentedAggregateRel::ParseTableT_
    SegmentedAggregateRel::_table_ =
        SegmentedAggregateRel::InternalGenerateParseTable_(SegmentedAggregateRel_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void SegmentedAggregateRel::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.substrait_ext.SegmentedAggregateRel)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.grouping_keys_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.segment_keys_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.measures_.Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL SegmentedAggregateRel::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const SegmentedAggregateRel& this_ = static_cast<const SegmentedAggregateRel&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL SegmentedAggregateRel::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const SegmentedAggregateRel& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.substrait_ext.SegmentedAggregateRel)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .substrait.Expression.FieldReference grouping_keys = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_grouping_keys_size());
         i < n; i++) {
      const auto& repfield = this_._internal_grouping_keys().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .substrait.Expression.FieldReference segment_keys = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_segment_keys_size());
         i < n; i++) {
      const auto& repfield = this_._internal_segment_keys().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              2, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .substrait.AggregateRel.Measure measures = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_measures_size());
         i < n; i++) {
      const auto& repfield = this_._internal_measures().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              3, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.substrait_ext.SegmentedAggregateRel)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t SegmentedAggregateRel::ByteSizeLong(const MessageLite& base) {
  const SegmentedAggregateRel& this_ = static_cast<const SegmentedAggregateRel&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t SegmentedAggregateRel::ByteSizeLong() const {
  const SegmentedAggregateRel& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.substrait_ext.SegmentedAggregateRel)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // repeated .substrait.Expression.FieldReference grouping_keys = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_grouping_keys_size();
      for (const auto& msg : this_._internal_grouping_keys()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .substrait.Expression.FieldReference segment_keys = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1UL * this_._internal_segment_keys_size();
      for (const auto& msg : this_._internal_segment_keys()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .substrait.AggregateRel.Measure measures = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1UL * this_._internal_measures_size();
      for (const auto& msg : this_._internal_measures()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void SegmentedAggregateRel::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<SegmentedAggregateRel*>(&to_msg);
  auto& from = static_cast<const SegmentedAggregateRel&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.substrait_ext.SegmentedAggregateRel)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_grouping_keys()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_grouping_keys());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_segment_keys()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_segment_keys());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_mutable_measures()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_measures());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void SegmentedAggregateRel::CopyFrom(const SegmentedAggregateRel& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.substrait_ext.SegmentedAggregateRel)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void SegmentedAggregateRel::InternalSwap(SegmentedAggregateRel* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.grouping_keys_.InternalSwap(&other->_impl_.grouping_keys_);
  _impl_.segment_keys_.InternalSwap(&other->_impl_.segment_keys_);
  _impl_.measures_.InternalSwap(&other->_impl_.measures_);
}

::google::protobuf::Metadata SegmentedAggregateRel::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace substrait_ext
}  // namespace arrow
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_substrait_2fextension_5frels_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
