// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: substrait/plan.proto
// Protobuf C++ Version: 7.35.1

#ifndef substrait_2fplan_2eproto_2epb_2eh
#define substrait_2fplan_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/unknown_field_set.h"
#include "substrait/algebra.pb.h"
#include "substrait/extensions/extensions.pb.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_substrait_2fplan_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_substrait_2fplan_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_substrait_2fplan_2eproto;
}  // extern "C"
namespace substrait {
class Plan;
struct PlanGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern PlanGlobalsTypeInternal Plan_globals_;
extern const ::google::protobuf::internal::ClassDataFull Plan_class_data_;
#else
extern const PlanGlobalsTypeInternal Plan_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class PlanRel;
struct PlanRelGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern PlanRelGlobalsTypeInternal PlanRel_globals_;
extern const ::google::protobuf::internal::ClassDataFull PlanRel_class_data_;
#else
extern const PlanRelGlobalsTypeInternal PlanRel_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class PlanVersion;
struct PlanVersionGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern PlanVersionGlobalsTypeInternal PlanVersion_globals_;
extern const ::google::protobuf::internal::ClassDataFull PlanVersion_class_data_;
#else
extern const PlanVersionGlobalsTypeInternal PlanVersion_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class Version;
struct VersionGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern VersionGlobalsTypeInternal Version_globals_;
extern const ::google::protobuf::internal::ClassDataFull Version_class_data_;
#else
extern const VersionGlobalsTypeInternal Version_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace substrait
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google

namespace substrait {

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED Version final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:substrait.Version) */ {
 public:
  inline Version() : Version(nullptr) {}
  ~Version() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(Version* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(Version));
  }
#endif

  template <typename = void>
  explicit constexpr Version(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline Version(const Version& from) : Version(nullptr, from) {}
  inline Version(Version&& from) noexcept : Version(nullptr, ::std::move(from)) {}
  inline Version& operator=(const Version& from) {
    CopyFrom(from);
    return *this;
  }
  inline Version& operator=(Version&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const Version& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<Version>(&Version_globals_);
  }
  static constexpr int kIndexInFileMessages = 3;
  friend void swap(Version& a, Version& b) { a.Swap(&b); }
  inline void Swap(Version* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(Version* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] Version* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<Version>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const Version& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const Version& from) { Version::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(Version* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "substrait.Version"; }

  explicit Version(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  Version(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Version& from);
  Version(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, Version&& from) noexcept
      : Version(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kGitHashFieldNumber = 4,
    kProducerFieldNumber = 5,
    kMajorNumberFieldNumber = 1,
    kMinorNumberFieldNumber = 2,
    kPatchNumberFieldNumber = 3,
  };
  // string git_hash = 4;
  void clear_git_hash() ;
  [[nodiscard]] const ::std::string& git_hash() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_git_hash(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_git_hash();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_git_hash();
  void set_allocated_git_hash(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_git_hash() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_git_hash(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_git_hash();

  public:
  // string producer = 5;
  void clear_producer() ;
  [[nodiscard]] const ::std::string& producer() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_producer(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_producer();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_producer();
  void set_allocated_producer(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_producer() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_producer(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_producer();

  public:
  // uint32 major_number = 1;
  void clear_major_number() ;
  [[nodiscard]] ::uint32_t major_number() const;
  void set_major_number(::uint32_t value);

  private:
  ::uint32_t _internal_major_number() const;
  void _internal_set_major_number(::uint32_t value);

  public:
  // uint32 minor_number = 2;
  void clear_minor_number() ;
  [[nodiscard]] ::uint32_t minor_number() const;
  void set_minor_number(::uint32_t value);

  private:
  ::uint32_t _internal_minor_number() const;
  void _internal_set_minor_number(::uint32_t value);

  public:
  // uint32 patch_number = 3;
  void clear_patch_number() ;
  [[nodiscard]] ::uint32_t patch_number() const;
  void set_patch_number(::uint32_t value);

  private:
  ::uint32_t _internal_patch_number() const;
  void _internal_set_patch_number(::uint32_t value);

  public:
  // @@protoc_insertion_point(class_scope:substrait.Version)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<3, 5,
                          0, 42,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const Version& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr git_hash_;
    ::google::protobuf::internal::ArenaStringPtr producer_;
    ::uint32_t major_number_;
    ::uint32_t minor_number_;
    ::uint32_t patch_number_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_substrait_2fplan_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED PlanVersion final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:substrait.PlanVersion) */ {
 public:
  inline PlanVersion() : PlanVersion(nullptr) {}
  ~PlanVersion() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(PlanVersion* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(PlanVersion));
  }
#endif

  template <typename = void>
  explicit constexpr PlanVersion(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline PlanVersion(const PlanVersion& from) : PlanVersion(nullptr, from) {}
  inline PlanVersion(PlanVersion&& from) noexcept : PlanVersion(nullptr, ::std::move(from)) {}
  inline PlanVersion& operator=(const PlanVersion& from) {
    CopyFrom(from);
    return *this;
  }
  inline PlanVersion& operator=(PlanVersion&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const PlanVersion& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<PlanVersion>(&PlanVersion_globals_);
  }
  static constexpr int kIndexInFileMessages = 2;
  friend void swap(PlanVersion& a, PlanVersion& b) { a.Swap(&b); }
  inline void Swap(PlanVersion* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(PlanVersion* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] PlanVersion* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<PlanVersion>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const PlanVersion& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const PlanVersion& from) { PlanVersion::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(PlanVersion* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "substrait.PlanVersion"; }

  explicit PlanVersion(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  PlanVersion(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const PlanVersion& from);
  PlanVersion(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, PlanVersion&& from) noexcept
      : PlanVersion(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kVersionFieldNumber = 6,
  };
  // .substrait.Version version = 6;
  [[nodiscard]] bool has_version()
      const;
  void clear_version() ;
  [[nodiscard]] const ::substrait::Version& version() const;
  [[nodiscard]] ::substrait::Version* PROTOBUF_NULLABLE release_version();
  ::substrait::Version* PROTOBUF_NONNULL mutable_version();
  void set_allocated_version(::substrait::Version* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_version(::substrait::Version* PROTOBUF_NULLABLE value);
  ::substrait::Version* PROTOBUF_NULLABLE unsafe_arena_release_version();

  private:
  const ::substrait::Version& _internal_version() const;
  ::substrait::Version* PROTOBUF_NONNULL _internal_mutable_version();

  public:
  // @@protoc_insertion_point(class_scope:substrait.PlanVersion)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<0, 1,
                          1, 0,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const PlanVersion& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::substrait::Version* PROTOBUF_NULLABLE version_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_substrait_2fplan_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED PlanRel final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:substrait.PlanRel) */ {
 public:
  inline PlanRel() : PlanRel(nullptr) {}
  ~PlanRel() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(PlanRel* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(PlanRel));
  }
#endif

  template <typename = void>
  explicit constexpr PlanRel(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline PlanRel(const PlanRel& from) : PlanRel(nullptr, from) {}
  inline PlanRel(PlanRel&& from) noexcept : PlanRel(nullptr, ::std::move(from)) {}
  inline PlanRel& operator=(const PlanRel& from) {
    CopyFrom(from);
    return *this;
  }
  inline PlanRel& operator=(PlanRel&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const PlanRel& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<PlanRel>(&PlanRel_globals_);
  }
  enum RelTypeCase {
    kRel = 1,
    kRoot = 2,
    REL_TYPE_NOT_SET = 0,
  };
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(PlanRel& a, PlanRel& b) { a.Swap(&b); }
  inline void Swap(PlanRel* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(PlanRel* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] PlanRel* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<PlanRel>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const PlanRel& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const PlanRel& from) { PlanRel::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(PlanRel* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "substrait.PlanRel"; }

  explicit PlanRel(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  PlanRel(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const PlanRel& from);
  PlanRel(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, PlanRel&& from) noexcept
      : PlanRel(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kRelFieldNumber = 1,
    kRootFieldNumber = 2,
  };
  // .substrait.Rel rel = 1;
  [[nodiscard]] bool has_rel()
      const;
  private:
  bool _internal_has_rel() const;

  public:
  void clear_rel() ;
  [[nodiscard]] const ::substrait::Rel& rel() const;
  [[nodiscard]] ::substrait::Rel* PROTOBUF_NULLABLE release_rel();
  ::substrait::Rel* PROTOBUF_NONNULL mutable_rel();
  void set_allocated_rel(::substrait::Rel* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_rel(::substrait::Rel* PROTOBUF_NULLABLE value);
  ::substrait::Rel* PROTOBUF_NULLABLE unsafe_arena_release_rel();

  private:
  const ::substrait::Rel& _internal_rel() const;
  ::substrait::Rel* PROTOBUF_NONNULL _internal_mutable_rel();

  public:
  // .substrait.RelRoot root = 2;
  [[nodiscard]] bool has_root()
      const;
  private:
  bool _internal_has_root() const;

  public:
  void clear_root() ;
  [[nodiscard]] const ::substrait::RelRoot& root() const;
  [[nodiscard]] ::substrait::RelRoot* PROTOBUF_NULLABLE release_root();
  ::substrait::RelRoot* PROTOBUF_NONNULL mutable_root();
  void set_allocated_root(::substrait::RelRoot* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_root(::substrait::RelRoot* PROTOBUF_NULLABLE value);
  ::substrait::RelRoot* PROTOBUF_NULLABLE unsafe_arena_release_root();

  private:
  const ::substrait::RelRoot& _internal_root() const;
  ::substrait::RelRoot* PROTOBUF_NONNULL _internal_mutable_root();

  public:
  void clear_rel_type();
  RelTypeCase rel_type_case() const;
  // @@protoc_insertion_point(class_scope:substrait.PlanRel)
 private:
  class _Internal;
  void set_has_rel();
  void set_has_root();
  [[nodiscard]] inline bool has_rel_type() const;
  inline void clear_has_rel_type();
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<0, 2,
                          2, 0,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const PlanRel& from_msg);
    union RelTypeUnion {
      constexpr RelTypeUnion() : _constinit_{} {}
      ::google::protobuf::internal::ConstantInitialized _constinit_;
      ::substrait::Rel* PROTOBUF_NULLABLE rel_;
      ::substrait::RelRoot* PROTOBUF_NULLABLE root_;
    } rel_type_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::uint32_t _oneof_case_[1];
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_substrait_2fplan_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED Plan final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:substrait.Plan) */ {
 public:
  inline Plan() : Plan(nullptr) {}
  ~Plan() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(Plan* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(Plan));
  }
#endif

  template <typename = void>
  explicit constexpr Plan(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline Plan(const Plan& from) : Plan(nullptr, from) {}
  inline Plan(Plan&& from) noexcept : Plan(nullptr, ::std::move(from)) {}
  inline Plan& operator=(const Plan& from) {
    CopyFrom(from);
    return *this;
  }
  inline Plan& operator=(Plan&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const Plan& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<Plan>(&Plan_globals_);
  }
  static constexpr int kIndexInFileMessages = 1;
  friend void swap(Plan& a, Plan& b) { a.Swap(&b); }
  inline void Swap(Plan* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(Plan* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] Plan* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<Plan>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const Plan& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const Plan& from) { Plan::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(Plan* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "substrait.Plan"; }

  explicit Plan(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  Plan(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Plan& from);
  Plan(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, Plan&& from) noexcept
      : Plan(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kExtensionUrisFieldNumber = 1,
    kExtensionsFieldNumber = 2,
    kRelationsFieldNumber = 3,
    kExpectedTypeUrlsFieldNumber = 5,
    kAdvancedExtensionsFieldNumber = 4,
    kVersionFieldNumber = 6,
  };
  // repeated .substrait.extensions.SimpleExtensionURI extension_uris = 1;
  [[nodiscard]] int extension_uris_size()
      const;
  private:
  int _internal_extension_uris_size() const;

  public:
  void clear_extension_uris() ;
  [[nodiscard]] const ::substrait::extensions::SimpleExtensionURI& extension_uris(int index) const;
  [[nodiscard]] ::substrait::extensions::SimpleExtensionURI* PROTOBUF_NONNULL mutable_extension_uris(int index);
  ::substrait::extensions::SimpleExtensionURI* PROTOBUF_NONNULL add_extension_uris();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>&
  extension_uris() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>* PROTOBUF_NONNULL
  mutable_extension_uris();

  private:
  const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>& _internal_extension_uris() const;
  ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>* PROTOBUF_NONNULL _internal_mutable_extension_uris();

  public:
  // repeated .substrait.extensions.SimpleExtensionDeclaration extensions = 2;
  [[nodiscard]] int extensions_size()
      const;
  private:
  int _internal_extensions_size() const;

  public:
  void clear_extensions() ;
  [[nodiscard]] const ::substrait::extensions::SimpleExtensionDeclaration& extensions(int index) const;
  [[nodiscard]] ::substrait::extensions::SimpleExtensionDeclaration* PROTOBUF_NONNULL mutable_extensions(int index);
  ::substrait::extensions::SimpleExtensionDeclaration* PROTOBUF_NONNULL add_extensions();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>&
  extensions() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>* PROTOBUF_NONNULL
  mutable_extensions();

  private:
  const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>& _internal_extensions() const;
  ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>* PROTOBUF_NONNULL _internal_mutable_extensions();

  public:
  // repeated .substrait.PlanRel relations = 3;
  [[nodiscard]] int relations_size()
      const;
  private:
  int _internal_relations_size() const;

  public:
  void clear_relations() ;
  [[nodiscard]] const ::substrait::PlanRel& relations(int index) const;
  [[nodiscard]] ::substrait::PlanRel* PROTOBUF_NONNULL mutable_relations(int index);
  ::substrait::PlanRel* PROTOBUF_NONNULL add_relations();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::substrait::PlanRel>&
  relations() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::substrait::PlanRel>* PROTOBUF_NONNULL
  mutable_relations();

  private:
  const ::google::protobuf::RepeatedPtrField<::substrait::PlanRel>& _internal_relations() const;
  ::google::protobuf::RepeatedPtrField<::substrait::PlanRel>* PROTOBUF_NONNULL _internal_mutable_relations();

  public:
  // repeated string expected_type_urls = 5;
  [[nodiscard]] int expected_type_urls_size()
      const;
  private:
  int _internal_expected_type_urls_size() const;

  public:
  void clear_expected_type_urls() ;
  [[nodiscard]] const ::std::string& expected_type_urls(int index) const;
  ::std::string* PROTOBUF_NONNULL mutable_expected_type_urls(int index);
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_expected_type_urls(int index, Arg_&& value, Args_... args);
  ::std::string* PROTOBUF_NONNULL add_expected_type_urls();
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void add_expected_type_urls(Arg_&& value, Args_... args);
  [[nodiscard]] const
      ::google::protobuf::RepeatedPtrField<::std::string>&
      expected_type_urls() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::std::string>*
      PROTOBUF_NONNULL
      mutable_expected_type_urls();

  private:
  const ::google::protobuf::RepeatedPtrField<::std::string>& _internal_expected_type_urls() const;
  ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL _internal_mutable_expected_type_urls();

  public:
  // .substrait.extensions.AdvancedExtension advanced_extensions = 4;
  [[nodiscard]] bool has_advanced_extensions()
      const;
  void clear_advanced_extensions() ;
  [[nodiscard]] const ::substrait::extensions::AdvancedExtension& advanced_extensions() const;
  [[nodiscard]] ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE release_advanced_extensions();
  ::substrait::extensions::AdvancedExtension* PROTOBUF_NONNULL mutable_advanced_extensions();
  void set_allocated_advanced_extensions(::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_advanced_extensions(::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE value);
  ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE unsafe_arena_release_advanced_extensions();

  private:
  const ::substrait::extensions::AdvancedExtension& _internal_advanced_extensions() const;
  ::substrait::extensions::AdvancedExtension* PROTOBUF_NONNULL _internal_mutable_advanced_extensions();

  public:
  // .substrait.Version version = 6;
  [[nodiscard]] bool has_version()
      const;
  void clear_version() ;
  [[nodiscard]] const ::substrait::Version& version() const;
  [[nodiscard]] ::substrait::Version* PROTOBUF_NULLABLE release_version();
  ::substrait::Version* PROTOBUF_NONNULL mutable_version();
  void set_allocated_version(::substrait::Version* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_version(::substrait::Version* PROTOBUF_NULLABLE value);
  ::substrait::Version* PROTOBUF_NULLABLE unsafe_arena_release_version();

  private:
  const ::substrait::Version& _internal_version() const;
  ::substrait::Version* PROTOBUF_NONNULL _internal_mutable_version();

  public:
  // @@protoc_insertion_point(class_scope:substrait.Plan)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<3, 6,
                          5, 41,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const Plan& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField< ::substrait::extensions::SimpleExtensionURI > extension_uris_;
    ::google::protobuf::RepeatedPtrField< ::substrait::extensions::SimpleExtensionDeclaration > extensions_;
    ::google::protobuf::RepeatedPtrField< ::substrait::PlanRel > relations_;
    ::google::protobuf::RepeatedPtrField<::std::string> expected_type_urls_;
    ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE advanced_extensions_;
    ::substrait::Version* PROTOBUF_NULLABLE version_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_substrait_2fplan_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// PlanRel

// .substrait.Rel rel = 1;
inline bool PlanRel::has_rel() const {
  return rel_type_case() == kRel;
}
inline bool PlanRel::_internal_has_rel() const {
  return rel_type_case() == kRel;
}
inline void PlanRel::set_has_rel() {
  _impl_._oneof_case_[0] = kRel;
}
inline ::substrait::Rel* PROTOBUF_NULLABLE PlanRel::release_rel() {
  // @@protoc_insertion_point(field_release:substrait.PlanRel.rel)
  if (rel_type_case() == kRel) {
    clear_has_rel_type();
    auto* temp = _impl_.rel_type_.rel_;
    if (GetArena() != nullptr) {
      temp = ::google::protobuf::internal::DuplicateIfNonNull(temp);
    }
    _impl_.rel_type_.rel_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline const ::substrait::Rel& PlanRel::_internal_rel() const {
  return rel_type_case() == kRel ? static_cast<const ::substrait::Rel&>(*_impl_.rel_type_.rel_)
                     : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::Rel>(&::substrait::Rel_globals_);
}
inline const ::substrait::Rel& PlanRel::rel() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.PlanRel.rel)
  return _internal_rel();
}
inline ::substrait::Rel* PROTOBUF_NULLABLE PlanRel::unsafe_arena_release_rel() {
  // @@protoc_insertion_point(field_unsafe_arena_release:substrait.PlanRel.rel)
  if (rel_type_case() == kRel) {
    clear_has_rel_type();
    auto* temp = _impl_.rel_type_.rel_;
    _impl_.rel_type_.rel_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline void PlanRel::unsafe_arena_set_allocated_rel(
    ::substrait::Rel* PROTOBUF_NULLABLE value) {
  // We rely on the oneof clear method to free the earlier contents
  // of this oneof. We can directly use the pointer we're given to
  // set the new value.
  clear_rel_type();
  if (value) {
    set_has_rel();
    _impl_.rel_type_.rel_ = value;
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.PlanRel.rel)
}
inline ::substrait::Rel* PROTOBUF_NONNULL PlanRel::_internal_mutable_rel() {
  if (rel_type_case() != kRel) {
    clear_rel_type();
    set_has_rel();
    _impl_.rel_type_.rel_ = 
        ::google::protobuf::Message::DefaultConstruct<::substrait::Rel>(GetArena());
  }
  return _impl_.rel_type_.rel_;
}
inline ::substrait::Rel* PROTOBUF_NONNULL PlanRel::mutable_rel()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::substrait::Rel* _msg = _internal_mutable_rel();
  // @@protoc_insertion_point(field_mutable:substrait.PlanRel.rel)
  return _msg;
}

// .substrait.RelRoot root = 2;
inline bool PlanRel::has_root() const {
  return rel_type_case() == kRoot;
}
inline bool PlanRel::_internal_has_root() const {
  return rel_type_case() == kRoot;
}
inline void PlanRel::set_has_root() {
  _impl_._oneof_case_[0] = kRoot;
}
inline ::substrait::RelRoot* PROTOBUF_NULLABLE PlanRel::release_root() {
  // @@protoc_insertion_point(field_release:substrait.PlanRel.root)
  if (rel_type_case() == kRoot) {
    clear_has_rel_type();
    auto* temp = _impl_.rel_type_.root_;
    if (GetArena() != nullptr) {
      temp = ::google::protobuf::internal::DuplicateIfNonNull(temp);
    }
    _impl_.rel_type_.root_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline const ::substrait::RelRoot& PlanRel::_internal_root() const {
  return rel_type_case() == kRoot ? static_cast<const ::substrait::RelRoot&>(*_impl_.rel_type_.root_)
                     : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::RelRoot>(&::substrait::RelRoot_globals_);
}
inline const ::substrait::RelRoot& PlanRel::root() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.PlanRel.root)
  return _internal_root();
}
inline ::substrait::RelRoot* PROTOBUF_NULLABLE PlanRel::unsafe_arena_release_root() {
  // @@protoc_insertion_point(field_unsafe_arena_release:substrait.PlanRel.root)
  if (rel_type_case() == kRoot) {
    clear_has_rel_type();
    auto* temp = _impl_.rel_type_.root_;
    _impl_.rel_type_.root_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline void PlanRel::unsafe_arena_set_allocated_root(
    ::substrait::RelRoot* PROTOBUF_NULLABLE value) {
  // We rely on the oneof clear method to free the earlier contents
  // of this oneof. We can directly use the pointer we're given to
  // set the new value.
  clear_rel_type();
  if (value) {
    set_has_root();
    _impl_.rel_type_.root_ = value;
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.PlanRel.root)
}
inline ::substrait::RelRoot* PROTOBUF_NONNULL PlanRel::_internal_mutable_root() {
  if (rel_type_case() != kRoot) {
    clear_rel_type();
    set_has_root();
    _impl_.rel_type_.root_ = 
        ::google::protobuf::Message::DefaultConstruct<::substrait::RelRoot>(GetArena());
  }
  return _impl_.rel_type_.root_;
}
inline ::substrait::RelRoot* PROTOBUF_NONNULL PlanRel::mutable_root()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::substrait::RelRoot* _msg = _internal_mutable_root();
  // @@protoc_insertion_point(field_mutable:substrait.PlanRel.root)
  return _msg;
}

inline bool PlanRel::has_rel_type() const {
  return rel_type_case() != REL_TYPE_NOT_SET;
}
inline void PlanRel::clear_has_rel_type() {
  _impl_._oneof_case_[0] = REL_TYPE_NOT_SET;
}
inline PlanRel::RelTypeCase PlanRel::rel_type_case() const {
  return PlanRel::RelTypeCase(_impl_._oneof_case_[0]);
}
// -------------------------------------------------------------------

// Plan

// .substrait.Version version = 6;
inline bool Plan::has_version() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000020U);
  PROTOBUF_ASSUME(!value || _impl_.version_ != nullptr);
  return value;
}
inline void Plan::clear_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.version_ != nullptr) _impl_.version_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
}
inline const ::substrait::Version& Plan::_internal_version() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::substrait::Version* p = _impl_.version_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::Version>(&::substrait::Version_globals_);
}
inline const ::substrait::Version& Plan::version() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.Plan.version)
  return _internal_version();
}
inline void Plan::unsafe_arena_set_allocated_version(
    ::substrait::Version* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.version_);
  }
  _impl_.version_ = reinterpret_cast<::substrait::Version*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.Plan.version)
}
inline ::substrait::Version* PROTOBUF_NULLABLE Plan::release_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
  ::substrait::Version* released = _impl_.version_;
  _impl_.version_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::substrait::Version* PROTOBUF_NULLABLE Plan::unsafe_arena_release_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:substrait.Plan.version)

  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
  ::substrait::Version* temp = _impl_.version_;
  _impl_.version_ = nullptr;
  return temp;
}
inline ::substrait::Version* PROTOBUF_NONNULL Plan::_internal_mutable_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.version_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::substrait::Version>(GetArena());
    _impl_.version_ = reinterpret_cast<::substrait::Version*>(p);
  }
  return _impl_.version_;
}
inline ::substrait::Version* PROTOBUF_NONNULL Plan::mutable_version()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  ::substrait::Version* _msg = _internal_mutable_version();
  // @@protoc_insertion_point(field_mutable:substrait.Plan.version)
  return _msg;
}
inline void Plan::set_allocated_version(::substrait::Version* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.version_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = value->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
  }

  _impl_.version_ = reinterpret_cast<::substrait::Version*>(value);
  // @@protoc_insertion_point(field_set_allocated:substrait.Plan.version)
}

// repeated .substrait.extensions.SimpleExtensionURI extension_uris = 1;
inline int Plan::_internal_extension_uris_size() const {
  return _internal_extension_uris().size();
}
inline int Plan::extension_uris_size() const {
  return _internal_extension_uris_size();
}
inline const ::substrait::extensions::SimpleExtensionURI& Plan::extension_uris(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.Plan.extension_uris)
  return _internal_extension_uris().Get(index);
}
inline ::substrait::extensions::SimpleExtensionURI* PROTOBUF_NONNULL Plan::mutable_extension_uris(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:substrait.Plan.extension_uris)
  return _internal_mutable_extension_uris()->Mutable(index);
}
inline ::substrait::extensions::SimpleExtensionURI* PROTOBUF_NONNULL Plan::add_extension_uris()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::substrait::extensions::SimpleExtensionURI* _add =
      _internal_mutable_extension_uris()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:substrait.Plan.extension_uris)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>& Plan::extension_uris() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:substrait.Plan.extension_uris)
  return _internal_extension_uris();
}
inline ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>* PROTOBUF_NONNULL
Plan::mutable_extension_uris() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:substrait.Plan.extension_uris)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_extension_uris();
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>&
Plan::_internal_extension_uris() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.extension_uris_;
}
inline ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>* PROTOBUF_NONNULL
Plan::_internal_mutable_extension_uris() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.extension_uris_;
}

// repeated .substrait.extensions.SimpleExtensionDeclaration extensions = 2;
inline int Plan::_internal_extensions_size() const {
  return _internal_extensions().size();
}
inline int Plan::extensions_size() const {
  return _internal_extensions_size();
}
inline const ::substrait::extensions::SimpleExtensionDeclaration& Plan::extensions(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.Plan.extensions)
  return _internal_extensions().Get(index);
}
inline ::substrait::extensions::SimpleExtensionDeclaration* PROTOBUF_NONNULL Plan::mutable_extensions(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:substrait.Plan.extensions)
  return _internal_mutable_extensions()->Mutable(index);
}
inline ::substrait::extensions::SimpleExtensionDeclaration* PROTOBUF_NONNULL Plan::add_extensions()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::substrait::extensions::SimpleExtensionDeclaration* _add =
      _internal_mutable_extensions()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  // @@protoc_insertion_point(field_add:substrait.Plan.extensions)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>& Plan::extensions() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:substrait.Plan.extensions)
  return _internal_extensions();
}
inline ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>* PROTOBUF_NONNULL
Plan::mutable_extensions() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  // @@protoc_insertion_point(field_mutable_list:substrait.Plan.extensions)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_extensions();
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>&
Plan::_internal_extensions() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.extensions_;
}
inline ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>* PROTOBUF_NONNULL
Plan::_internal_mutable_extensions() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.extensions_;
}

// repeated .substrait.PlanRel relations = 3;
inline int Plan::_internal_relations_size() const {
  return _internal_relations().size();
}
inline int Plan::relations_size() const {
  return _internal_relations_size();
}
inline void Plan::clear_relations() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.relations_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline const ::substrait::PlanRel& Plan::relations(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.Plan.relations)
  return _internal_relations().Get(index);
}
inline ::substrait::PlanRel* PROTOBUF_NONNULL Plan::mutable_relations(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:substrait.Plan.relations)
  return _internal_mutable_relations()->Mutable(index);
}
inline ::substrait::PlanRel* PROTOBUF_NONNULL Plan::add_relations()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::substrait::PlanRel* _add =
      _internal_mutable_relations()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_add:substrait.Plan.relations)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::PlanRel>& Plan::relations() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:substrait.Plan.relations)
  return _internal_relations();
}
inline ::google::protobuf::RepeatedPtrField<::substrait::PlanRel>* PROTOBUF_NONNULL
Plan::mutable_relations() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_mutable_list:substrait.Plan.relations)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_relations();
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::PlanRel>&
Plan::_internal_relations() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.relations_;
}
inline ::google::protobuf::RepeatedPtrField<::substrait::PlanRel>* PROTOBUF_NONNULL
Plan::_internal_mutable_relations() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.relations_;
}

// .substrait.extensions.AdvancedExtension advanced_extensions = 4;
inline bool Plan::has_advanced_extensions() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000010U);
  PROTOBUF_ASSUME(!value || _impl_.advanced_extensions_ != nullptr);
  return value;
}
inline const ::substrait::extensions::AdvancedExtension& Plan::_internal_advanced_extensions() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::substrait::extensions::AdvancedExtension* p = _impl_.advanced_extensions_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::extensions::AdvancedExtension>(&::substrait::extensions::AdvancedExtension_globals_);
}
inline const ::substrait::extensions::AdvancedExtension& Plan::advanced_extensions() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.Plan.advanced_extensions)
  return _internal_advanced_extensions();
}
inline void Plan::unsafe_arena_set_allocated_advanced_extensions(
    ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.advanced_extensions_);
  }
  _impl_.advanced_extensions_ = reinterpret_cast<::substrait::extensions::AdvancedExtension*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.Plan.advanced_extensions)
}
inline ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE Plan::release_advanced_extensions() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::substrait::extensions::AdvancedExtension* released = _impl_.advanced_extensions_;
  _impl_.advanced_extensions_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE Plan::unsafe_arena_release_advanced_extensions() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:substrait.Plan.advanced_extensions)

  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::substrait::extensions::AdvancedExtension* temp = _impl_.advanced_extensions_;
  _impl_.advanced_extensions_ = nullptr;
  return temp;
}
inline ::substrait::extensions::AdvancedExtension* PROTOBUF_NONNULL Plan::_internal_mutable_advanced_extensions() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.advanced_extensions_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::substrait::extensions::AdvancedExtension>(GetArena());
    _impl_.advanced_extensions_ = reinterpret_cast<::substrait::extensions::AdvancedExtension*>(p);
  }
  return _impl_.advanced_extensions_;
}
inline ::substrait::extensions::AdvancedExtension* PROTOBUF_NONNULL Plan::mutable_advanced_extensions()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::substrait::extensions::AdvancedExtension* _msg = _internal_mutable_advanced_extensions();
  // @@protoc_insertion_point(field_mutable:substrait.Plan.advanced_extensions)
  return _msg;
}
inline void Plan::set_allocated_advanced_extensions(::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.advanced_extensions_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  }

  _impl_.advanced_extensions_ = reinterpret_cast<::substrait::extensions::AdvancedExtension*>(value);
  // @@protoc_insertion_point(field_set_allocated:substrait.Plan.advanced_extensions)
}

// repeated string expected_type_urls = 5;
inline int Plan::_internal_expected_type_urls_size() const {
  return _internal_expected_type_urls().size();
}
inline int Plan::expected_type_urls_size() const {
  return _internal_expected_type_urls_size();
}
inline void Plan::clear_expected_type_urls() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.expected_type_urls_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline ::std::string* PROTOBUF_NONNULL Plan::add_expected_type_urls()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::std::string* _s =
      _internal_mutable_expected_type_urls()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_add_mutable:substrait.Plan.expected_type_urls)
  return _s;
}
inline const ::std::string& Plan::expected_type_urls(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.Plan.expected_type_urls)
  return _internal_expected_type_urls().Get(index);
}
inline ::std::string* PROTOBUF_NONNULL Plan::mutable_expected_type_urls(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:substrait.Plan.expected_type_urls)
  return _internal_mutable_expected_type_urls()->Mutable(index);
}
template <typename Arg_, typename... Args_>
inline void Plan::set_expected_type_urls(int index, Arg_&& value, Args_... args) {
  ::google::protobuf::internal::AssignToString(
      *_internal_mutable_expected_type_urls()->Mutable(index),
      ::std::forward<Arg_>(value), args... );
  // @@protoc_insertion_point(field_set:substrait.Plan.expected_type_urls)
}
template <typename Arg_, typename... Args_>
inline void Plan::add_expected_type_urls(Arg_&& value, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::google::protobuf::internal::AddToRepeatedPtrField(
      ::google::protobuf::MessageLite::internal_visibility(), GetArena(),
      *_internal_mutable_expected_type_urls(), ::std::forward<Arg_>(value),
      args... );
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_add:substrait.Plan.expected_type_urls)
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>& Plan::expected_type_urls()
    const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:substrait.Plan.expected_type_urls)
  return _internal_expected_type_urls();
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
Plan::mutable_expected_type_urls() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_mutable_list:substrait.Plan.expected_type_urls)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_expected_type_urls();
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>&
Plan::_internal_expected_type_urls() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.expected_type_urls_;
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
Plan::_internal_mutable_expected_type_urls() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.expected_type_urls_;
}

// -------------------------------------------------------------------

// PlanVersion

// .substrait.Version version = 6;
inline bool PlanVersion::has_version() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000001U);
  PROTOBUF_ASSUME(!value || _impl_.version_ != nullptr);
  return value;
}
inline void PlanVersion::clear_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.version_ != nullptr) _impl_.version_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::substrait::Version& PlanVersion::_internal_version() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::substrait::Version* p = _impl_.version_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::Version>(&::substrait::Version_globals_);
}
inline const ::substrait::Version& PlanVersion::version() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.PlanVersion.version)
  return _internal_version();
}
inline void PlanVersion::unsafe_arena_set_allocated_version(
    ::substrait::Version* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.version_);
  }
  _impl_.version_ = reinterpret_cast<::substrait::Version*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.PlanVersion.version)
}
inline ::substrait::Version* PROTOBUF_NULLABLE PlanVersion::release_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::substrait::Version* released = _impl_.version_;
  _impl_.version_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::substrait::Version* PROTOBUF_NULLABLE PlanVersion::unsafe_arena_release_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:substrait.PlanVersion.version)

  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::substrait::Version* temp = _impl_.version_;
  _impl_.version_ = nullptr;
  return temp;
}
inline ::substrait::Version* PROTOBUF_NONNULL PlanVersion::_internal_mutable_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.version_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::substrait::Version>(GetArena());
    _impl_.version_ = reinterpret_cast<::substrait::Version*>(p);
  }
  return _impl_.version_;
}
inline ::substrait::Version* PROTOBUF_NONNULL PlanVersion::mutable_version()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::substrait::Version* _msg = _internal_mutable_version();
  // @@protoc_insertion_point(field_mutable:substrait.PlanVersion.version)
  return _msg;
}
inline void PlanVersion::set_allocated_version(::substrait::Version* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.version_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = value->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }

  _impl_.version_ = reinterpret_cast<::substrait::Version*>(value);
  // @@protoc_insertion_point(field_set_allocated:substrait.PlanVersion.version)
}

// -------------------------------------------------------------------

// Version

// uint32 major_number = 1;
inline void Version::clear_major_number() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.major_number_ = 0u;
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline ::uint32_t Version::major_number() const {
  // @@protoc_insertion_point(field_get:substrait.Version.major_number)
  return _internal_major_number();
}
inline void Version::set_major_number(::uint32_t value) {
  _internal_set_major_number(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_set:substrait.Version.major_number)
}
inline ::uint32_t Version::_internal_major_number() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.major_number_;
}
inline void Version::_internal_set_major_number(::uint32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.major_number_ = value;
}

// uint32 minor_number = 2;
inline void Version::clear_minor_number() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.minor_number_ = 0u;
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline ::uint32_t Version::minor_number() const {
  // @@protoc_insertion_point(field_get:substrait.Version.minor_number)
  return _internal_minor_number();
}
inline void Version::set_minor_number(::uint32_t value) {
  _internal_set_minor_number(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_set:substrait.Version.minor_number)
}
inline ::uint32_t Version::_internal_minor_number() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.minor_number_;
}
inline void Version::_internal_set_minor_number(::uint32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.minor_number_ = value;
}

// uint32 patch_number = 3;
inline void Version::clear_patch_number() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.patch_number_ = 0u;
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
}
inline ::uint32_t Version::patch_number() const {
  // @@protoc_insertion_point(field_get:substrait.Version.patch_number)
  return _internal_patch_number();
}
inline void Version::set_patch_number(::uint32_t value) {
  _internal_set_patch_number(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  // @@protoc_insertion_point(field_set:substrait.Version.patch_number)
}
inline ::uint32_t Version::_internal_patch_number() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.patch_number_;
}
inline void Version::_internal_set_patch_number(::uint32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.patch_number_ = value;
}

// string git_hash = 4;
inline void Version::clear_git_hash() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.git_hash_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& Version::git_hash() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.Version.git_hash)
  return _internal_git_hash();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void Version::set_git_hash(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.git_hash_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:substrait.Version.git_hash)
}
inline ::std::string* PROTOBUF_NONNULL Version::mutable_git_hash()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_git_hash();
  // @@protoc_insertion_point(field_mutable:substrait.Version.git_hash)
  return _s;
}
inline const ::std::string& Version::_internal_git_hash() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.git_hash_.Get();
}
inline void Version::_internal_set_git_hash(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.git_hash_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL Version::_internal_mutable_git_hash() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.git_hash_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE Version::release_git_hash() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:substrait.Version.git_hash)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.git_hash_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.git_hash_.Set("", GetArena());
  }
  return released;
}
inline void Version::set_allocated_git_hash(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.git_hash_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.git_hash_.IsDefault()) {
    _impl_.git_hash_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:substrait.Version.git_hash)
}

// string producer = 5;
inline void Version::clear_producer() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.producer_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& Version::producer() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.Version.producer)
  return _internal_producer();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void Version::set_producer(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.producer_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:substrait.Version.producer)
}
inline ::std::string* PROTOBUF_NONNULL Version::mutable_producer()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_producer();
  // @@protoc_insertion_point(field_mutable:substrait.Version.producer)
  return _s;
}
inline const ::std::string& Version::_internal_producer() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.producer_.Get();
}
inline void Version::_internal_set_producer(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.producer_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL Version::_internal_mutable_producer() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.producer_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE Version::release_producer() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:substrait.Version.producer)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.producer_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.producer_.Set("", GetArena());
  }
  return released;
}
inline void Version::set_allocated_producer(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.producer_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.producer_.IsDefault()) {
    _impl_.producer_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:substrait.Version.producer)
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace substrait


// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // substrait_2fplan_2eproto_2epb_2eh
