// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: substrait/extended_expression.proto
// Protobuf C++ Version: 7.35.1

#ifndef substrait_2fextended_5fexpression_2eproto_2epb_2eh
#define substrait_2fextended_5fexpression_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/unknown_field_set.h"
#include "substrait/algebra.pb.h"
#include "substrait/extensions/extensions.pb.h"
#include "substrait/plan.pb.h"
#include "substrait/type.pb.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_substrait_2fextended_5fexpression_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_substrait_2fextended_5fexpression_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_substrait_2fextended_5fexpression_2eproto;
}  // extern "C"
namespace substrait {
class ExpressionReference;
struct ExpressionReferenceGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern ExpressionReferenceGlobalsTypeInternal ExpressionReference_globals_;
extern const ::google::protobuf::internal::ClassDataFull ExpressionReference_class_data_;
#else
extern const ExpressionReferenceGlobalsTypeInternal ExpressionReference_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class ExtendedExpression;
struct ExtendedExpressionGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern ExtendedExpressionGlobalsTypeInternal ExtendedExpression_globals_;
extern const ::google::protobuf::internal::ClassDataFull ExtendedExpression_class_data_;
#else
extern const ExtendedExpressionGlobalsTypeInternal ExtendedExpression_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace substrait
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google

namespace substrait {

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED ExpressionReference final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:substrait.ExpressionReference) */ {
 public:
  inline ExpressionReference() : ExpressionReference(nullptr) {}
  ~ExpressionReference() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(ExpressionReference* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(ExpressionReference));
  }
#endif

  template <typename = void>
  explicit constexpr ExpressionReference(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline ExpressionReference(const ExpressionReference& from) : ExpressionReference(nullptr, from) {}
  inline ExpressionReference(ExpressionReference&& from) noexcept : ExpressionReference(nullptr, ::std::move(from)) {}
  inline ExpressionReference& operator=(const ExpressionReference& from) {
    CopyFrom(from);
    return *this;
  }
  inline ExpressionReference& operator=(ExpressionReference&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const ExpressionReference& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<ExpressionReference>(&ExpressionReference_globals_);
  }
  enum ExprTypeCase {
    kExpression = 1,
    kMeasure = 2,
    EXPR_TYPE_NOT_SET = 0,
  };
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(ExpressionReference& a, ExpressionReference& b) { a.Swap(&b); }
  inline void Swap(ExpressionReference* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(ExpressionReference* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] ExpressionReference* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<ExpressionReference>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const ExpressionReference& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const ExpressionReference& from) { ExpressionReference::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(ExpressionReference* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "substrait.ExpressionReference"; }

  explicit ExpressionReference(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  ExpressionReference(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ExpressionReference& from);
  ExpressionReference(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, ExpressionReference&& from) noexcept
      : ExpressionReference(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kOutputNamesFieldNumber = 3,
    kExpressionFieldNumber = 1,
    kMeasureFieldNumber = 2,
  };
  // repeated string output_names = 3;
  [[nodiscard]] int output_names_size()
      const;
  private:
  int _internal_output_names_size() const;

  public:
  void clear_output_names() ;
  [[nodiscard]] const ::std::string& output_names(int index) const;
  ::std::string* PROTOBUF_NONNULL mutable_output_names(int index);
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_output_names(int index, Arg_&& value, Args_... args);
  ::std::string* PROTOBUF_NONNULL add_output_names();
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void add_output_names(Arg_&& value, Args_... args);
  [[nodiscard]] const
      ::google::protobuf::RepeatedPtrField<::std::string>&
      output_names() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::std::string>*
      PROTOBUF_NONNULL
      mutable_output_names();

  private:
  const ::google::protobuf::RepeatedPtrField<::std::string>& _internal_output_names() const;
  ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL _internal_mutable_output_names();

  public:
  // .substrait.Expression expression = 1;
  [[nodiscard]] bool has_expression()
      const;
  private:
  bool _internal_has_expression() const;

  public:
  void clear_expression() ;
  [[nodiscard]] const ::substrait::Expression& expression() const;
  [[nodiscard]] ::substrait::Expression* PROTOBUF_NULLABLE release_expression();
  ::substrait::Expression* PROTOBUF_NONNULL mutable_expression();
  void set_allocated_expression(::substrait::Expression* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_expression(::substrait::Expression* PROTOBUF_NULLABLE value);
  ::substrait::Expression* PROTOBUF_NULLABLE unsafe_arena_release_expression();

  private:
  const ::substrait::Expression& _internal_expression() const;
  ::substrait::Expression* PROTOBUF_NONNULL _internal_mutable_expression();

  public:
  // .substrait.AggregateFunction measure = 2;
  [[nodiscard]] bool has_measure()
      const;
  private:
  bool _internal_has_measure() const;

  public:
  void clear_measure() ;
  [[nodiscard]] const ::substrait::AggregateFunction& measure() const;
  [[nodiscard]] ::substrait::AggregateFunction* PROTOBUF_NULLABLE release_measure();
  ::substrait::AggregateFunction* PROTOBUF_NONNULL mutable_measure();
  void set_allocated_measure(::substrait::AggregateFunction* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_measure(::substrait::AggregateFunction* PROTOBUF_NULLABLE value);
  ::substrait::AggregateFunction* PROTOBUF_NULLABLE unsafe_arena_release_measure();

  private:
  const ::substrait::AggregateFunction& _internal_measure() const;
  ::substrait::AggregateFunction* PROTOBUF_NONNULL _internal_mutable_measure();

  public:
  void clear_expr_type();
  ExprTypeCase expr_type_case() const;
  // @@protoc_insertion_point(class_scope:substrait.ExpressionReference)
 private:
  class _Internal;
  void set_has_expression();
  void set_has_measure();
  [[nodiscard]] inline bool has_expr_type() const;
  inline void clear_has_expr_type();
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<0, 3,
                          2, 50,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const ExpressionReference& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField<::std::string> output_names_;
    union ExprTypeUnion {
      constexpr ExprTypeUnion() : _constinit_{} {}
      ::google::protobuf::internal::ConstantInitialized _constinit_;
      ::substrait::Expression* PROTOBUF_NULLABLE expression_;
      ::substrait::AggregateFunction* PROTOBUF_NULLABLE measure_;
    } expr_type_;
    ::uint32_t _oneof_case_[1];
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_substrait_2fextended_5fexpression_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED ExtendedExpression final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:substrait.ExtendedExpression) */ {
 public:
  inline ExtendedExpression() : ExtendedExpression(nullptr) {}
  ~ExtendedExpression() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(ExtendedExpression* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(ExtendedExpression));
  }
#endif

  template <typename = void>
  explicit constexpr ExtendedExpression(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline ExtendedExpression(const ExtendedExpression& from) : ExtendedExpression(nullptr, from) {}
  inline ExtendedExpression(ExtendedExpression&& from) noexcept : ExtendedExpression(nullptr, ::std::move(from)) {}
  inline ExtendedExpression& operator=(const ExtendedExpression& from) {
    CopyFrom(from);
    return *this;
  }
  inline ExtendedExpression& operator=(ExtendedExpression&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const ExtendedExpression& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<ExtendedExpression>(&ExtendedExpression_globals_);
  }
  static constexpr int kIndexInFileMessages = 1;
  friend void swap(ExtendedExpression& a, ExtendedExpression& b) { a.Swap(&b); }
  inline void Swap(ExtendedExpression* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(ExtendedExpression* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] ExtendedExpression* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<ExtendedExpression>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const ExtendedExpression& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const ExtendedExpression& from) { ExtendedExpression::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(ExtendedExpression* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "substrait.ExtendedExpression"; }

  explicit ExtendedExpression(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  ExtendedExpression(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ExtendedExpression& from);
  ExtendedExpression(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, ExtendedExpression&& from) noexcept
      : ExtendedExpression(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kExtensionUrisFieldNumber = 1,
    kExtensionsFieldNumber = 2,
    kReferredExprFieldNumber = 3,
    kExpectedTypeUrlsFieldNumber = 6,
    kBaseSchemaFieldNumber = 4,
    kAdvancedExtensionsFieldNumber = 5,
    kVersionFieldNumber = 7,
  };
  // repeated .substrait.extensions.SimpleExtensionURI extension_uris = 1;
  [[nodiscard]] int extension_uris_size()
      const;
  private:
  int _internal_extension_uris_size() const;

  public:
  void clear_extension_uris() ;
  [[nodiscard]] const ::substrait::extensions::SimpleExtensionURI& extension_uris(int index) const;
  [[nodiscard]] ::substrait::extensions::SimpleExtensionURI* PROTOBUF_NONNULL mutable_extension_uris(int index);
  ::substrait::extensions::SimpleExtensionURI* PROTOBUF_NONNULL add_extension_uris();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>&
  extension_uris() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>* PROTOBUF_NONNULL
  mutable_extension_uris();

  private:
  const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>& _internal_extension_uris() const;
  ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>* PROTOBUF_NONNULL _internal_mutable_extension_uris();

  public:
  // repeated .substrait.extensions.SimpleExtensionDeclaration extensions = 2;
  [[nodiscard]] int extensions_size()
      const;
  private:
  int _internal_extensions_size() const;

  public:
  void clear_extensions() ;
  [[nodiscard]] const ::substrait::extensions::SimpleExtensionDeclaration& extensions(int index) const;
  [[nodiscard]] ::substrait::extensions::SimpleExtensionDeclaration* PROTOBUF_NONNULL mutable_extensions(int index);
  ::substrait::extensions::SimpleExtensionDeclaration* PROTOBUF_NONNULL add_extensions();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>&
  extensions() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>* PROTOBUF_NONNULL
  mutable_extensions();

  private:
  const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>& _internal_extensions() const;
  ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>* PROTOBUF_NONNULL _internal_mutable_extensions();

  public:
  // repeated .substrait.ExpressionReference referred_expr = 3;
  [[nodiscard]] int referred_expr_size()
      const;
  private:
  int _internal_referred_expr_size() const;

  public:
  void clear_referred_expr() ;
  [[nodiscard]] const ::substrait::ExpressionReference& referred_expr(int index) const;
  [[nodiscard]] ::substrait::ExpressionReference* PROTOBUF_NONNULL mutable_referred_expr(int index);
  ::substrait::ExpressionReference* PROTOBUF_NONNULL add_referred_expr();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::substrait::ExpressionReference>&
  referred_expr() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::substrait::ExpressionReference>* PROTOBUF_NONNULL
  mutable_referred_expr();

  private:
  const ::google::protobuf::RepeatedPtrField<::substrait::ExpressionReference>& _internal_referred_expr() const;
  ::google::protobuf::RepeatedPtrField<::substrait::ExpressionReference>* PROTOBUF_NONNULL _internal_mutable_referred_expr();

  public:
  // repeated string expected_type_urls = 6;
  [[nodiscard]] int expected_type_urls_size()
      const;
  private:
  int _internal_expected_type_urls_size() const;

  public:
  void clear_expected_type_urls() ;
  [[nodiscard]] const ::std::string& expected_type_urls(int index) const;
  ::std::string* PROTOBUF_NONNULL mutable_expected_type_urls(int index);
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_expected_type_urls(int index, Arg_&& value, Args_... args);
  ::std::string* PROTOBUF_NONNULL add_expected_type_urls();
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void add_expected_type_urls(Arg_&& value, Args_... args);
  [[nodiscard]] const
      ::google::protobuf::RepeatedPtrField<::std::string>&
      expected_type_urls() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::std::string>*
      PROTOBUF_NONNULL
      mutable_expected_type_urls();

  private:
  const ::google::protobuf::RepeatedPtrField<::std::string>& _internal_expected_type_urls() const;
  ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL _internal_mutable_expected_type_urls();

  public:
  // .substrait.NamedStruct base_schema = 4;
  [[nodiscard]] bool has_base_schema()
      const;
  void clear_base_schema() ;
  [[nodiscard]] const ::substrait::NamedStruct& base_schema() const;
  [[nodiscard]] ::substrait::NamedStruct* PROTOBUF_NULLABLE release_base_schema();
  ::substrait::NamedStruct* PROTOBUF_NONNULL mutable_base_schema();
  void set_allocated_base_schema(::substrait::NamedStruct* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_base_schema(::substrait::NamedStruct* PROTOBUF_NULLABLE value);
  ::substrait::NamedStruct* PROTOBUF_NULLABLE unsafe_arena_release_base_schema();

  private:
  const ::substrait::NamedStruct& _internal_base_schema() const;
  ::substrait::NamedStruct* PROTOBUF_NONNULL _internal_mutable_base_schema();

  public:
  // .substrait.extensions.AdvancedExtension advanced_extensions = 5;
  [[nodiscard]] bool has_advanced_extensions()
      const;
  void clear_advanced_extensions() ;
  [[nodiscard]] const ::substrait::extensions::AdvancedExtension& advanced_extensions() const;
  [[nodiscard]] ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE release_advanced_extensions();
  ::substrait::extensions::AdvancedExtension* PROTOBUF_NONNULL mutable_advanced_extensions();
  void set_allocated_advanced_extensions(::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_advanced_extensions(::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE value);
  ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE unsafe_arena_release_advanced_extensions();

  private:
  const ::substrait::extensions::AdvancedExtension& _internal_advanced_extensions() const;
  ::substrait::extensions::AdvancedExtension* PROTOBUF_NONNULL _internal_mutable_advanced_extensions();

  public:
  // .substrait.Version version = 7;
  [[nodiscard]] bool has_version()
      const;
  void clear_version() ;
  [[nodiscard]] const ::substrait::Version& version() const;
  [[nodiscard]] ::substrait::Version* PROTOBUF_NULLABLE release_version();
  ::substrait::Version* PROTOBUF_NONNULL mutable_version();
  void set_allocated_version(::substrait::Version* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_version(::substrait::Version* PROTOBUF_NULLABLE value);
  ::substrait::Version* PROTOBUF_NULLABLE unsafe_arena_release_version();

  private:
  const ::substrait::Version& _internal_version() const;
  ::substrait::Version* PROTOBUF_NONNULL _internal_mutable_version();

  public:
  // @@protoc_insertion_point(class_scope:substrait.ExtendedExpression)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<3, 7,
                          6, 55,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const ExtendedExpression& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField< ::substrait::extensions::SimpleExtensionURI > extension_uris_;
    ::google::protobuf::RepeatedPtrField< ::substrait::extensions::SimpleExtensionDeclaration > extensions_;
    ::google::protobuf::RepeatedPtrField< ::substrait::ExpressionReference > referred_expr_;
    ::google::protobuf::RepeatedPtrField<::std::string> expected_type_urls_;
    ::substrait::NamedStruct* PROTOBUF_NULLABLE base_schema_;
    ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE advanced_extensions_;
    ::substrait::Version* PROTOBUF_NULLABLE version_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_substrait_2fextended_5fexpression_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// ExpressionReference

// .substrait.Expression expression = 1;
inline bool ExpressionReference::has_expression() const {
  return expr_type_case() == kExpression;
}
inline bool ExpressionReference::_internal_has_expression() const {
  return expr_type_case() == kExpression;
}
inline void ExpressionReference::set_has_expression() {
  _impl_._oneof_case_[0] = kExpression;
}
inline ::substrait::Expression* PROTOBUF_NULLABLE ExpressionReference::release_expression() {
  // @@protoc_insertion_point(field_release:substrait.ExpressionReference.expression)
  if (expr_type_case() == kExpression) {
    clear_has_expr_type();
    auto* temp = _impl_.expr_type_.expression_;
    if (GetArena() != nullptr) {
      temp = ::google::protobuf::internal::DuplicateIfNonNull(temp);
    }
    _impl_.expr_type_.expression_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline const ::substrait::Expression& ExpressionReference::_internal_expression() const {
  return expr_type_case() == kExpression ? static_cast<const ::substrait::Expression&>(*_impl_.expr_type_.expression_)
                     : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::Expression>(&::substrait::Expression_globals_);
}
inline const ::substrait::Expression& ExpressionReference::expression() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExpressionReference.expression)
  return _internal_expression();
}
inline ::substrait::Expression* PROTOBUF_NULLABLE ExpressionReference::unsafe_arena_release_expression() {
  // @@protoc_insertion_point(field_unsafe_arena_release:substrait.ExpressionReference.expression)
  if (expr_type_case() == kExpression) {
    clear_has_expr_type();
    auto* temp = _impl_.expr_type_.expression_;
    _impl_.expr_type_.expression_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline void ExpressionReference::unsafe_arena_set_allocated_expression(
    ::substrait::Expression* PROTOBUF_NULLABLE value) {
  // We rely on the oneof clear method to free the earlier contents
  // of this oneof. We can directly use the pointer we're given to
  // set the new value.
  clear_expr_type();
  if (value) {
    set_has_expression();
    _impl_.expr_type_.expression_ = value;
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.ExpressionReference.expression)
}
inline ::substrait::Expression* PROTOBUF_NONNULL ExpressionReference::_internal_mutable_expression() {
  if (expr_type_case() != kExpression) {
    clear_expr_type();
    set_has_expression();
    _impl_.expr_type_.expression_ = 
        ::google::protobuf::Message::DefaultConstruct<::substrait::Expression>(GetArena());
  }
  return _impl_.expr_type_.expression_;
}
inline ::substrait::Expression* PROTOBUF_NONNULL ExpressionReference::mutable_expression()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::substrait::Expression* _msg = _internal_mutable_expression();
  // @@protoc_insertion_point(field_mutable:substrait.ExpressionReference.expression)
  return _msg;
}

// .substrait.AggregateFunction measure = 2;
inline bool ExpressionReference::has_measure() const {
  return expr_type_case() == kMeasure;
}
inline bool ExpressionReference::_internal_has_measure() const {
  return expr_type_case() == kMeasure;
}
inline void ExpressionReference::set_has_measure() {
  _impl_._oneof_case_[0] = kMeasure;
}
inline ::substrait::AggregateFunction* PROTOBUF_NULLABLE ExpressionReference::release_measure() {
  // @@protoc_insertion_point(field_release:substrait.ExpressionReference.measure)
  if (expr_type_case() == kMeasure) {
    clear_has_expr_type();
    auto* temp = _impl_.expr_type_.measure_;
    if (GetArena() != nullptr) {
      temp = ::google::protobuf::internal::DuplicateIfNonNull(temp);
    }
    _impl_.expr_type_.measure_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline const ::substrait::AggregateFunction& ExpressionReference::_internal_measure() const {
  return expr_type_case() == kMeasure ? static_cast<const ::substrait::AggregateFunction&>(*_impl_.expr_type_.measure_)
                     : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::AggregateFunction>(&::substrait::AggregateFunction_globals_);
}
inline const ::substrait::AggregateFunction& ExpressionReference::measure() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExpressionReference.measure)
  return _internal_measure();
}
inline ::substrait::AggregateFunction* PROTOBUF_NULLABLE ExpressionReference::unsafe_arena_release_measure() {
  // @@protoc_insertion_point(field_unsafe_arena_release:substrait.ExpressionReference.measure)
  if (expr_type_case() == kMeasure) {
    clear_has_expr_type();
    auto* temp = _impl_.expr_type_.measure_;
    _impl_.expr_type_.measure_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline void ExpressionReference::unsafe_arena_set_allocated_measure(
    ::substrait::AggregateFunction* PROTOBUF_NULLABLE value) {
  // We rely on the oneof clear method to free the earlier contents
  // of this oneof. We can directly use the pointer we're given to
  // set the new value.
  clear_expr_type();
  if (value) {
    set_has_measure();
    _impl_.expr_type_.measure_ = value;
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.ExpressionReference.measure)
}
inline ::substrait::AggregateFunction* PROTOBUF_NONNULL ExpressionReference::_internal_mutable_measure() {
  if (expr_type_case() != kMeasure) {
    clear_expr_type();
    set_has_measure();
    _impl_.expr_type_.measure_ = 
        ::google::protobuf::Message::DefaultConstruct<::substrait::AggregateFunction>(GetArena());
  }
  return _impl_.expr_type_.measure_;
}
inline ::substrait::AggregateFunction* PROTOBUF_NONNULL ExpressionReference::mutable_measure()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::substrait::AggregateFunction* _msg = _internal_mutable_measure();
  // @@protoc_insertion_point(field_mutable:substrait.ExpressionReference.measure)
  return _msg;
}

// repeated string output_names = 3;
inline int ExpressionReference::_internal_output_names_size() const {
  return _internal_output_names().size();
}
inline int ExpressionReference::output_names_size() const {
  return _internal_output_names_size();
}
inline void ExpressionReference::clear_output_names() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.output_names_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline ::std::string* PROTOBUF_NONNULL ExpressionReference::add_output_names()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::std::string* _s =
      _internal_mutable_output_names()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add_mutable:substrait.ExpressionReference.output_names)
  return _s;
}
inline const ::std::string& ExpressionReference::output_names(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExpressionReference.output_names)
  return _internal_output_names().Get(index);
}
inline ::std::string* PROTOBUF_NONNULL ExpressionReference::mutable_output_names(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:substrait.ExpressionReference.output_names)
  return _internal_mutable_output_names()->Mutable(index);
}
template <typename Arg_, typename... Args_>
inline void ExpressionReference::set_output_names(int index, Arg_&& value, Args_... args) {
  ::google::protobuf::internal::AssignToString(
      *_internal_mutable_output_names()->Mutable(index),
      ::std::forward<Arg_>(value), args... );
  // @@protoc_insertion_point(field_set:substrait.ExpressionReference.output_names)
}
template <typename Arg_, typename... Args_>
inline void ExpressionReference::add_output_names(Arg_&& value, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::google::protobuf::internal::AddToRepeatedPtrField(
      ::google::protobuf::MessageLite::internal_visibility(), GetArena(),
      *_internal_mutable_output_names(), ::std::forward<Arg_>(value),
      args... );
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:substrait.ExpressionReference.output_names)
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>& ExpressionReference::output_names()
    const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:substrait.ExpressionReference.output_names)
  return _internal_output_names();
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
ExpressionReference::mutable_output_names() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:substrait.ExpressionReference.output_names)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_output_names();
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>&
ExpressionReference::_internal_output_names() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.output_names_;
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
ExpressionReference::_internal_mutable_output_names() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.output_names_;
}

inline bool ExpressionReference::has_expr_type() const {
  return expr_type_case() != EXPR_TYPE_NOT_SET;
}
inline void ExpressionReference::clear_has_expr_type() {
  _impl_._oneof_case_[0] = EXPR_TYPE_NOT_SET;
}
inline ExpressionReference::ExprTypeCase ExpressionReference::expr_type_case() const {
  return ExpressionReference::ExprTypeCase(_impl_._oneof_case_[0]);
}
// -------------------------------------------------------------------

// ExtendedExpression

// .substrait.Version version = 7;
inline bool ExtendedExpression::has_version() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000040U);
  PROTOBUF_ASSUME(!value || _impl_.version_ != nullptr);
  return value;
}
inline const ::substrait::Version& ExtendedExpression::_internal_version() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::substrait::Version* p = _impl_.version_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::Version>(&::substrait::Version_globals_);
}
inline const ::substrait::Version& ExtendedExpression::version() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExtendedExpression.version)
  return _internal_version();
}
inline void ExtendedExpression::unsafe_arena_set_allocated_version(
    ::substrait::Version* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.version_);
  }
  _impl_.version_ = reinterpret_cast<::substrait::Version*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000040U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000040U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.ExtendedExpression.version)
}
inline ::substrait::Version* PROTOBUF_NULLABLE ExtendedExpression::release_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000040U);
  ::substrait::Version* released = _impl_.version_;
  _impl_.version_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::substrait::Version* PROTOBUF_NULLABLE ExtendedExpression::unsafe_arena_release_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:substrait.ExtendedExpression.version)

  ClearHasBit(_impl_._has_bits_[0], 0x00000040U);
  ::substrait::Version* temp = _impl_.version_;
  _impl_.version_ = nullptr;
  return temp;
}
inline ::substrait::Version* PROTOBUF_NONNULL ExtendedExpression::_internal_mutable_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.version_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::substrait::Version>(GetArena());
    _impl_.version_ = reinterpret_cast<::substrait::Version*>(p);
  }
  return _impl_.version_;
}
inline ::substrait::Version* PROTOBUF_NONNULL ExtendedExpression::mutable_version()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000040U);
  ::substrait::Version* _msg = _internal_mutable_version();
  // @@protoc_insertion_point(field_mutable:substrait.ExtendedExpression.version)
  return _msg;
}
inline void ExtendedExpression::set_allocated_version(::substrait::Version* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.version_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000040U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000040U);
  }

  _impl_.version_ = reinterpret_cast<::substrait::Version*>(value);
  // @@protoc_insertion_point(field_set_allocated:substrait.ExtendedExpression.version)
}

// repeated .substrait.extensions.SimpleExtensionURI extension_uris = 1;
inline int ExtendedExpression::_internal_extension_uris_size() const {
  return _internal_extension_uris().size();
}
inline int ExtendedExpression::extension_uris_size() const {
  return _internal_extension_uris_size();
}
inline const ::substrait::extensions::SimpleExtensionURI& ExtendedExpression::extension_uris(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExtendedExpression.extension_uris)
  return _internal_extension_uris().Get(index);
}
inline ::substrait::extensions::SimpleExtensionURI* PROTOBUF_NONNULL ExtendedExpression::mutable_extension_uris(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:substrait.ExtendedExpression.extension_uris)
  return _internal_mutable_extension_uris()->Mutable(index);
}
inline ::substrait::extensions::SimpleExtensionURI* PROTOBUF_NONNULL ExtendedExpression::add_extension_uris()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::substrait::extensions::SimpleExtensionURI* _add =
      _internal_mutable_extension_uris()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:substrait.ExtendedExpression.extension_uris)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>& ExtendedExpression::extension_uris() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:substrait.ExtendedExpression.extension_uris)
  return _internal_extension_uris();
}
inline ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>* PROTOBUF_NONNULL
ExtendedExpression::mutable_extension_uris() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:substrait.ExtendedExpression.extension_uris)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_extension_uris();
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>&
ExtendedExpression::_internal_extension_uris() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.extension_uris_;
}
inline ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionURI>* PROTOBUF_NONNULL
ExtendedExpression::_internal_mutable_extension_uris() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.extension_uris_;
}

// repeated .substrait.extensions.SimpleExtensionDeclaration extensions = 2;
inline int ExtendedExpression::_internal_extensions_size() const {
  return _internal_extensions().size();
}
inline int ExtendedExpression::extensions_size() const {
  return _internal_extensions_size();
}
inline const ::substrait::extensions::SimpleExtensionDeclaration& ExtendedExpression::extensions(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExtendedExpression.extensions)
  return _internal_extensions().Get(index);
}
inline ::substrait::extensions::SimpleExtensionDeclaration* PROTOBUF_NONNULL ExtendedExpression::mutable_extensions(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:substrait.ExtendedExpression.extensions)
  return _internal_mutable_extensions()->Mutable(index);
}
inline ::substrait::extensions::SimpleExtensionDeclaration* PROTOBUF_NONNULL ExtendedExpression::add_extensions()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::substrait::extensions::SimpleExtensionDeclaration* _add =
      _internal_mutable_extensions()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  // @@protoc_insertion_point(field_add:substrait.ExtendedExpression.extensions)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>& ExtendedExpression::extensions() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:substrait.ExtendedExpression.extensions)
  return _internal_extensions();
}
inline ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>* PROTOBUF_NONNULL
ExtendedExpression::mutable_extensions() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  // @@protoc_insertion_point(field_mutable_list:substrait.ExtendedExpression.extensions)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_extensions();
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>&
ExtendedExpression::_internal_extensions() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.extensions_;
}
inline ::google::protobuf::RepeatedPtrField<::substrait::extensions::SimpleExtensionDeclaration>* PROTOBUF_NONNULL
ExtendedExpression::_internal_mutable_extensions() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.extensions_;
}

// repeated .substrait.ExpressionReference referred_expr = 3;
inline int ExtendedExpression::_internal_referred_expr_size() const {
  return _internal_referred_expr().size();
}
inline int ExtendedExpression::referred_expr_size() const {
  return _internal_referred_expr_size();
}
inline void ExtendedExpression::clear_referred_expr() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.referred_expr_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline const ::substrait::ExpressionReference& ExtendedExpression::referred_expr(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExtendedExpression.referred_expr)
  return _internal_referred_expr().Get(index);
}
inline ::substrait::ExpressionReference* PROTOBUF_NONNULL ExtendedExpression::mutable_referred_expr(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:substrait.ExtendedExpression.referred_expr)
  return _internal_mutable_referred_expr()->Mutable(index);
}
inline ::substrait::ExpressionReference* PROTOBUF_NONNULL ExtendedExpression::add_referred_expr()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::substrait::ExpressionReference* _add =
      _internal_mutable_referred_expr()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_add:substrait.ExtendedExpression.referred_expr)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::ExpressionReference>& ExtendedExpression::referred_expr() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:substrait.ExtendedExpression.referred_expr)
  return _internal_referred_expr();
}
inline ::google::protobuf::RepeatedPtrField<::substrait::ExpressionReference>* PROTOBUF_NONNULL
ExtendedExpression::mutable_referred_expr() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_mutable_list:substrait.ExtendedExpression.referred_expr)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_referred_expr();
}
inline const ::google::protobuf::RepeatedPtrField<::substrait::ExpressionReference>&
ExtendedExpression::_internal_referred_expr() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.referred_expr_;
}
inline ::google::protobuf::RepeatedPtrField<::substrait::ExpressionReference>* PROTOBUF_NONNULL
ExtendedExpression::_internal_mutable_referred_expr() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.referred_expr_;
}

// .substrait.NamedStruct base_schema = 4;
inline bool ExtendedExpression::has_base_schema() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000010U);
  PROTOBUF_ASSUME(!value || _impl_.base_schema_ != nullptr);
  return value;
}
inline const ::substrait::NamedStruct& ExtendedExpression::_internal_base_schema() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::substrait::NamedStruct* p = _impl_.base_schema_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::NamedStruct>(&::substrait::NamedStruct_globals_);
}
inline const ::substrait::NamedStruct& ExtendedExpression::base_schema() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExtendedExpression.base_schema)
  return _internal_base_schema();
}
inline void ExtendedExpression::unsafe_arena_set_allocated_base_schema(
    ::substrait::NamedStruct* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.base_schema_);
  }
  _impl_.base_schema_ = reinterpret_cast<::substrait::NamedStruct*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.ExtendedExpression.base_schema)
}
inline ::substrait::NamedStruct* PROTOBUF_NULLABLE ExtendedExpression::release_base_schema() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::substrait::NamedStruct* released = _impl_.base_schema_;
  _impl_.base_schema_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::substrait::NamedStruct* PROTOBUF_NULLABLE ExtendedExpression::unsafe_arena_release_base_schema() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:substrait.ExtendedExpression.base_schema)

  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::substrait::NamedStruct* temp = _impl_.base_schema_;
  _impl_.base_schema_ = nullptr;
  return temp;
}
inline ::substrait::NamedStruct* PROTOBUF_NONNULL ExtendedExpression::_internal_mutable_base_schema() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.base_schema_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::substrait::NamedStruct>(GetArena());
    _impl_.base_schema_ = reinterpret_cast<::substrait::NamedStruct*>(p);
  }
  return _impl_.base_schema_;
}
inline ::substrait::NamedStruct* PROTOBUF_NONNULL ExtendedExpression::mutable_base_schema()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::substrait::NamedStruct* _msg = _internal_mutable_base_schema();
  // @@protoc_insertion_point(field_mutable:substrait.ExtendedExpression.base_schema)
  return _msg;
}
inline void ExtendedExpression::set_allocated_base_schema(::substrait::NamedStruct* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.base_schema_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  }

  _impl_.base_schema_ = reinterpret_cast<::substrait::NamedStruct*>(value);
  // @@protoc_insertion_point(field_set_allocated:substrait.ExtendedExpression.base_schema)
}

// .substrait.extensions.AdvancedExtension advanced_extensions = 5;
inline bool ExtendedExpression::has_advanced_extensions() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000020U);
  PROTOBUF_ASSUME(!value || _impl_.advanced_extensions_ != nullptr);
  return value;
}
inline const ::substrait::extensions::AdvancedExtension& ExtendedExpression::_internal_advanced_extensions() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::substrait::extensions::AdvancedExtension* p = _impl_.advanced_extensions_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::substrait::extensions::AdvancedExtension>(&::substrait::extensions::AdvancedExtension_globals_);
}
inline const ::substrait::extensions::AdvancedExtension& ExtendedExpression::advanced_extensions() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExtendedExpression.advanced_extensions)
  return _internal_advanced_extensions();
}
inline void ExtendedExpression::unsafe_arena_set_allocated_advanced_extensions(
    ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.advanced_extensions_);
  }
  _impl_.advanced_extensions_ = reinterpret_cast<::substrait::extensions::AdvancedExtension*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:substrait.ExtendedExpression.advanced_extensions)
}
inline ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE ExtendedExpression::release_advanced_extensions() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
  ::substrait::extensions::AdvancedExtension* released = _impl_.advanced_extensions_;
  _impl_.advanced_extensions_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE ExtendedExpression::unsafe_arena_release_advanced_extensions() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:substrait.ExtendedExpression.advanced_extensions)

  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
  ::substrait::extensions::AdvancedExtension* temp = _impl_.advanced_extensions_;
  _impl_.advanced_extensions_ = nullptr;
  return temp;
}
inline ::substrait::extensions::AdvancedExtension* PROTOBUF_NONNULL ExtendedExpression::_internal_mutable_advanced_extensions() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.advanced_extensions_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::substrait::extensions::AdvancedExtension>(GetArena());
    _impl_.advanced_extensions_ = reinterpret_cast<::substrait::extensions::AdvancedExtension*>(p);
  }
  return _impl_.advanced_extensions_;
}
inline ::substrait::extensions::AdvancedExtension* PROTOBUF_NONNULL ExtendedExpression::mutable_advanced_extensions()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  ::substrait::extensions::AdvancedExtension* _msg = _internal_mutable_advanced_extensions();
  // @@protoc_insertion_point(field_mutable:substrait.ExtendedExpression.advanced_extensions)
  return _msg;
}
inline void ExtendedExpression::set_allocated_advanced_extensions(::substrait::extensions::AdvancedExtension* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.advanced_extensions_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
  }

  _impl_.advanced_extensions_ = reinterpret_cast<::substrait::extensions::AdvancedExtension*>(value);
  // @@protoc_insertion_point(field_set_allocated:substrait.ExtendedExpression.advanced_extensions)
}

// repeated string expected_type_urls = 6;
inline int ExtendedExpression::_internal_expected_type_urls_size() const {
  return _internal_expected_type_urls().size();
}
inline int ExtendedExpression::expected_type_urls_size() const {
  return _internal_expected_type_urls_size();
}
inline void ExtendedExpression::clear_expected_type_urls() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.expected_type_urls_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline ::std::string* PROTOBUF_NONNULL ExtendedExpression::add_expected_type_urls()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::std::string* _s =
      _internal_mutable_expected_type_urls()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_add_mutable:substrait.ExtendedExpression.expected_type_urls)
  return _s;
}
inline const ::std::string& ExtendedExpression::expected_type_urls(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:substrait.ExtendedExpression.expected_type_urls)
  return _internal_expected_type_urls().Get(index);
}
inline ::std::string* PROTOBUF_NONNULL ExtendedExpression::mutable_expected_type_urls(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:substrait.ExtendedExpression.expected_type_urls)
  return _internal_mutable_expected_type_urls()->Mutable(index);
}
template <typename Arg_, typename... Args_>
inline void ExtendedExpression::set_expected_type_urls(int index, Arg_&& value, Args_... args) {
  ::google::protobuf::internal::AssignToString(
      *_internal_mutable_expected_type_urls()->Mutable(index),
      ::std::forward<Arg_>(value), args... );
  // @@protoc_insertion_point(field_set:substrait.ExtendedExpression.expected_type_urls)
}
template <typename Arg_, typename... Args_>
inline void ExtendedExpression::add_expected_type_urls(Arg_&& value, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::google::protobuf::internal::AddToRepeatedPtrField(
      ::google::protobuf::MessageLite::internal_visibility(), GetArena(),
      *_internal_mutable_expected_type_urls(), ::std::forward<Arg_>(value),
      args... );
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_add:substrait.ExtendedExpression.expected_type_urls)
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>& ExtendedExpression::expected_type_urls()
    const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:substrait.ExtendedExpression.expected_type_urls)
  return _internal_expected_type_urls();
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
ExtendedExpression::mutable_expected_type_urls() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_mutable_list:substrait.ExtendedExpression.expected_type_urls)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_expected_type_urls();
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>&
ExtendedExpression::_internal_expected_type_urls() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.expected_type_urls_;
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
ExtendedExpression::_internal_mutable_expected_type_urls() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.expected_type_urls_;
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace substrait


// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // substrait_2fextended_5fexpression_2eproto_2epb_2eh
