// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: FlightSql.proto
// Protobuf C++ Version: 7.35.1

#include "FlightSql.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::arrow::flight::protocol::sql::CommandGetSqlInfo
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandGetXdbcTypeInfo
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandGetCatalogs
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandGetDbSchemas
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandGetTables
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandGetTableTypes
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandGetPrimaryKeys
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandGetExportedKeys
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandGetImportedKeys
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandGetCrossReference
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionCreatePreparedStatementRequest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::SubstraitPlan
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionCreatePreparedSubstraitPlanRequest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionCreatePreparedStatementResult
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionClosePreparedStatementRequest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionBeginTransactionRequest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionBeginSavepointRequest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionBeginTransactionResult
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionBeginSavepointResult
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionEndTransactionRequest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionEndSavepointRequest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandStatementQuery
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandStatementSubstraitPlan
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::TicketStatementQuery
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandPreparedStatementQuery
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandStatementUpdate
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandPreparedStatementUpdate
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandStatementIngest_OptionsEntry_DoNotUse
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::CommandStatementIngest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::DoPutUpdateResult
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::DoPutPreparedStatementResult
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionCancelQueryRequest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
        // ::arrow::flight::protocol::sql::ActionCancelQueryResult
        {&::_pbi::kDescriptorMethods, &::descriptor_table_FlightSql_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace arrow {
namespace flight {
namespace protocol {
namespace sql {
class TicketStatementQuery::_Internal {
 public:
  using HasBits = decltype(::std::declval<TicketStatementQuery>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(TicketStatementQuery, _impl_._has_bits_);
};

constexpr TicketStatementQuery::ParseTableT_ TicketStatementQuery::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(TicketStatementQuery, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::TicketStatementQuery>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bytes statement_handle = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(TicketStatementQuery, _impl_.statement_handle_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes statement_handle = 1;
      {PROTOBUF_FIELD_OFFSET(TicketStatementQuery, _impl_.statement_handle_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr TicketStatementQuery::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        statement_handle_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr TicketStatementQuery::TicketStatementQuery(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL TicketStatementQuery::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) TicketStatementQuery(arena);
}
constexpr auto TicketStatementQuery::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(TicketStatementQuery), alignof(TicketStatementQuery));
}
constexpr auto TicketStatementQuery::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &TicketStatementQuery::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<TicketStatementQuery>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &TicketStatementQuery::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<TicketStatementQuery>(), &TicketStatementQuery::ByteSizeLong,
              &TicketStatementQuery::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(TicketStatementQuery, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[23],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct TicketStatementQueryGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr TicketStatementQueryGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 TicketStatementQuery_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(TicketStatementQuery::InternalGenerateClassData_(
            _default, &TicketStatementQuery_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<TicketStatementQuery>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~TicketStatementQueryGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) TicketStatementQuery _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<TicketStatementQuery>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(TicketStatementQueryGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST TicketStatementQueryGlobalsTypeInternal TicketStatementQuery_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* TicketStatementQuery_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return TicketStatementQuery_globals_.GetClassData();
#else
  return TicketStatementQuery_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class SubstraitPlan::_Internal {
 public:
  using HasBits = decltype(::std::declval<SubstraitPlan>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(SubstraitPlan, _impl_._has_bits_);
};

constexpr SubstraitPlan::ParseTableT_ SubstraitPlan::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(SubstraitPlan, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::SubstraitPlan>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // string version = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(SubstraitPlan, _impl_.version_)}},
      // bytes plan = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(SubstraitPlan, _impl_.plan_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes plan = 1;
      {PROTOBUF_FIELD_OFFSET(SubstraitPlan, _impl_.plan_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // string version = 2;
      {PROTOBUF_FIELD_OFFSET(SubstraitPlan, _impl_.version_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\47\0\7\0\0\0\0\0"
      "arrow.flight.protocol.sql.SubstraitPlan"
      "version"
    }},
  };
}


inline constexpr SubstraitPlan::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        plan_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        version_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr SubstraitPlan::SubstraitPlan(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL SubstraitPlan::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) SubstraitPlan(arena);
}
constexpr auto SubstraitPlan::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(SubstraitPlan), alignof(SubstraitPlan));
}
constexpr auto SubstraitPlan::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &SubstraitPlan::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<SubstraitPlan>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &SubstraitPlan::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<SubstraitPlan>(), &SubstraitPlan::ByteSizeLong,
              &SubstraitPlan::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(SubstraitPlan, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[11],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct SubstraitPlanGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr SubstraitPlanGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 SubstraitPlan_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(SubstraitPlan::InternalGenerateClassData_(
            _default, &SubstraitPlan_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<SubstraitPlan>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~SubstraitPlanGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) SubstraitPlan _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<SubstraitPlan>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(SubstraitPlanGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST SubstraitPlanGlobalsTypeInternal SubstraitPlan_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* SubstraitPlan_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return SubstraitPlan_globals_.GetClassData();
#else
  return SubstraitPlan_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class DoPutUpdateResult::_Internal {
 public:
  using HasBits = decltype(::std::declval<DoPutUpdateResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(DoPutUpdateResult, _impl_._has_bits_);
};

constexpr DoPutUpdateResult::ParseTableT_ DoPutUpdateResult::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(DoPutUpdateResult, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::DoPutUpdateResult>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // int64 record_count = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(DoPutUpdateResult, _impl_.record_count_), 0>(),
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(DoPutUpdateResult, _impl_.record_count_)}},
    }}, {{
      65535, 65535
    }}, {{
      // int64 record_count = 1;
      {PROTOBUF_FIELD_OFFSET(DoPutUpdateResult, _impl_.record_count_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt64)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr DoPutUpdateResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        record_count_{::int64_t{0}} {}

template <typename>
constexpr DoPutUpdateResult::DoPutUpdateResult(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL DoPutUpdateResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) DoPutUpdateResult(arena);
}
constexpr auto DoPutUpdateResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(DoPutUpdateResult), alignof(DoPutUpdateResult));
}
constexpr auto DoPutUpdateResult::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &DoPutUpdateResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<DoPutUpdateResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &DoPutUpdateResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<DoPutUpdateResult>(), &DoPutUpdateResult::ByteSizeLong,
              &DoPutUpdateResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(DoPutUpdateResult, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[30],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct DoPutUpdateResultGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr DoPutUpdateResultGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 DoPutUpdateResult_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(DoPutUpdateResult::InternalGenerateClassData_(
            _default, &DoPutUpdateResult_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<DoPutUpdateResult>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~DoPutUpdateResultGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) DoPutUpdateResult _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<DoPutUpdateResult>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(DoPutUpdateResultGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST DoPutUpdateResultGlobalsTypeInternal DoPutUpdateResult_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* DoPutUpdateResult_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return DoPutUpdateResult_globals_.GetClassData();
#else
  return DoPutUpdateResult_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class DoPutPreparedStatementResult::_Internal {
 public:
  using HasBits = decltype(::std::declval<DoPutPreparedStatementResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(DoPutPreparedStatementResult, _impl_._has_bits_);
};

constexpr DoPutPreparedStatementResult::ParseTableT_ DoPutPreparedStatementResult::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(DoPutPreparedStatementResult, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::DoPutPreparedStatementResult>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional bytes prepared_statement_handle = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(DoPutPreparedStatementResult, _impl_.prepared_statement_handle_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional bytes prepared_statement_handle = 1;
      {PROTOBUF_FIELD_OFFSET(DoPutPreparedStatementResult, _impl_.prepared_statement_handle_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr DoPutPreparedStatementResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        prepared_statement_handle_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr DoPutPreparedStatementResult::DoPutPreparedStatementResult(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL DoPutPreparedStatementResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) DoPutPreparedStatementResult(arena);
}
constexpr auto DoPutPreparedStatementResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(DoPutPreparedStatementResult), alignof(DoPutPreparedStatementResult));
}
constexpr auto DoPutPreparedStatementResult::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &DoPutPreparedStatementResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<DoPutPreparedStatementResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &DoPutPreparedStatementResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<DoPutPreparedStatementResult>(), &DoPutPreparedStatementResult::ByteSizeLong,
              &DoPutPreparedStatementResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(DoPutPreparedStatementResult, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[31],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct DoPutPreparedStatementResultGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr DoPutPreparedStatementResultGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 DoPutPreparedStatementResult_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(DoPutPreparedStatementResult::InternalGenerateClassData_(
            _default, &DoPutPreparedStatementResult_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<DoPutPreparedStatementResult>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~DoPutPreparedStatementResultGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) DoPutPreparedStatementResult _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<DoPutPreparedStatementResult>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(DoPutPreparedStatementResultGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST DoPutPreparedStatementResultGlobalsTypeInternal DoPutPreparedStatementResult_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* DoPutPreparedStatementResult_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return DoPutPreparedStatementResult_globals_.GetClassData();
#else
  return DoPutPreparedStatementResult_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandStatementUpdate::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandStatementUpdate>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandStatementUpdate, _impl_._has_bits_);
};

constexpr CommandStatementUpdate::ParseTableT_ CommandStatementUpdate::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandStatementUpdate, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandStatementUpdate>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional bytes transaction_id = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementUpdate, _impl_.transaction_id_)}},
      // string query = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementUpdate, _impl_.query_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string query = 1;
      {PROTOBUF_FIELD_OFFSET(CommandStatementUpdate, _impl_.query_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional bytes transaction_id = 2;
      {PROTOBUF_FIELD_OFFSET(CommandStatementUpdate, _impl_.transaction_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\60\5\0\0\0\0\0\0"
      "arrow.flight.protocol.sql.CommandStatementUpdate"
      "query"
    }},
  };
}


inline constexpr CommandStatementUpdate::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        query_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        transaction_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CommandStatementUpdate::CommandStatementUpdate(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandStatementUpdate::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandStatementUpdate(arena);
}
constexpr auto CommandStatementUpdate::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandStatementUpdate), alignof(CommandStatementUpdate));
}
constexpr auto CommandStatementUpdate::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandStatementUpdate::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandStatementUpdate>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandStatementUpdate::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandStatementUpdate>(), &CommandStatementUpdate::ByteSizeLong,
              &CommandStatementUpdate::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandStatementUpdate, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[25],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandStatementUpdateGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandStatementUpdateGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandStatementUpdate_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandStatementUpdate::InternalGenerateClassData_(
            _default, &CommandStatementUpdate_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementUpdate>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandStatementUpdateGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandStatementUpdate _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementUpdate>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandStatementUpdateGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandStatementUpdateGlobalsTypeInternal CommandStatementUpdate_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandStatementUpdate_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandStatementUpdate_globals_.GetClassData();
#else
  return CommandStatementUpdate_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandStatementQuery::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandStatementQuery>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandStatementQuery, _impl_._has_bits_);
};

constexpr CommandStatementQuery::ParseTableT_ CommandStatementQuery::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandStatementQuery, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandStatementQuery>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional bytes transaction_id = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementQuery, _impl_.transaction_id_)}},
      // string query = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementQuery, _impl_.query_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string query = 1;
      {PROTOBUF_FIELD_OFFSET(CommandStatementQuery, _impl_.query_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional bytes transaction_id = 2;
      {PROTOBUF_FIELD_OFFSET(CommandStatementQuery, _impl_.transaction_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\57\5\0\0\0\0\0\0"
      "arrow.flight.protocol.sql.CommandStatementQuery"
      "query"
    }},
  };
}


inline constexpr CommandStatementQuery::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        query_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        transaction_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CommandStatementQuery::CommandStatementQuery(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandStatementQuery::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandStatementQuery(arena);
}
constexpr auto CommandStatementQuery::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandStatementQuery), alignof(CommandStatementQuery));
}
constexpr auto CommandStatementQuery::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandStatementQuery::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandStatementQuery>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandStatementQuery::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandStatementQuery>(), &CommandStatementQuery::ByteSizeLong,
              &CommandStatementQuery::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandStatementQuery, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[21],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandStatementQueryGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandStatementQueryGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandStatementQuery_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandStatementQuery::InternalGenerateClassData_(
            _default, &CommandStatementQuery_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementQuery>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandStatementQueryGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandStatementQuery _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementQuery>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandStatementQueryGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandStatementQueryGlobalsTypeInternal CommandStatementQuery_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandStatementQuery_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandStatementQuery_globals_.GetClassData();
#else
  return CommandStatementQuery_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandStatementIngest_TableDefinitionOptions::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandStatementIngest_TableDefinitionOptions>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptions, _impl_._has_bits_);
};

constexpr CommandStatementIngest_TableDefinitionOptions::ParseTableT_ CommandStatementIngest_TableDefinitionOptions::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptions, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions.TableExistsOption if_exists = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(CommandStatementIngest_TableDefinitionOptions, _impl_.if_exists_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptions, _impl_.if_exists_)}},
      // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions.TableNotExistOption if_not_exist = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(CommandStatementIngest_TableDefinitionOptions, _impl_.if_not_exist_), 0>(),
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptions, _impl_.if_not_exist_)}},
    }}, {{
      65535, 65535
    }}, {{
      // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions.TableNotExistOption if_not_exist = 1;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptions, _impl_.if_not_exist_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
      // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions.TableExistsOption if_exists = 2;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptions, _impl_.if_exists_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr CommandStatementIngest_TableDefinitionOptions::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        if_not_exist_{static_cast< ::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions_TableNotExistOption >(0)},
        if_exists_{static_cast< ::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions_TableExistsOption >(0)} {}

template <typename>
constexpr CommandStatementIngest_TableDefinitionOptions::CommandStatementIngest_TableDefinitionOptions(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandStatementIngest_TableDefinitionOptions::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandStatementIngest_TableDefinitionOptions(arena);
}
constexpr auto CommandStatementIngest_TableDefinitionOptions::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(CommandStatementIngest_TableDefinitionOptions), alignof(CommandStatementIngest_TableDefinitionOptions));
}
constexpr auto CommandStatementIngest_TableDefinitionOptions::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandStatementIngest_TableDefinitionOptions::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandStatementIngest_TableDefinitionOptions>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandStatementIngest_TableDefinitionOptions::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandStatementIngest_TableDefinitionOptions>(), &CommandStatementIngest_TableDefinitionOptions::ByteSizeLong,
              &CommandStatementIngest_TableDefinitionOptions::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptions, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[27],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandStatementIngest_TableDefinitionOptionsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandStatementIngest_TableDefinitionOptionsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandStatementIngest_TableDefinitionOptions_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandStatementIngest_TableDefinitionOptions::InternalGenerateClassData_(
            _default, &CommandStatementIngest_TableDefinitionOptions_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementIngest_TableDefinitionOptions>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandStatementIngest_TableDefinitionOptionsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandStatementIngest_TableDefinitionOptions _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementIngest_TableDefinitionOptions>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptionsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandStatementIngest_TableDefinitionOptionsGlobalsTypeInternal CommandStatementIngest_TableDefinitionOptions_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandStatementIngest_TableDefinitionOptions_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandStatementIngest_TableDefinitionOptions_globals_.GetClassData();
#else
  return CommandStatementIngest_TableDefinitionOptions_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
constexpr CommandStatementIngest_OptionsEntry_DoNotUse::ParseTableT_ CommandStatementIngest_OptionsEntry_DoNotUse::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandStatementIngest_OptionsEntry_DoNotUse, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::DiscardEverythingFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandStatementIngest_OptionsEntry_DoNotUse>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // string value = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest_OptionsEntry_DoNotUse, _impl_.value_)}},
      // string key = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest_OptionsEntry_DoNotUse, _impl_.key_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string key = 1;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest_OptionsEntry_DoNotUse, _impl_.key_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string value = 2;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest_OptionsEntry_DoNotUse, _impl_.value_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\75\3\5\0\0\0\0\0"
      "arrow.flight.protocol.sql.CommandStatementIngest.OptionsEntry"
      "key"
      "value"
    }},
  };
}

template <typename>
constexpr CommandStatementIngest_OptionsEntry_DoNotUse::CommandStatementIngest_OptionsEntry_DoNotUse(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : CommandStatementIngest_OptionsEntry_DoNotUse::MapEntry(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
      ) {
}
inline void* PROTOBUF_NONNULL CommandStatementIngest_OptionsEntry_DoNotUse::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandStatementIngest_OptionsEntry_DoNotUse(arena);
}
constexpr auto CommandStatementIngest_OptionsEntry_DoNotUse::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandStatementIngest_OptionsEntry_DoNotUse), alignof(CommandStatementIngest_OptionsEntry_DoNotUse));
}
constexpr auto CommandStatementIngest_OptionsEntry_DoNotUse::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandStatementIngest_OptionsEntry_DoNotUse::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandStatementIngest_OptionsEntry_DoNotUse>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandStatementIngest_OptionsEntry_DoNotUse::SharedDtor,
          static_cast<void (::google::protobuf::MessageLite::*)()>(&CommandStatementIngest_OptionsEntry_DoNotUse::ClearImpl),
              ::google::protobuf::Message::ByteSizeLongImpl, ::google::protobuf::Message::_InternalSerializeImpl
              ,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandStatementIngest_OptionsEntry_DoNotUse, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[28],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandStatementIngest_OptionsEntry_DoNotUseGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandStatementIngest_OptionsEntry_DoNotUseGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandStatementIngest_OptionsEntry_DoNotUse_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandStatementIngest_OptionsEntry_DoNotUse::InternalGenerateClassData_(
            _default, &CommandStatementIngest_OptionsEntry_DoNotUse_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementIngest_OptionsEntry_DoNotUse>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandStatementIngest_OptionsEntry_DoNotUseGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandStatementIngest_OptionsEntry_DoNotUse _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementIngest_OptionsEntry_DoNotUse>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandStatementIngest_OptionsEntry_DoNotUseGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandStatementIngest_OptionsEntry_DoNotUseGlobalsTypeInternal CommandStatementIngest_OptionsEntry_DoNotUse_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandStatementIngest_OptionsEntry_DoNotUse_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandStatementIngest_OptionsEntry_DoNotUse_globals_.GetClassData();
#else
  return CommandStatementIngest_OptionsEntry_DoNotUse_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandPreparedStatementUpdate::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandPreparedStatementUpdate>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandPreparedStatementUpdate, _impl_._has_bits_);
};

constexpr CommandPreparedStatementUpdate::ParseTableT_ CommandPreparedStatementUpdate::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandPreparedStatementUpdate, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandPreparedStatementUpdate>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bytes prepared_statement_handle = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandPreparedStatementUpdate, _impl_.prepared_statement_handle_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes prepared_statement_handle = 1;
      {PROTOBUF_FIELD_OFFSET(CommandPreparedStatementUpdate, _impl_.prepared_statement_handle_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr CommandPreparedStatementUpdate::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        prepared_statement_handle_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CommandPreparedStatementUpdate::CommandPreparedStatementUpdate(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandPreparedStatementUpdate::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandPreparedStatementUpdate(arena);
}
constexpr auto CommandPreparedStatementUpdate::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandPreparedStatementUpdate), alignof(CommandPreparedStatementUpdate));
}
constexpr auto CommandPreparedStatementUpdate::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandPreparedStatementUpdate::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandPreparedStatementUpdate>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandPreparedStatementUpdate::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandPreparedStatementUpdate>(), &CommandPreparedStatementUpdate::ByteSizeLong,
              &CommandPreparedStatementUpdate::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandPreparedStatementUpdate, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[26],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandPreparedStatementUpdateGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandPreparedStatementUpdateGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandPreparedStatementUpdate_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandPreparedStatementUpdate::InternalGenerateClassData_(
            _default, &CommandPreparedStatementUpdate_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandPreparedStatementUpdate>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandPreparedStatementUpdateGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandPreparedStatementUpdate _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandPreparedStatementUpdate>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandPreparedStatementUpdateGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandPreparedStatementUpdateGlobalsTypeInternal CommandPreparedStatementUpdate_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandPreparedStatementUpdate_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandPreparedStatementUpdate_globals_.GetClassData();
#else
  return CommandPreparedStatementUpdate_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandPreparedStatementQuery::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandPreparedStatementQuery>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandPreparedStatementQuery, _impl_._has_bits_);
};

constexpr CommandPreparedStatementQuery::ParseTableT_ CommandPreparedStatementQuery::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandPreparedStatementQuery, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandPreparedStatementQuery>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bytes prepared_statement_handle = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandPreparedStatementQuery, _impl_.prepared_statement_handle_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes prepared_statement_handle = 1;
      {PROTOBUF_FIELD_OFFSET(CommandPreparedStatementQuery, _impl_.prepared_statement_handle_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr CommandPreparedStatementQuery::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        prepared_statement_handle_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CommandPreparedStatementQuery::CommandPreparedStatementQuery(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandPreparedStatementQuery::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandPreparedStatementQuery(arena);
}
constexpr auto CommandPreparedStatementQuery::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandPreparedStatementQuery), alignof(CommandPreparedStatementQuery));
}
constexpr auto CommandPreparedStatementQuery::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandPreparedStatementQuery::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandPreparedStatementQuery>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandPreparedStatementQuery::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandPreparedStatementQuery>(), &CommandPreparedStatementQuery::ByteSizeLong,
              &CommandPreparedStatementQuery::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandPreparedStatementQuery, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[24],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandPreparedStatementQueryGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandPreparedStatementQueryGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandPreparedStatementQuery_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandPreparedStatementQuery::InternalGenerateClassData_(
            _default, &CommandPreparedStatementQuery_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandPreparedStatementQuery>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandPreparedStatementQueryGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandPreparedStatementQuery _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandPreparedStatementQuery>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandPreparedStatementQueryGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandPreparedStatementQueryGlobalsTypeInternal CommandPreparedStatementQuery_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandPreparedStatementQuery_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandPreparedStatementQuery_globals_.GetClassData();
#else
  return CommandPreparedStatementQuery_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetXdbcTypeInfo::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandGetXdbcTypeInfo>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandGetXdbcTypeInfo, _impl_._has_bits_);
};

constexpr CommandGetXdbcTypeInfo::ParseTableT_ CommandGetXdbcTypeInfo::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetXdbcTypeInfo, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetXdbcTypeInfo>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional int32 data_type = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(CommandGetXdbcTypeInfo, _impl_.data_type_), 0>(),
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetXdbcTypeInfo, _impl_.data_type_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional int32 data_type = 1;
      {PROTOBUF_FIELD_OFFSET(CommandGetXdbcTypeInfo, _impl_.data_type_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr CommandGetXdbcTypeInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_type_{0} {}

template <typename>
constexpr CommandGetXdbcTypeInfo::CommandGetXdbcTypeInfo(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandGetXdbcTypeInfo::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetXdbcTypeInfo(arena);
}
constexpr auto CommandGetXdbcTypeInfo::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(CommandGetXdbcTypeInfo), alignof(CommandGetXdbcTypeInfo));
}
constexpr auto CommandGetXdbcTypeInfo::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetXdbcTypeInfo::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandGetXdbcTypeInfo>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetXdbcTypeInfo::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandGetXdbcTypeInfo>(), &CommandGetXdbcTypeInfo::ByteSizeLong,
              &CommandGetXdbcTypeInfo::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetXdbcTypeInfo, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetXdbcTypeInfoGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetXdbcTypeInfoGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetXdbcTypeInfo_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetXdbcTypeInfo::InternalGenerateClassData_(
            _default, &CommandGetXdbcTypeInfo_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetXdbcTypeInfo>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetXdbcTypeInfoGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetXdbcTypeInfo _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetXdbcTypeInfo>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetXdbcTypeInfoGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetXdbcTypeInfoGlobalsTypeInternal CommandGetXdbcTypeInfo_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetXdbcTypeInfo_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetXdbcTypeInfo_globals_.GetClassData();
#else
  return CommandGetXdbcTypeInfo_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetTables::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandGetTables>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_._has_bits_);
};

constexpr CommandGetTables::ParseTableT_ CommandGetTables::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_._has_bits_),
      0, // no _extensions_
      5, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967264,  // skipmap
      offsetof(ParseTableT_, field_entries),
      5,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetTables>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional string catalog = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.catalog_)}},
      // optional string db_schema_filter_pattern = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 2, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.db_schema_filter_pattern_)}},
      // optional string table_name_filter_pattern = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 3, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.table_name_filter_pattern_)}},
      // repeated string table_types = 4;
      {::_pbi::TcParser::FastUR1,
       {34, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.table_types_)}},
      // bool include_schema = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(CommandGetTables, _impl_.include_schema_), 4>(),
       {40, 4, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.include_schema_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string catalog = 1;
      {PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.catalog_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string db_schema_filter_pattern = 2;
      {PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.db_schema_filter_pattern_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string table_name_filter_pattern = 3;
      {PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.table_name_filter_pattern_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // repeated string table_types = 4;
      {PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.table_types_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
      // bool include_schema = 5;
      {PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_.include_schema_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    // no aux_entries
    {{
      "\52\7\30\31\13\0\0\0"
      "arrow.flight.protocol.sql.CommandGetTables"
      "catalog"
      "db_schema_filter_pattern"
      "table_name_filter_pattern"
      "table_types"
    }},
  };
}


inline constexpr CommandGetTables::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        table_types_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::sql::CommandGetTables,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetTables, _impl_.table_types_)>()
         }
        ,
        catalog_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        db_schema_filter_pattern_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        table_name_filter_pattern_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        include_schema_{false} {}

template <typename>
constexpr CommandGetTables::CommandGetTables(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandGetTables::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetTables(arena);
}
constexpr auto CommandGetTables::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandGetTables), alignof(CommandGetTables));
}
constexpr auto CommandGetTables::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetTables::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandGetTables>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetTables::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandGetTables>(), &CommandGetTables::ByteSizeLong,
              &CommandGetTables::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetTables, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[4],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetTablesGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetTablesGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetTables_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetTables::InternalGenerateClassData_(
            _default, &CommandGetTables_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetTables>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetTablesGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetTables _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetTables>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetTablesGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetTablesGlobalsTypeInternal CommandGetTables_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetTables_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetTables_globals_.GetClassData();
#else
  return CommandGetTables_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetTableTypes::_Internal {
 public:
};

constexpr CommandGetTableTypes::ParseTableT_ CommandGetTableTypes::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetTableTypes,
                            _impl_._cached_size_),  // no hasbits
      0, // no _extensions_
      0, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967295,  // skipmap
      offsetof(ParseTableT_, field_names),  // no field_entries
      0,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetTableTypes>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, // no field_entries, or aux_entries
    {{
    }},
  };
}

template <typename>
constexpr CommandGetTableTypes::CommandGetTableTypes(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::internal::ZeroFieldsBase(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
      ) {
}
inline void* PROTOBUF_NONNULL CommandGetTableTypes::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetTableTypes(arena);
}
constexpr auto CommandGetTableTypes::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(CommandGetTableTypes), alignof(CommandGetTableTypes));
}
constexpr auto CommandGetTableTypes::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetTableTypes::MergeImpl,
          ::google::protobuf::internal::ZeroFieldsBase::GetNewImpl<CommandGetTableTypes>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetTableTypes::SharedDtor,
          ::google::protobuf::internal::ZeroFieldsBase::GetClearImpl<CommandGetTableTypes>(), &CommandGetTableTypes::ByteSizeLong,
              &CommandGetTableTypes::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetTableTypes, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[5],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetTableTypesGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetTableTypesGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetTableTypes_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetTableTypes::InternalGenerateClassData_(
            _default, &CommandGetTableTypes_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetTableTypes>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetTableTypesGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetTableTypes _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetTableTypes>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetTableTypesGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetTableTypesGlobalsTypeInternal CommandGetTableTypes_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetTableTypes_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetTableTypes_globals_.GetClassData();
#else
  return CommandGetTableTypes_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetSqlInfo::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandGetSqlInfo>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandGetSqlInfo, _impl_._has_bits_);
};

constexpr CommandGetSqlInfo::ParseTableT_ CommandGetSqlInfo::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetSqlInfo, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetSqlInfo>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated uint32 info = 1;
      {::_pbi::TcParser::FastV32P1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetSqlInfo, _impl_.info_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated uint32 info = 1;
      {PROTOBUF_FIELD_OFFSET(CommandGetSqlInfo, _impl_.info_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kPackedUInt32)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr CommandGetSqlInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        info_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::sql::CommandGetSqlInfo,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetSqlInfo, _impl_.info_)>()
         }
        ,
        _info_cached_byte_size_{0} {}

template <typename>
constexpr CommandGetSqlInfo::CommandGetSqlInfo(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandGetSqlInfo::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetSqlInfo(arena);
}
constexpr auto CommandGetSqlInfo::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandGetSqlInfo), alignof(CommandGetSqlInfo));
}
constexpr auto CommandGetSqlInfo::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetSqlInfo::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandGetSqlInfo>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetSqlInfo::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandGetSqlInfo>(), &CommandGetSqlInfo::ByteSizeLong,
              &CommandGetSqlInfo::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetSqlInfo, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetSqlInfoGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetSqlInfoGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetSqlInfo_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetSqlInfo::InternalGenerateClassData_(
            _default, &CommandGetSqlInfo_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetSqlInfo>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetSqlInfoGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetSqlInfo _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetSqlInfo>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetSqlInfoGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetSqlInfoGlobalsTypeInternal CommandGetSqlInfo_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetSqlInfo_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetSqlInfo_globals_.GetClassData();
#else
  return CommandGetSqlInfo_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetPrimaryKeys::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandGetPrimaryKeys>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeys, _impl_._has_bits_);
};

constexpr CommandGetPrimaryKeys::ParseTableT_ CommandGetPrimaryKeys::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeys, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetPrimaryKeys>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional string catalog = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeys, _impl_.catalog_)}},
      // optional string db_schema = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeys, _impl_.db_schema_)}},
      // string table = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeys, _impl_.table_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string catalog = 1;
      {PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeys, _impl_.catalog_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string db_schema = 2;
      {PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeys, _impl_.db_schema_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string table = 3;
      {PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeys, _impl_.table_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\57\7\11\5\0\0\0\0"
      "arrow.flight.protocol.sql.CommandGetPrimaryKeys"
      "catalog"
      "db_schema"
      "table"
    }},
  };
}


inline constexpr CommandGetPrimaryKeys::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        catalog_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        db_schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        table_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CommandGetPrimaryKeys::CommandGetPrimaryKeys(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandGetPrimaryKeys::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetPrimaryKeys(arena);
}
constexpr auto CommandGetPrimaryKeys::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandGetPrimaryKeys), alignof(CommandGetPrimaryKeys));
}
constexpr auto CommandGetPrimaryKeys::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetPrimaryKeys::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandGetPrimaryKeys>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetPrimaryKeys::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandGetPrimaryKeys>(), &CommandGetPrimaryKeys::ByteSizeLong,
              &CommandGetPrimaryKeys::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeys, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[6],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetPrimaryKeysGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetPrimaryKeysGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetPrimaryKeys_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetPrimaryKeys::InternalGenerateClassData_(
            _default, &CommandGetPrimaryKeys_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetPrimaryKeys>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetPrimaryKeysGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetPrimaryKeys _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetPrimaryKeys>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetPrimaryKeysGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetPrimaryKeysGlobalsTypeInternal CommandGetPrimaryKeys_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetPrimaryKeys_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetPrimaryKeys_globals_.GetClassData();
#else
  return CommandGetPrimaryKeys_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetImportedKeys::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandGetImportedKeys>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandGetImportedKeys, _impl_._has_bits_);
};

constexpr CommandGetImportedKeys::ParseTableT_ CommandGetImportedKeys::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetImportedKeys, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetImportedKeys>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional string catalog = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetImportedKeys, _impl_.catalog_)}},
      // optional string db_schema = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetImportedKeys, _impl_.db_schema_)}},
      // string table = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetImportedKeys, _impl_.table_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string catalog = 1;
      {PROTOBUF_FIELD_OFFSET(CommandGetImportedKeys, _impl_.catalog_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string db_schema = 2;
      {PROTOBUF_FIELD_OFFSET(CommandGetImportedKeys, _impl_.db_schema_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string table = 3;
      {PROTOBUF_FIELD_OFFSET(CommandGetImportedKeys, _impl_.table_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\60\7\11\5\0\0\0\0"
      "arrow.flight.protocol.sql.CommandGetImportedKeys"
      "catalog"
      "db_schema"
      "table"
    }},
  };
}


inline constexpr CommandGetImportedKeys::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        catalog_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        db_schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        table_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CommandGetImportedKeys::CommandGetImportedKeys(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandGetImportedKeys::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetImportedKeys(arena);
}
constexpr auto CommandGetImportedKeys::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandGetImportedKeys), alignof(CommandGetImportedKeys));
}
constexpr auto CommandGetImportedKeys::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetImportedKeys::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandGetImportedKeys>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetImportedKeys::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandGetImportedKeys>(), &CommandGetImportedKeys::ByteSizeLong,
              &CommandGetImportedKeys::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetImportedKeys, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[8],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetImportedKeysGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetImportedKeysGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetImportedKeys_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetImportedKeys::InternalGenerateClassData_(
            _default, &CommandGetImportedKeys_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetImportedKeys>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetImportedKeysGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetImportedKeys _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetImportedKeys>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetImportedKeysGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetImportedKeysGlobalsTypeInternal CommandGetImportedKeys_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetImportedKeys_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetImportedKeys_globals_.GetClassData();
#else
  return CommandGetImportedKeys_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetExportedKeys::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandGetExportedKeys>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandGetExportedKeys, _impl_._has_bits_);
};

constexpr CommandGetExportedKeys::ParseTableT_ CommandGetExportedKeys::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetExportedKeys, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetExportedKeys>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional string catalog = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetExportedKeys, _impl_.catalog_)}},
      // optional string db_schema = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetExportedKeys, _impl_.db_schema_)}},
      // string table = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetExportedKeys, _impl_.table_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string catalog = 1;
      {PROTOBUF_FIELD_OFFSET(CommandGetExportedKeys, _impl_.catalog_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string db_schema = 2;
      {PROTOBUF_FIELD_OFFSET(CommandGetExportedKeys, _impl_.db_schema_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string table = 3;
      {PROTOBUF_FIELD_OFFSET(CommandGetExportedKeys, _impl_.table_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\60\7\11\5\0\0\0\0"
      "arrow.flight.protocol.sql.CommandGetExportedKeys"
      "catalog"
      "db_schema"
      "table"
    }},
  };
}


inline constexpr CommandGetExportedKeys::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        catalog_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        db_schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        table_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CommandGetExportedKeys::CommandGetExportedKeys(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandGetExportedKeys::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetExportedKeys(arena);
}
constexpr auto CommandGetExportedKeys::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandGetExportedKeys), alignof(CommandGetExportedKeys));
}
constexpr auto CommandGetExportedKeys::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetExportedKeys::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandGetExportedKeys>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetExportedKeys::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandGetExportedKeys>(), &CommandGetExportedKeys::ByteSizeLong,
              &CommandGetExportedKeys::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetExportedKeys, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[7],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetExportedKeysGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetExportedKeysGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetExportedKeys_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetExportedKeys::InternalGenerateClassData_(
            _default, &CommandGetExportedKeys_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetExportedKeys>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetExportedKeysGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetExportedKeys _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetExportedKeys>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetExportedKeysGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetExportedKeysGlobalsTypeInternal CommandGetExportedKeys_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetExportedKeys_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetExportedKeys_globals_.GetClassData();
#else
  return CommandGetExportedKeys_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetDbSchemas::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandGetDbSchemas>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandGetDbSchemas, _impl_._has_bits_);
};

constexpr CommandGetDbSchemas::ParseTableT_ CommandGetDbSchemas::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetDbSchemas, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetDbSchemas>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional string db_schema_filter_pattern = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetDbSchemas, _impl_.db_schema_filter_pattern_)}},
      // optional string catalog = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetDbSchemas, _impl_.catalog_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string catalog = 1;
      {PROTOBUF_FIELD_OFFSET(CommandGetDbSchemas, _impl_.catalog_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string db_schema_filter_pattern = 2;
      {PROTOBUF_FIELD_OFFSET(CommandGetDbSchemas, _impl_.db_schema_filter_pattern_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\55\7\30\0\0\0\0\0"
      "arrow.flight.protocol.sql.CommandGetDbSchemas"
      "catalog"
      "db_schema_filter_pattern"
    }},
  };
}


inline constexpr CommandGetDbSchemas::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        catalog_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        db_schema_filter_pattern_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CommandGetDbSchemas::CommandGetDbSchemas(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandGetDbSchemas::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetDbSchemas(arena);
}
constexpr auto CommandGetDbSchemas::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandGetDbSchemas), alignof(CommandGetDbSchemas));
}
constexpr auto CommandGetDbSchemas::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetDbSchemas::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandGetDbSchemas>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetDbSchemas::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandGetDbSchemas>(), &CommandGetDbSchemas::ByteSizeLong,
              &CommandGetDbSchemas::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetDbSchemas, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[3],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetDbSchemasGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetDbSchemasGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetDbSchemas_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetDbSchemas::InternalGenerateClassData_(
            _default, &CommandGetDbSchemas_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetDbSchemas>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetDbSchemasGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetDbSchemas _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetDbSchemas>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetDbSchemasGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetDbSchemasGlobalsTypeInternal CommandGetDbSchemas_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetDbSchemas_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetDbSchemas_globals_.GetClassData();
#else
  return CommandGetDbSchemas_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetCrossReference::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandGetCrossReference>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_._has_bits_);
};

constexpr CommandGetCrossReference::ParseTableT_ CommandGetCrossReference::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_._has_bits_),
      0, // no _extensions_
      6, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967232,  // skipmap
      offsetof(ParseTableT_, field_entries),
      6,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetCrossReference>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional string pk_catalog = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.pk_catalog_)}},
      // optional string pk_db_schema = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.pk_db_schema_)}},
      // string pk_table = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.pk_table_)}},
      // optional string fk_catalog = 4;
      {::_pbi::TcParser::FastUS1,
       {34, 3, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.fk_catalog_)}},
      // optional string fk_db_schema = 5;
      {::_pbi::TcParser::FastUS1,
       {42, 4, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.fk_db_schema_)}},
      // string fk_table = 6;
      {::_pbi::TcParser::FastUS1,
       {50, 5, 0,
        PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.fk_table_)}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string pk_catalog = 1;
      {PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.pk_catalog_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string pk_db_schema = 2;
      {PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.pk_db_schema_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string pk_table = 3;
      {PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.pk_table_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string fk_catalog = 4;
      {PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.fk_catalog_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string fk_db_schema = 5;
      {PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.fk_db_schema_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string fk_table = 6;
      {PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_.fk_table_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\62\12\14\10\12\14\10\0"
      "arrow.flight.protocol.sql.CommandGetCrossReference"
      "pk_catalog"
      "pk_db_schema"
      "pk_table"
      "fk_catalog"
      "fk_db_schema"
      "fk_table"
    }},
  };
}


inline constexpr CommandGetCrossReference::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        pk_catalog_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        pk_db_schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        pk_table_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        fk_catalog_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        fk_db_schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        fk_table_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CommandGetCrossReference::CommandGetCrossReference(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandGetCrossReference::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetCrossReference(arena);
}
constexpr auto CommandGetCrossReference::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandGetCrossReference), alignof(CommandGetCrossReference));
}
constexpr auto CommandGetCrossReference::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetCrossReference::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandGetCrossReference>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetCrossReference::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandGetCrossReference>(), &CommandGetCrossReference::ByteSizeLong,
              &CommandGetCrossReference::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetCrossReference, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[9],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetCrossReferenceGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetCrossReferenceGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetCrossReference_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetCrossReference::InternalGenerateClassData_(
            _default, &CommandGetCrossReference_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetCrossReference>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetCrossReferenceGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetCrossReference _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetCrossReference>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetCrossReferenceGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetCrossReferenceGlobalsTypeInternal CommandGetCrossReference_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetCrossReference_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetCrossReference_globals_.GetClassData();
#else
  return CommandGetCrossReference_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandGetCatalogs::_Internal {
 public:
};

constexpr CommandGetCatalogs::ParseTableT_ CommandGetCatalogs::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandGetCatalogs,
                            _impl_._cached_size_),  // no hasbits
      0, // no _extensions_
      0, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967295,  // skipmap
      offsetof(ParseTableT_, field_names),  // no field_entries
      0,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandGetCatalogs>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, // no field_entries, or aux_entries
    {{
    }},
  };
}

template <typename>
constexpr CommandGetCatalogs::CommandGetCatalogs(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::internal::ZeroFieldsBase(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
      ) {
}
inline void* PROTOBUF_NONNULL CommandGetCatalogs::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandGetCatalogs(arena);
}
constexpr auto CommandGetCatalogs::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(CommandGetCatalogs), alignof(CommandGetCatalogs));
}
constexpr auto CommandGetCatalogs::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandGetCatalogs::MergeImpl,
          ::google::protobuf::internal::ZeroFieldsBase::GetNewImpl<CommandGetCatalogs>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandGetCatalogs::SharedDtor,
          ::google::protobuf::internal::ZeroFieldsBase::GetClearImpl<CommandGetCatalogs>(), &CommandGetCatalogs::ByteSizeLong,
              &CommandGetCatalogs::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandGetCatalogs, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[2],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandGetCatalogsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandGetCatalogsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandGetCatalogs_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandGetCatalogs::InternalGenerateClassData_(
            _default, &CommandGetCatalogs_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandGetCatalogs>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandGetCatalogsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandGetCatalogs _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandGetCatalogs>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandGetCatalogsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandGetCatalogsGlobalsTypeInternal CommandGetCatalogs_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandGetCatalogs_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandGetCatalogs_globals_.GetClassData();
#else
  return CommandGetCatalogs_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionEndTransactionRequest::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionEndTransactionRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionEndTransactionRequest, _impl_._has_bits_);
};

constexpr ActionEndTransactionRequest::ParseTableT_ ActionEndTransactionRequest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionEndTransactionRequest, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionEndTransactionRequest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // .arrow.flight.protocol.sql.ActionEndTransactionRequest.EndTransaction action = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(ActionEndTransactionRequest, _impl_.action_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(ActionEndTransactionRequest, _impl_.action_)}},
      // bytes transaction_id = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionEndTransactionRequest, _impl_.transaction_id_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes transaction_id = 1;
      {PROTOBUF_FIELD_OFFSET(ActionEndTransactionRequest, _impl_.transaction_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // .arrow.flight.protocol.sql.ActionEndTransactionRequest.EndTransaction action = 2;
      {PROTOBUF_FIELD_OFFSET(ActionEndTransactionRequest, _impl_.action_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr ActionEndTransactionRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        transaction_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        action_{static_cast< ::arrow::flight::protocol::sql::ActionEndTransactionRequest_EndTransaction >(0)} {}

template <typename>
constexpr ActionEndTransactionRequest::ActionEndTransactionRequest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionEndTransactionRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionEndTransactionRequest(arena);
}
constexpr auto ActionEndTransactionRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionEndTransactionRequest), alignof(ActionEndTransactionRequest));
}
constexpr auto ActionEndTransactionRequest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionEndTransactionRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionEndTransactionRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionEndTransactionRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionEndTransactionRequest>(), &ActionEndTransactionRequest::ByteSizeLong,
              &ActionEndTransactionRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionEndTransactionRequest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[19],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionEndTransactionRequestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionEndTransactionRequestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionEndTransactionRequest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionEndTransactionRequest::InternalGenerateClassData_(
            _default, &ActionEndTransactionRequest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionEndTransactionRequest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionEndTransactionRequestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionEndTransactionRequest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionEndTransactionRequest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionEndTransactionRequestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionEndTransactionRequestGlobalsTypeInternal ActionEndTransactionRequest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionEndTransactionRequest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionEndTransactionRequest_globals_.GetClassData();
#else
  return ActionEndTransactionRequest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionEndSavepointRequest::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionEndSavepointRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionEndSavepointRequest, _impl_._has_bits_);
};

constexpr ActionEndSavepointRequest::ParseTableT_ ActionEndSavepointRequest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionEndSavepointRequest, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionEndSavepointRequest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // .arrow.flight.protocol.sql.ActionEndSavepointRequest.EndSavepoint action = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(ActionEndSavepointRequest, _impl_.action_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(ActionEndSavepointRequest, _impl_.action_)}},
      // bytes savepoint_id = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionEndSavepointRequest, _impl_.savepoint_id_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes savepoint_id = 1;
      {PROTOBUF_FIELD_OFFSET(ActionEndSavepointRequest, _impl_.savepoint_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // .arrow.flight.protocol.sql.ActionEndSavepointRequest.EndSavepoint action = 2;
      {PROTOBUF_FIELD_OFFSET(ActionEndSavepointRequest, _impl_.action_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr ActionEndSavepointRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        savepoint_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        action_{static_cast< ::arrow::flight::protocol::sql::ActionEndSavepointRequest_EndSavepoint >(0)} {}

template <typename>
constexpr ActionEndSavepointRequest::ActionEndSavepointRequest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionEndSavepointRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionEndSavepointRequest(arena);
}
constexpr auto ActionEndSavepointRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionEndSavepointRequest), alignof(ActionEndSavepointRequest));
}
constexpr auto ActionEndSavepointRequest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionEndSavepointRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionEndSavepointRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionEndSavepointRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionEndSavepointRequest>(), &ActionEndSavepointRequest::ByteSizeLong,
              &ActionEndSavepointRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionEndSavepointRequest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[20],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionEndSavepointRequestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionEndSavepointRequestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionEndSavepointRequest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionEndSavepointRequest::InternalGenerateClassData_(
            _default, &ActionEndSavepointRequest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionEndSavepointRequest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionEndSavepointRequestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionEndSavepointRequest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionEndSavepointRequest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionEndSavepointRequestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionEndSavepointRequestGlobalsTypeInternal ActionEndSavepointRequest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionEndSavepointRequest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionEndSavepointRequest_globals_.GetClassData();
#else
  return ActionEndSavepointRequest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionCreatePreparedStatementResult::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionCreatePreparedStatementResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResult, _impl_._has_bits_);
};

constexpr ActionCreatePreparedStatementResult::ParseTableT_ ActionCreatePreparedStatementResult::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResult, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionCreatePreparedStatementResult>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // bytes prepared_statement_handle = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResult, _impl_.prepared_statement_handle_)}},
      // bytes dataset_schema = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResult, _impl_.dataset_schema_)}},
      // bytes parameter_schema = 3;
      {::_pbi::TcParser::FastBS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResult, _impl_.parameter_schema_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes prepared_statement_handle = 1;
      {PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResult, _impl_.prepared_statement_handle_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // bytes dataset_schema = 2;
      {PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResult, _impl_.dataset_schema_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // bytes parameter_schema = 3;
      {PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResult, _impl_.parameter_schema_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr ActionCreatePreparedStatementResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        prepared_statement_handle_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        dataset_schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        parameter_schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr ActionCreatePreparedStatementResult::ActionCreatePreparedStatementResult(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionCreatePreparedStatementResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionCreatePreparedStatementResult(arena);
}
constexpr auto ActionCreatePreparedStatementResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionCreatePreparedStatementResult), alignof(ActionCreatePreparedStatementResult));
}
constexpr auto ActionCreatePreparedStatementResult::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionCreatePreparedStatementResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionCreatePreparedStatementResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionCreatePreparedStatementResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionCreatePreparedStatementResult>(), &ActionCreatePreparedStatementResult::ByteSizeLong,
              &ActionCreatePreparedStatementResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResult, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[13],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionCreatePreparedStatementResultGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionCreatePreparedStatementResultGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionCreatePreparedStatementResult_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionCreatePreparedStatementResult::InternalGenerateClassData_(
            _default, &ActionCreatePreparedStatementResult_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionCreatePreparedStatementResult>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionCreatePreparedStatementResultGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionCreatePreparedStatementResult _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionCreatePreparedStatementResult>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementResultGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionCreatePreparedStatementResultGlobalsTypeInternal ActionCreatePreparedStatementResult_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionCreatePreparedStatementResult_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionCreatePreparedStatementResult_globals_.GetClassData();
#else
  return ActionCreatePreparedStatementResult_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionCreatePreparedStatementRequest::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionCreatePreparedStatementRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementRequest, _impl_._has_bits_);
};

constexpr ActionCreatePreparedStatementRequest::ParseTableT_ ActionCreatePreparedStatementRequest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementRequest, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionCreatePreparedStatementRequest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional bytes transaction_id = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementRequest, _impl_.transaction_id_)}},
      // string query = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementRequest, _impl_.query_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string query = 1;
      {PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementRequest, _impl_.query_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional bytes transaction_id = 2;
      {PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementRequest, _impl_.transaction_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\76\5\0\0\0\0\0\0"
      "arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest"
      "query"
    }},
  };
}


inline constexpr ActionCreatePreparedStatementRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        query_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        transaction_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr ActionCreatePreparedStatementRequest::ActionCreatePreparedStatementRequest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionCreatePreparedStatementRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionCreatePreparedStatementRequest(arena);
}
constexpr auto ActionCreatePreparedStatementRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionCreatePreparedStatementRequest), alignof(ActionCreatePreparedStatementRequest));
}
constexpr auto ActionCreatePreparedStatementRequest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionCreatePreparedStatementRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionCreatePreparedStatementRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionCreatePreparedStatementRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionCreatePreparedStatementRequest>(), &ActionCreatePreparedStatementRequest::ByteSizeLong,
              &ActionCreatePreparedStatementRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementRequest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[10],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionCreatePreparedStatementRequestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionCreatePreparedStatementRequestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionCreatePreparedStatementRequest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionCreatePreparedStatementRequest::InternalGenerateClassData_(
            _default, &ActionCreatePreparedStatementRequest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionCreatePreparedStatementRequest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionCreatePreparedStatementRequestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionCreatePreparedStatementRequest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionCreatePreparedStatementRequest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionCreatePreparedStatementRequestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionCreatePreparedStatementRequestGlobalsTypeInternal ActionCreatePreparedStatementRequest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionCreatePreparedStatementRequest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionCreatePreparedStatementRequest_globals_.GetClassData();
#else
  return ActionCreatePreparedStatementRequest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionClosePreparedStatementRequest::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionClosePreparedStatementRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionClosePreparedStatementRequest, _impl_._has_bits_);
};

constexpr ActionClosePreparedStatementRequest::ParseTableT_ ActionClosePreparedStatementRequest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionClosePreparedStatementRequest, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionClosePreparedStatementRequest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bytes prepared_statement_handle = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionClosePreparedStatementRequest, _impl_.prepared_statement_handle_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes prepared_statement_handle = 1;
      {PROTOBUF_FIELD_OFFSET(ActionClosePreparedStatementRequest, _impl_.prepared_statement_handle_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr ActionClosePreparedStatementRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        prepared_statement_handle_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr ActionClosePreparedStatementRequest::ActionClosePreparedStatementRequest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionClosePreparedStatementRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionClosePreparedStatementRequest(arena);
}
constexpr auto ActionClosePreparedStatementRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionClosePreparedStatementRequest), alignof(ActionClosePreparedStatementRequest));
}
constexpr auto ActionClosePreparedStatementRequest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionClosePreparedStatementRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionClosePreparedStatementRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionClosePreparedStatementRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionClosePreparedStatementRequest>(), &ActionClosePreparedStatementRequest::ByteSizeLong,
              &ActionClosePreparedStatementRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionClosePreparedStatementRequest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[14],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionClosePreparedStatementRequestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionClosePreparedStatementRequestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionClosePreparedStatementRequest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionClosePreparedStatementRequest::InternalGenerateClassData_(
            _default, &ActionClosePreparedStatementRequest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionClosePreparedStatementRequest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionClosePreparedStatementRequestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionClosePreparedStatementRequest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionClosePreparedStatementRequest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionClosePreparedStatementRequestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionClosePreparedStatementRequestGlobalsTypeInternal ActionClosePreparedStatementRequest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionClosePreparedStatementRequest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionClosePreparedStatementRequest_globals_.GetClassData();
#else
  return ActionClosePreparedStatementRequest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionCancelQueryResult::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionCancelQueryResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionCancelQueryResult, _impl_._has_bits_);
};

constexpr ActionCancelQueryResult::ParseTableT_ ActionCancelQueryResult::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionCancelQueryResult, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionCancelQueryResult>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // .arrow.flight.protocol.sql.ActionCancelQueryResult.CancelResult result = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(ActionCancelQueryResult, _impl_.result_), 0>(),
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionCancelQueryResult, _impl_.result_)}},
    }}, {{
      65535, 65535
    }}, {{
      // .arrow.flight.protocol.sql.ActionCancelQueryResult.CancelResult result = 1;
      {PROTOBUF_FIELD_OFFSET(ActionCancelQueryResult, _impl_.result_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr ActionCancelQueryResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        result_{static_cast< ::arrow::flight::protocol::sql::ActionCancelQueryResult_CancelResult >(0)} {}

template <typename>
constexpr ActionCancelQueryResult::ActionCancelQueryResult(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionCancelQueryResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionCancelQueryResult(arena);
}
constexpr auto ActionCancelQueryResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(ActionCancelQueryResult), alignof(ActionCancelQueryResult));
}
constexpr auto ActionCancelQueryResult::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionCancelQueryResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionCancelQueryResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionCancelQueryResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionCancelQueryResult>(), &ActionCancelQueryResult::ByteSizeLong,
              &ActionCancelQueryResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionCancelQueryResult, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[33],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionCancelQueryResultGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionCancelQueryResultGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionCancelQueryResult_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionCancelQueryResult::InternalGenerateClassData_(
            _default, &ActionCancelQueryResult_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionCancelQueryResult>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionCancelQueryResultGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionCancelQueryResult _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionCancelQueryResult>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionCancelQueryResultGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionCancelQueryResultGlobalsTypeInternal ActionCancelQueryResult_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionCancelQueryResult_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionCancelQueryResult_globals_.GetClassData();
#else
  return ActionCancelQueryResult_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionCancelQueryRequest::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionCancelQueryRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionCancelQueryRequest, _impl_._has_bits_);
};

constexpr ActionCancelQueryRequest::ParseTableT_ ActionCancelQueryRequest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionCancelQueryRequest, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionCancelQueryRequest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bytes info = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionCancelQueryRequest, _impl_.info_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes info = 1;
      {PROTOBUF_FIELD_OFFSET(ActionCancelQueryRequest, _impl_.info_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr ActionCancelQueryRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        info_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr ActionCancelQueryRequest::ActionCancelQueryRequest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionCancelQueryRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionCancelQueryRequest(arena);
}
constexpr auto ActionCancelQueryRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionCancelQueryRequest), alignof(ActionCancelQueryRequest));
}
constexpr auto ActionCancelQueryRequest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionCancelQueryRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionCancelQueryRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionCancelQueryRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionCancelQueryRequest>(), &ActionCancelQueryRequest::ByteSizeLong,
              &ActionCancelQueryRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionCancelQueryRequest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[32],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionCancelQueryRequestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionCancelQueryRequestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionCancelQueryRequest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionCancelQueryRequest::InternalGenerateClassData_(
            _default, &ActionCancelQueryRequest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionCancelQueryRequest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionCancelQueryRequestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionCancelQueryRequest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionCancelQueryRequest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionCancelQueryRequestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionCancelQueryRequestGlobalsTypeInternal ActionCancelQueryRequest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionCancelQueryRequest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionCancelQueryRequest_globals_.GetClassData();
#else
  return ActionCancelQueryRequest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionBeginTransactionResult::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionBeginTransactionResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionBeginTransactionResult, _impl_._has_bits_);
};

constexpr ActionBeginTransactionResult::ParseTableT_ ActionBeginTransactionResult::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionBeginTransactionResult, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionBeginTransactionResult>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bytes transaction_id = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionBeginTransactionResult, _impl_.transaction_id_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes transaction_id = 1;
      {PROTOBUF_FIELD_OFFSET(ActionBeginTransactionResult, _impl_.transaction_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr ActionBeginTransactionResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        transaction_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr ActionBeginTransactionResult::ActionBeginTransactionResult(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionBeginTransactionResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionBeginTransactionResult(arena);
}
constexpr auto ActionBeginTransactionResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionBeginTransactionResult), alignof(ActionBeginTransactionResult));
}
constexpr auto ActionBeginTransactionResult::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionBeginTransactionResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionBeginTransactionResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionBeginTransactionResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionBeginTransactionResult>(), &ActionBeginTransactionResult::ByteSizeLong,
              &ActionBeginTransactionResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionBeginTransactionResult, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[17],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionBeginTransactionResultGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionBeginTransactionResultGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionBeginTransactionResult_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionBeginTransactionResult::InternalGenerateClassData_(
            _default, &ActionBeginTransactionResult_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionBeginTransactionResult>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionBeginTransactionResultGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionBeginTransactionResult _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionBeginTransactionResult>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionBeginTransactionResultGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionBeginTransactionResultGlobalsTypeInternal ActionBeginTransactionResult_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionBeginTransactionResult_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionBeginTransactionResult_globals_.GetClassData();
#else
  return ActionBeginTransactionResult_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionBeginTransactionRequest::_Internal {
 public:
};

constexpr ActionBeginTransactionRequest::ParseTableT_ ActionBeginTransactionRequest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionBeginTransactionRequest,
                            _impl_._cached_size_),  // no hasbits
      0, // no _extensions_
      0, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967295,  // skipmap
      offsetof(ParseTableT_, field_names),  // no field_entries
      0,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionBeginTransactionRequest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, // no field_entries, or aux_entries
    {{
    }},
  };
}

template <typename>
constexpr ActionBeginTransactionRequest::ActionBeginTransactionRequest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::internal::ZeroFieldsBase(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
      ) {
}
inline void* PROTOBUF_NONNULL ActionBeginTransactionRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionBeginTransactionRequest(arena);
}
constexpr auto ActionBeginTransactionRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(ActionBeginTransactionRequest), alignof(ActionBeginTransactionRequest));
}
constexpr auto ActionBeginTransactionRequest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionBeginTransactionRequest::MergeImpl,
          ::google::protobuf::internal::ZeroFieldsBase::GetNewImpl<ActionBeginTransactionRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionBeginTransactionRequest::SharedDtor,
          ::google::protobuf::internal::ZeroFieldsBase::GetClearImpl<ActionBeginTransactionRequest>(), &ActionBeginTransactionRequest::ByteSizeLong,
              &ActionBeginTransactionRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionBeginTransactionRequest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[15],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionBeginTransactionRequestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionBeginTransactionRequestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionBeginTransactionRequest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionBeginTransactionRequest::InternalGenerateClassData_(
            _default, &ActionBeginTransactionRequest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionBeginTransactionRequest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionBeginTransactionRequestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionBeginTransactionRequest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionBeginTransactionRequest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionBeginTransactionRequestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionBeginTransactionRequestGlobalsTypeInternal ActionBeginTransactionRequest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionBeginTransactionRequest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionBeginTransactionRequest_globals_.GetClassData();
#else
  return ActionBeginTransactionRequest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionBeginSavepointResult::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionBeginSavepointResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionBeginSavepointResult, _impl_._has_bits_);
};

constexpr ActionBeginSavepointResult::ParseTableT_ ActionBeginSavepointResult::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionBeginSavepointResult, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionBeginSavepointResult>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // bytes savepoint_id = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionBeginSavepointResult, _impl_.savepoint_id_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes savepoint_id = 1;
      {PROTOBUF_FIELD_OFFSET(ActionBeginSavepointResult, _impl_.savepoint_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr ActionBeginSavepointResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        savepoint_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr ActionBeginSavepointResult::ActionBeginSavepointResult(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionBeginSavepointResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionBeginSavepointResult(arena);
}
constexpr auto ActionBeginSavepointResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionBeginSavepointResult), alignof(ActionBeginSavepointResult));
}
constexpr auto ActionBeginSavepointResult::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionBeginSavepointResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionBeginSavepointResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionBeginSavepointResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionBeginSavepointResult>(), &ActionBeginSavepointResult::ByteSizeLong,
              &ActionBeginSavepointResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionBeginSavepointResult, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[18],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionBeginSavepointResultGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionBeginSavepointResultGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionBeginSavepointResult_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionBeginSavepointResult::InternalGenerateClassData_(
            _default, &ActionBeginSavepointResult_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionBeginSavepointResult>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionBeginSavepointResultGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionBeginSavepointResult _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionBeginSavepointResult>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionBeginSavepointResultGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionBeginSavepointResultGlobalsTypeInternal ActionBeginSavepointResult_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionBeginSavepointResult_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionBeginSavepointResult_globals_.GetClassData();
#else
  return ActionBeginSavepointResult_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionBeginSavepointRequest::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionBeginSavepointRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionBeginSavepointRequest, _impl_._has_bits_);
};

constexpr ActionBeginSavepointRequest::ParseTableT_ ActionBeginSavepointRequest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionBeginSavepointRequest, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionBeginSavepointRequest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // string name = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(ActionBeginSavepointRequest, _impl_.name_)}},
      // bytes transaction_id = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionBeginSavepointRequest, _impl_.transaction_id_)}},
    }}, {{
      65535, 65535
    }}, {{
      // bytes transaction_id = 1;
      {PROTOBUF_FIELD_OFFSET(ActionBeginSavepointRequest, _impl_.transaction_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // string name = 2;
      {PROTOBUF_FIELD_OFFSET(ActionBeginSavepointRequest, _impl_.name_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\65\0\4\0\0\0\0\0"
      "arrow.flight.protocol.sql.ActionBeginSavepointRequest"
      "name"
    }},
  };
}


inline constexpr ActionBeginSavepointRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        transaction_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr ActionBeginSavepointRequest::ActionBeginSavepointRequest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionBeginSavepointRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionBeginSavepointRequest(arena);
}
constexpr auto ActionBeginSavepointRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionBeginSavepointRequest), alignof(ActionBeginSavepointRequest));
}
constexpr auto ActionBeginSavepointRequest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionBeginSavepointRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionBeginSavepointRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionBeginSavepointRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionBeginSavepointRequest>(), &ActionBeginSavepointRequest::ByteSizeLong,
              &ActionBeginSavepointRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionBeginSavepointRequest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[16],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionBeginSavepointRequestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionBeginSavepointRequestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionBeginSavepointRequest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionBeginSavepointRequest::InternalGenerateClassData_(
            _default, &ActionBeginSavepointRequest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionBeginSavepointRequest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionBeginSavepointRequestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionBeginSavepointRequest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionBeginSavepointRequest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionBeginSavepointRequestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionBeginSavepointRequestGlobalsTypeInternal ActionBeginSavepointRequest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionBeginSavepointRequest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionBeginSavepointRequest_globals_.GetClassData();
#else
  return ActionBeginSavepointRequest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandStatementSubstraitPlan::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandStatementSubstraitPlan>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandStatementSubstraitPlan, _impl_._has_bits_);
};

constexpr CommandStatementSubstraitPlan::ParseTableT_ CommandStatementSubstraitPlan::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandStatementSubstraitPlan, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandStatementSubstraitPlan>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional bytes transaction_id = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementSubstraitPlan, _impl_.transaction_id_)}},
      // .arrow.flight.protocol.sql.SubstraitPlan plan = 1;
      {::_pbi::TcParser::FastMtS1,
       {10, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementSubstraitPlan, _impl_.plan_)}},
    }}, {{
      65535, 65535
    }}, {{
      // .arrow.flight.protocol.sql.SubstraitPlan plan = 1;
      {PROTOBUF_FIELD_OFFSET(CommandStatementSubstraitPlan, _impl_.plan_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional bytes transaction_id = 2;
      {PROTOBUF_FIELD_OFFSET(CommandStatementSubstraitPlan, _impl_.transaction_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::SubstraitPlan>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::arrow::flight::protocol::sql::SubstraitPlan_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr CommandStatementSubstraitPlan::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        transaction_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        plan_{nullptr} {}

template <typename>
constexpr CommandStatementSubstraitPlan::CommandStatementSubstraitPlan(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandStatementSubstraitPlan::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandStatementSubstraitPlan(arena);
}
constexpr auto CommandStatementSubstraitPlan::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandStatementSubstraitPlan), alignof(CommandStatementSubstraitPlan));
}
constexpr auto CommandStatementSubstraitPlan::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandStatementSubstraitPlan::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandStatementSubstraitPlan>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandStatementSubstraitPlan::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandStatementSubstraitPlan>(), &CommandStatementSubstraitPlan::ByteSizeLong,
              &CommandStatementSubstraitPlan::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandStatementSubstraitPlan, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[22],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandStatementSubstraitPlanGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandStatementSubstraitPlanGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandStatementSubstraitPlan_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandStatementSubstraitPlan::InternalGenerateClassData_(
            _default, &CommandStatementSubstraitPlan_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementSubstraitPlan>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandStatementSubstraitPlanGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandStatementSubstraitPlan _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementSubstraitPlan>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandStatementSubstraitPlanGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandStatementSubstraitPlanGlobalsTypeInternal CommandStatementSubstraitPlan_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandStatementSubstraitPlan_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandStatementSubstraitPlan_globals_.GetClassData();
#else
  return CommandStatementSubstraitPlan_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CommandStatementIngest::_Internal {
 public:
  using HasBits = decltype(::std::declval<CommandStatementIngest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_._has_bits_);
};

constexpr CommandStatementIngest::ParseTableT_ CommandStatementIngest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_._has_bits_),
      0, // no _extensions_
      1000, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967232,  // skipmap
      offsetof(ParseTableT_, field_entries),
      7,  // num_field_entries
      2,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandStatementIngest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions table_definition_options = 1;
      {::_pbi::TcParser::FastMtS1,
       {10, 4, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.table_definition_options_)}},
      // string table = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.table_)}},
      // optional string schema = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 1, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.schema_)}},
      // optional string catalog = 4;
      {::_pbi::TcParser::FastUS1,
       {34, 2, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.catalog_)}},
      // bool temporary = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(CommandStatementIngest, _impl_.temporary_), 5>(),
       {40, 5, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.temporary_)}},
      // optional bytes transaction_id = 6;
      {::_pbi::TcParser::FastBS1,
       {50, 3, 0,
        PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.transaction_id_)}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      1000, 0, 1,
      65534, 6,
      65535, 65535
    }}, {{
      // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions table_definition_options = 1;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.table_definition_options_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // string table = 2;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.table_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string schema = 3;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.schema_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // optional string catalog = 4;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.catalog_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool temporary = 5;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.temporary_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // optional bytes transaction_id = 6;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.transaction_id_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // map<string, string> options = 1000;
      {PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.options_), _Internal::kHasBitsOffset + 6, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMap)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions_globals_},
        #endif
        {::_pbi::TcParser::GetMapAuxInfo(
            1, 0, 9, 9, 0)},
    }},
    {{
      "\60\0\5\6\7\0\0\7"
      "arrow.flight.protocol.sql.CommandStatementIngest"
      "table"
      "schema"
      "catalog"
      "options"
    }},
  };
}


inline constexpr CommandStatementIngest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        table_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        schema_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        catalog_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        transaction_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        table_definition_options_{nullptr},
        temporary_{false},
        options_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::sql::CommandStatementIngest,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.options_)>()
         }
     {}

template <typename>
constexpr CommandStatementIngest::CommandStatementIngest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CommandStatementIngest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CommandStatementIngest(arena);
}
constexpr auto CommandStatementIngest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CommandStatementIngest), alignof(CommandStatementIngest));
}
constexpr auto CommandStatementIngest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CommandStatementIngest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CommandStatementIngest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CommandStatementIngest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CommandStatementIngest>(), &CommandStatementIngest::ByteSizeLong,
              &CommandStatementIngest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[29],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CommandStatementIngestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CommandStatementIngestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CommandStatementIngest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CommandStatementIngest::InternalGenerateClassData_(
            _default, &CommandStatementIngest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementIngest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CommandStatementIngestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CommandStatementIngest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CommandStatementIngest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CommandStatementIngestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CommandStatementIngestGlobalsTypeInternal CommandStatementIngest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CommandStatementIngest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CommandStatementIngest_globals_.GetClassData();
#else
  return CommandStatementIngest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ActionCreatePreparedSubstraitPlanRequest::_Internal {
 public:
  using HasBits = decltype(::std::declval<ActionCreatePreparedSubstraitPlanRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ActionCreatePreparedSubstraitPlanRequest, _impl_._has_bits_);
};

constexpr ActionCreatePreparedSubstraitPlanRequest::ParseTableT_ ActionCreatePreparedSubstraitPlanRequest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ActionCreatePreparedSubstraitPlanRequest, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::ActionCreatePreparedSubstraitPlanRequest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional bytes transaction_id = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(ActionCreatePreparedSubstraitPlanRequest, _impl_.transaction_id_)}},
      // .arrow.flight.protocol.sql.SubstraitPlan plan = 1;
      {::_pbi::TcParser::FastMtS1,
       {10, 1, 0,
        PROTOBUF_FIELD_OFFSET(ActionCreatePreparedSubstraitPlanRequest, _impl_.plan_)}},
    }}, {{
      65535, 65535
    }}, {{
      // .arrow.flight.protocol.sql.SubstraitPlan plan = 1;
      {PROTOBUF_FIELD_OFFSET(ActionCreatePreparedSubstraitPlanRequest, _impl_.plan_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional bytes transaction_id = 2;
      {PROTOBUF_FIELD_OFFSET(ActionCreatePreparedSubstraitPlanRequest, _impl_.transaction_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::arrow::flight::protocol::sql::SubstraitPlan>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::arrow::flight::protocol::sql::SubstraitPlan_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr ActionCreatePreparedSubstraitPlanRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        transaction_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        plan_{nullptr} {}

template <typename>
constexpr ActionCreatePreparedSubstraitPlanRequest::ActionCreatePreparedSubstraitPlanRequest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ActionCreatePreparedSubstraitPlanRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ActionCreatePreparedSubstraitPlanRequest(arena);
}
constexpr auto ActionCreatePreparedSubstraitPlanRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ActionCreatePreparedSubstraitPlanRequest), alignof(ActionCreatePreparedSubstraitPlanRequest));
}
constexpr auto ActionCreatePreparedSubstraitPlanRequest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ActionCreatePreparedSubstraitPlanRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ActionCreatePreparedSubstraitPlanRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ActionCreatePreparedSubstraitPlanRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ActionCreatePreparedSubstraitPlanRequest>(), &ActionCreatePreparedSubstraitPlanRequest::ByteSizeLong,
              &ActionCreatePreparedSubstraitPlanRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ActionCreatePreparedSubstraitPlanRequest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[12],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_FlightSql_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ActionCreatePreparedSubstraitPlanRequestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ActionCreatePreparedSubstraitPlanRequestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ActionCreatePreparedSubstraitPlanRequest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ActionCreatePreparedSubstraitPlanRequest::InternalGenerateClassData_(
            _default, &ActionCreatePreparedSubstraitPlanRequest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ActionCreatePreparedSubstraitPlanRequest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ActionCreatePreparedSubstraitPlanRequestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ActionCreatePreparedSubstraitPlanRequest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ActionCreatePreparedSubstraitPlanRequest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ActionCreatePreparedSubstraitPlanRequestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ActionCreatePreparedSubstraitPlanRequestGlobalsTypeInternal ActionCreatePreparedSubstraitPlanRequest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ActionCreatePreparedSubstraitPlanRequest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ActionCreatePreparedSubstraitPlanRequest_globals_.GetClassData();
#else
  return ActionCreatePreparedSubstraitPlanRequest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace sql
}  // namespace protocol
}  // namespace flight
}  // namespace arrow
static const ::_pb::EnumDescriptor* PROTOBUF_NONNULL
    file_level_enum_descriptors_FlightSql_2eproto[27];
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_FlightSql_2eproto = nullptr;
const ::uint32_t
    TableStruct_FlightSql_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetSqlInfo, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetSqlInfo, _impl_.info_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetXdbcTypeInfo, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetXdbcTypeInfo, _impl_.data_type_),
        0,
        0x000, // bitmap
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetDbSchemas, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetDbSchemas, _impl_.catalog_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetDbSchemas, _impl_.db_schema_filter_pattern_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetTables, _impl_._has_bits_),
        8, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetTables, _impl_.catalog_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetTables, _impl_.db_schema_filter_pattern_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetTables, _impl_.table_name_filter_pattern_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetTables, _impl_.table_types_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetTables, _impl_.include_schema_),
        1,
        2,
        3,
        0,
        4,
        0x000, // bitmap
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetPrimaryKeys, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetPrimaryKeys, _impl_.catalog_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetPrimaryKeys, _impl_.db_schema_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetPrimaryKeys, _impl_.table_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetExportedKeys, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetExportedKeys, _impl_.catalog_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetExportedKeys, _impl_.db_schema_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetExportedKeys, _impl_.table_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetImportedKeys, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetImportedKeys, _impl_.catalog_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetImportedKeys, _impl_.db_schema_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetImportedKeys, _impl_.table_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetCrossReference, _impl_._has_bits_),
        9, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetCrossReference, _impl_.pk_catalog_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetCrossReference, _impl_.pk_db_schema_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetCrossReference, _impl_.pk_table_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetCrossReference, _impl_.fk_catalog_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetCrossReference, _impl_.fk_db_schema_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetCrossReference, _impl_.fk_table_),
        0,
        1,
        2,
        3,
        4,
        5,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedStatementRequest, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedStatementRequest, _impl_.query_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedStatementRequest, _impl_.transaction_id_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::SubstraitPlan, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::SubstraitPlan, _impl_.plan_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::SubstraitPlan, _impl_.version_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedSubstraitPlanRequest, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedSubstraitPlanRequest, _impl_.plan_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedSubstraitPlanRequest, _impl_.transaction_id_),
        1,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedStatementResult, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedStatementResult, _impl_.prepared_statement_handle_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedStatementResult, _impl_.dataset_schema_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCreatePreparedStatementResult, _impl_.parameter_schema_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionClosePreparedStatementRequest, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionClosePreparedStatementRequest, _impl_.prepared_statement_handle_),
        0,
        0x000, // bitmap
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionBeginSavepointRequest, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionBeginSavepointRequest, _impl_.transaction_id_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionBeginSavepointRequest, _impl_.name_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionBeginTransactionResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionBeginTransactionResult, _impl_.transaction_id_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionBeginSavepointResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionBeginSavepointResult, _impl_.savepoint_id_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionEndTransactionRequest, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionEndTransactionRequest, _impl_.transaction_id_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionEndTransactionRequest, _impl_.action_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionEndSavepointRequest, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionEndSavepointRequest, _impl_.savepoint_id_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionEndSavepointRequest, _impl_.action_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementQuery, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementQuery, _impl_.query_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementQuery, _impl_.transaction_id_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementSubstraitPlan, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementSubstraitPlan, _impl_.plan_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementSubstraitPlan, _impl_.transaction_id_),
        1,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::TicketStatementQuery, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::TicketStatementQuery, _impl_.statement_handle_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandPreparedStatementQuery, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandPreparedStatementQuery, _impl_.prepared_statement_handle_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementUpdate, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementUpdate, _impl_.query_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementUpdate, _impl_.transaction_id_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandPreparedStatementUpdate, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandPreparedStatementUpdate, _impl_.prepared_statement_handle_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions, _impl_.if_not_exist_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions, _impl_.if_exists_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest_OptionsEntry_DoNotUse, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest_OptionsEntry_DoNotUse, _impl_.key_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest_OptionsEntry_DoNotUse, _impl_.value_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_._has_bits_),
        10, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.table_definition_options_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.table_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.schema_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.catalog_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.temporary_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.transaction_id_),
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.options_),
        4,
        0,
        1,
        2,
        5,
        3,
        6,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::DoPutUpdateResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::DoPutUpdateResult, _impl_.record_count_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::DoPutPreparedStatementResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::DoPutPreparedStatementResult, _impl_.prepared_statement_handle_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCancelQueryRequest, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCancelQueryRequest, _impl_.info_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCancelQueryResult, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::ActionCancelQueryResult, _impl_.result_),
        0,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::arrow::flight::protocol::sql::CommandGetSqlInfo)},
        {5, sizeof(::arrow::flight::protocol::sql::CommandGetXdbcTypeInfo)},
        {10, sizeof(::arrow::flight::protocol::sql::CommandGetCatalogs)},
        {11, sizeof(::arrow::flight::protocol::sql::CommandGetDbSchemas)},
        {18, sizeof(::arrow::flight::protocol::sql::CommandGetTables)},
        {31, sizeof(::arrow::flight::protocol::sql::CommandGetTableTypes)},
        {32, sizeof(::arrow::flight::protocol::sql::CommandGetPrimaryKeys)},
        {41, sizeof(::arrow::flight::protocol::sql::CommandGetExportedKeys)},
        {50, sizeof(::arrow::flight::protocol::sql::CommandGetImportedKeys)},
        {59, sizeof(::arrow::flight::protocol::sql::CommandGetCrossReference)},
        {74, sizeof(::arrow::flight::protocol::sql::ActionCreatePreparedStatementRequest)},
        {81, sizeof(::arrow::flight::protocol::sql::SubstraitPlan)},
        {88, sizeof(::arrow::flight::protocol::sql::ActionCreatePreparedSubstraitPlanRequest)},
        {95, sizeof(::arrow::flight::protocol::sql::ActionCreatePreparedStatementResult)},
        {104, sizeof(::arrow::flight::protocol::sql::ActionClosePreparedStatementRequest)},
        {109, sizeof(::arrow::flight::protocol::sql::ActionBeginTransactionRequest)},
        {110, sizeof(::arrow::flight::protocol::sql::ActionBeginSavepointRequest)},
        {117, sizeof(::arrow::flight::protocol::sql::ActionBeginTransactionResult)},
        {122, sizeof(::arrow::flight::protocol::sql::ActionBeginSavepointResult)},
        {127, sizeof(::arrow::flight::protocol::sql::ActionEndTransactionRequest)},
        {134, sizeof(::arrow::flight::protocol::sql::ActionEndSavepointRequest)},
        {141, sizeof(::arrow::flight::protocol::sql::CommandStatementQuery)},
        {148, sizeof(::arrow::flight::protocol::sql::CommandStatementSubstraitPlan)},
        {155, sizeof(::arrow::flight::protocol::sql::TicketStatementQuery)},
        {160, sizeof(::arrow::flight::protocol::sql::CommandPreparedStatementQuery)},
        {165, sizeof(::arrow::flight::protocol::sql::CommandStatementUpdate)},
        {172, sizeof(::arrow::flight::protocol::sql::CommandPreparedStatementUpdate)},
        {177, sizeof(::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions)},
        {184, sizeof(::arrow::flight::protocol::sql::CommandStatementIngest_OptionsEntry_DoNotUse)},
        {191, sizeof(::arrow::flight::protocol::sql::CommandStatementIngest)},
        {208, sizeof(::arrow::flight::protocol::sql::DoPutUpdateResult)},
        {213, sizeof(::arrow::flight::protocol::sql::DoPutPreparedStatementResult)},
        {218, sizeof(::arrow::flight::protocol::sql::ActionCancelQueryRequest)},
        {223, sizeof(::arrow::flight::protocol::sql::ActionCancelQueryResult)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::arrow::flight::protocol::sql::CommandGetSqlInfo_globals_,
        &::arrow::flight::protocol::sql::CommandGetXdbcTypeInfo_globals_,
        &::arrow::flight::protocol::sql::CommandGetCatalogs_globals_,
        &::arrow::flight::protocol::sql::CommandGetDbSchemas_globals_,
        &::arrow::flight::protocol::sql::CommandGetTables_globals_,
        &::arrow::flight::protocol::sql::CommandGetTableTypes_globals_,
        &::arrow::flight::protocol::sql::CommandGetPrimaryKeys_globals_,
        &::arrow::flight::protocol::sql::CommandGetExportedKeys_globals_,
        &::arrow::flight::protocol::sql::CommandGetImportedKeys_globals_,
        &::arrow::flight::protocol::sql::CommandGetCrossReference_globals_,
        &::arrow::flight::protocol::sql::ActionCreatePreparedStatementRequest_globals_,
        &::arrow::flight::protocol::sql::SubstraitPlan_globals_,
        &::arrow::flight::protocol::sql::ActionCreatePreparedSubstraitPlanRequest_globals_,
        &::arrow::flight::protocol::sql::ActionCreatePreparedStatementResult_globals_,
        &::arrow::flight::protocol::sql::ActionClosePreparedStatementRequest_globals_,
        &::arrow::flight::protocol::sql::ActionBeginTransactionRequest_globals_,
        &::arrow::flight::protocol::sql::ActionBeginSavepointRequest_globals_,
        &::arrow::flight::protocol::sql::ActionBeginTransactionResult_globals_,
        &::arrow::flight::protocol::sql::ActionBeginSavepointResult_globals_,
        &::arrow::flight::protocol::sql::ActionEndTransactionRequest_globals_,
        &::arrow::flight::protocol::sql::ActionEndSavepointRequest_globals_,
        &::arrow::flight::protocol::sql::CommandStatementQuery_globals_,
        &::arrow::flight::protocol::sql::CommandStatementSubstraitPlan_globals_,
        &::arrow::flight::protocol::sql::TicketStatementQuery_globals_,
        &::arrow::flight::protocol::sql::CommandPreparedStatementQuery_globals_,
        &::arrow::flight::protocol::sql::CommandStatementUpdate_globals_,
        &::arrow::flight::protocol::sql::CommandPreparedStatementUpdate_globals_,
        &::arrow::flight::protocol::sql::CommandStatementIngest_TableDefinitionOptions_globals_,
        &::arrow::flight::protocol::sql::CommandStatementIngest_OptionsEntry_DoNotUse_globals_,
        &::arrow::flight::protocol::sql::CommandStatementIngest_globals_,
        &::arrow::flight::protocol::sql::DoPutUpdateResult_globals_,
        &::arrow::flight::protocol::sql::DoPutPreparedStatementResult_globals_,
        &::arrow::flight::protocol::sql::ActionCancelQueryRequest_globals_,
        &::arrow::flight::protocol::sql::ActionCancelQueryResult_globals_,
};
const char descriptor_table_protodef_FlightSql_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\017FlightSql.proto\022\031arrow.flight.protocol"
    ".sql\032 google/protobuf/descriptor.proto\"!"
    "\n\021CommandGetSqlInfo\022\014\n\004info\030\001 \003(\r\">\n\026Com"
    "mandGetXdbcTypeInfo\022\026\n\tdata_type\030\001 \001(\005H\000"
    "\210\001\001B\014\n\n_data_type\"\024\n\022CommandGetCatalogs\""
    "{\n\023CommandGetDbSchemas\022\024\n\007catalog\030\001 \001(\tH"
    "\000\210\001\001\022%\n\030db_schema_filter_pattern\030\002 \001(\tH\001"
    "\210\001\001B\n\n\010_catalogB\033\n\031_db_schema_filter_pat"
    "tern\"\353\001\n\020CommandGetTables\022\024\n\007catalog\030\001 \001"
    "(\tH\000\210\001\001\022%\n\030db_schema_filter_pattern\030\002 \001("
    "\tH\001\210\001\001\022&\n\031table_name_filter_pattern\030\003 \001("
    "\tH\002\210\001\001\022\023\n\013table_types\030\004 \003(\t\022\026\n\016include_s"
    "chema\030\005 \001(\010B\n\n\010_catalogB\033\n\031_db_schema_fi"
    "lter_patternB\034\n\032_table_name_filter_patte"
    "rn\"\026\n\024CommandGetTableTypes\"n\n\025CommandGet"
    "PrimaryKeys\022\024\n\007catalog\030\001 \001(\tH\000\210\001\001\022\026\n\tdb_"
    "schema\030\002 \001(\tH\001\210\001\001\022\r\n\005table\030\003 \001(\tB\n\n\010_cat"
    "alogB\014\n\n_db_schema\"o\n\026CommandGetExported"
    "Keys\022\024\n\007catalog\030\001 \001(\tH\000\210\001\001\022\026\n\tdb_schema\030"
    "\002 \001(\tH\001\210\001\001\022\r\n\005table\030\003 \001(\tB\n\n\010_catalogB\014\n"
    "\n_db_schema\"o\n\026CommandGetImportedKeys\022\024\n"
    "\007catalog\030\001 \001(\tH\000\210\001\001\022\026\n\tdb_schema\030\002 \001(\tH\001"
    "\210\001\001\022\r\n\005table\030\003 \001(\tB\n\n\010_catalogB\014\n\n_db_sc"
    "hema\"\346\001\n\030CommandGetCrossReference\022\027\n\npk_"
    "catalog\030\001 \001(\tH\000\210\001\001\022\031\n\014pk_db_schema\030\002 \001(\t"
    "H\001\210\001\001\022\020\n\010pk_table\030\003 \001(\t\022\027\n\nfk_catalog\030\004 "
    "\001(\tH\002\210\001\001\022\031\n\014fk_db_schema\030\005 \001(\tH\003\210\001\001\022\020\n\010f"
    "k_table\030\006 \001(\tB\r\n\013_pk_catalogB\017\n\r_pk_db_s"
    "chemaB\r\n\013_fk_catalogB\017\n\r_fk_db_schema\"e\n"
    "$ActionCreatePreparedStatementRequest\022\r\n"
    "\005query\030\001 \001(\t\022\033\n\016transaction_id\030\002 \001(\014H\000\210\001"
    "\001B\021\n\017_transaction_id\".\n\rSubstraitPlan\022\014\n"
    "\004plan\030\001 \001(\014\022\017\n\007version\030\002 \001(\t\"\222\001\n(ActionC"
    "reatePreparedSubstraitPlanRequest\0226\n\004pla"
    "n\030\001 \001(\0132(.arrow.flight.protocol.sql.Subs"
    "traitPlan\022\033\n\016transaction_id\030\002 \001(\014H\000\210\001\001B\021"
    "\n\017_transaction_id\"z\n#ActionCreatePrepare"
    "dStatementResult\022!\n\031prepared_statement_h"
    "andle\030\001 \001(\014\022\026\n\016dataset_schema\030\002 \001(\014\022\030\n\020p"
    "arameter_schema\030\003 \001(\014\"H\n#ActionClosePrep"
    "aredStatementRequest\022!\n\031prepared_stateme"
    "nt_handle\030\001 \001(\014\"\037\n\035ActionBeginTransactio"
    "nRequest\"C\n\033ActionBeginSavepointRequest\022"
    "\026\n\016transaction_id\030\001 \001(\014\022\014\n\004name\030\002 \001(\t\"6\n"
    "\034ActionBeginTransactionResult\022\026\n\016transac"
    "tion_id\030\001 \001(\014\"2\n\032ActionBeginSavepointRes"
    "ult\022\024\n\014savepoint_id\030\001 \001(\014\"\371\001\n\033ActionEndT"
    "ransactionRequest\022\026\n\016transaction_id\030\001 \001("
    "\014\022U\n\006action\030\002 \001(\0162E.arrow.flight.protoco"
    "l.sql.ActionEndTransactionRequest.EndTra"
    "nsaction\"k\n\016EndTransaction\022\037\n\033END_TRANSA"
    "CTION_UNSPECIFIED\020\000\022\032\n\026END_TRANSACTION_C"
    "OMMIT\020\001\022\034\n\030END_TRANSACTION_ROLLBACK\020\002\"\352\001"
    "\n\031ActionEndSavepointRequest\022\024\n\014savepoint"
    "_id\030\001 \001(\014\022Q\n\006action\030\002 \001(\0162A.arrow.flight"
    ".protocol.sql.ActionEndSavepointRequest."
    "EndSavepoint\"d\n\014EndSavepoint\022\035\n\031END_SAVE"
    "POINT_UNSPECIFIED\020\000\022\031\n\025END_SAVEPOINT_REL"
    "EASE\020\001\022\032\n\026END_SAVEPOINT_ROLLBACK\020\002\"V\n\025Co"
    "mmandStatementQuery\022\r\n\005query\030\001 \001(\t\022\033\n\016tr"
    "ansaction_id\030\002 \001(\014H\000\210\001\001B\021\n\017_transaction_"
    "id\"\207\001\n\035CommandStatementSubstraitPlan\0226\n\004"
    "plan\030\001 \001(\0132(.arrow.flight.protocol.sql.S"
    "ubstraitPlan\022\033\n\016transaction_id\030\002 \001(\014H\000\210\001"
    "\001B\021\n\017_transaction_id\"0\n\024TicketStatementQ"
    "uery\022\030\n\020statement_handle\030\001 \001(\014\"B\n\035Comman"
    "dPreparedStatementQuery\022!\n\031prepared_stat"
    "ement_handle\030\001 \001(\014\"W\n\026CommandStatementUp"
    "date\022\r\n\005query\030\001 \001(\t\022\033\n\016transaction_id\030\002 "
    "\001(\014H\000\210\001\001B\021\n\017_transaction_id\"C\n\036CommandPr"
    "eparedStatementUpdate\022!\n\031prepared_statem"
    "ent_handle\030\001 \001(\014\"\266\007\n\026CommandStatementIng"
    "est\022j\n\030table_definition_options\030\001 \001(\0132H."
    "arrow.flight.protocol.sql.CommandStateme"
    "ntIngest.TableDefinitionOptions\022\r\n\005table"
    "\030\002 \001(\t\022\023\n\006schema\030\003 \001(\tH\000\210\001\001\022\024\n\007catalog\030\004"
    " \001(\tH\001\210\001\001\022\021\n\ttemporary\030\005 \001(\010\022\033\n\016transact"
    "ion_id\030\006 \001(\014H\002\210\001\001\022P\n\007options\030\350\007 \003(\0132>.ar"
    "row.flight.protocol.sql.CommandStatement"
    "Ingest.OptionsEntry\032\231\004\n\026TableDefinitionO"
    "ptions\022r\n\014if_not_exist\030\001 \001(\0162\\.arrow.fli"
    "ght.protocol.sql.CommandStatementIngest."
    "TableDefinitionOptions.TableNotExistOpti"
    "on\022m\n\tif_exists\030\002 \001(\0162Z.arrow.flight.pro"
    "tocol.sql.CommandStatementIngest.TableDe"
    "finitionOptions.TableExistsOption\"\201\001\n\023Ta"
    "bleNotExistOption\022&\n\"TABLE_NOT_EXIST_OPT"
    "ION_UNSPECIFIED\020\000\022!\n\035TABLE_NOT_EXIST_OPT"
    "ION_CREATE\020\001\022\037\n\033TABLE_NOT_EXIST_OPTION_F"
    "AIL\020\002\"\227\001\n\021TableExistsOption\022#\n\037TABLE_EXI"
    "STS_OPTION_UNSPECIFIED\020\000\022\034\n\030TABLE_EXISTS"
    "_OPTION_FAIL\020\001\022\036\n\032TABLE_EXISTS_OPTION_AP"
    "PEND\020\002\022\037\n\033TABLE_EXISTS_OPTION_REPLACE\020\003\032"
    ".\n\014OptionsEntry\022\013\n\003key\030\001 \001(\t\022\r\n\005value\030\002 "
    "\001(\t:\0028\001B\t\n\007_schemaB\n\n\010_catalogB\021\n\017_trans"
    "action_id\")\n\021DoPutUpdateResult\022\024\n\014record"
    "_count\030\001 \001(\003\"d\n\034DoPutPreparedStatementRe"
    "sult\022&\n\031prepared_statement_handle\030\001 \001(\014H"
    "\000\210\001\001B\034\n\032_prepared_statement_handle\",\n\030Ac"
    "tionCancelQueryRequest\022\014\n\004info\030\001 \001(\014:\002\030\001"
    "\"\374\001\n\027ActionCancelQueryResult\022O\n\006result\030\001"
    " \001(\0162\?.arrow.flight.protocol.sql.ActionC"
    "ancelQueryResult.CancelResult\"\213\001\n\014Cancel"
    "Result\022\035\n\031CANCEL_RESULT_UNSPECIFIED\020\000\022\033\n"
    "\027CANCEL_RESULT_CANCELLED\020\001\022\034\n\030CANCEL_RES"
    "ULT_CANCELLING\020\002\022!\n\035CANCEL_RESULT_NOT_CA"
    "NCELLABLE\020\003:\002\030\001*\222\031\n\007SqlInfo\022\032\n\026FLIGHT_SQ"
    "L_SERVER_NAME\020\000\022\035\n\031FLIGHT_SQL_SERVER_VER"
    "SION\020\001\022#\n\037FLIGHT_SQL_SERVER_ARROW_VERSIO"
    "N\020\002\022\037\n\033FLIGHT_SQL_SERVER_READ_ONLY\020\003\022\031\n\025"
    "FLIGHT_SQL_SERVER_SQL\020\004\022\037\n\033FLIGHT_SQL_SE"
    "RVER_SUBSTRAIT\020\005\022+\n\'FLIGHT_SQL_SERVER_SU"
    "BSTRAIT_MIN_VERSION\020\006\022+\n\'FLIGHT_SQL_SERV"
    "ER_SUBSTRAIT_MAX_VERSION\020\007\022!\n\035FLIGHT_SQL"
    "_SERVER_TRANSACTION\020\010\022\034\n\030FLIGHT_SQL_SERV"
    "ER_CANCEL\020\t\022$\n FLIGHT_SQL_SERVER_BULK_IN"
    "GESTION\020\n\0223\n/FLIGHT_SQL_SERVER_INGEST_TR"
    "ANSACTIONS_SUPPORTED\020\013\022\'\n#FLIGHT_SQL_SER"
    "VER_STATEMENT_TIMEOUT\020d\022)\n%FLIGHT_SQL_SE"
    "RVER_TRANSACTION_TIMEOUT\020e\022\024\n\017SQL_DDL_CA"
    "TALOG\020\364\003\022\023\n\016SQL_DDL_SCHEMA\020\365\003\022\022\n\rSQL_DDL"
    "_TABLE\020\366\003\022\030\n\023SQL_IDENTIFIER_CASE\020\367\003\022\036\n\031S"
    "QL_IDENTIFIER_QUOTE_CHAR\020\370\003\022\037\n\032SQL_QUOTE"
    "D_IDENTIFIER_CASE\020\371\003\022\"\n\035SQL_ALL_TABLES_A"
    "RE_SELECTABLE\020\372\003\022\026\n\021SQL_NULL_ORDERING\020\373\003"
    "\022\021\n\014SQL_KEYWORDS\020\374\003\022\032\n\025SQL_NUMERIC_FUNCT"
    "IONS\020\375\003\022\031\n\024SQL_STRING_FUNCTIONS\020\376\003\022\031\n\024SQ"
    "L_SYSTEM_FUNCTIONS\020\377\003\022\033\n\026SQL_DATETIME_FU"
    "NCTIONS\020\200\004\022\035\n\030SQL_SEARCH_STRING_ESCAPE\020\201"
    "\004\022\036\n\031SQL_EXTRA_NAME_CHARACTERS\020\202\004\022!\n\034SQL"
    "_SUPPORTS_COLUMN_ALIASING\020\203\004\022\037\n\032SQL_NULL"
    "_PLUS_NULL_IS_NULL\020\204\004\022\031\n\024SQL_SUPPORTS_CO"
    "NVERT\020\205\004\022)\n$SQL_SUPPORTS_TABLE_CORRELATI"
    "ON_NAMES\020\206\004\0223\n.SQL_SUPPORTS_DIFFERENT_TA"
    "BLE_CORRELATION_NAMES\020\207\004\022)\n$SQL_SUPPORTS"
    "_EXPRESSIONS_IN_ORDER_BY\020\210\004\022$\n\037SQL_SUPPO"
    "RTS_ORDER_BY_UNRELATED\020\211\004\022\033\n\026SQL_SUPPORT"
    "ED_GROUP_BY\020\212\004\022$\n\037SQL_SUPPORTS_LIKE_ESCA"
    "PE_CLAUSE\020\213\004\022&\n!SQL_SUPPORTS_NON_NULLABL"
    "E_COLUMNS\020\214\004\022\032\n\025SQL_SUPPORTED_GRAMMAR\020\215\004"
    "\022\037\n\032SQL_ANSI92_SUPPORTED_LEVEL\020\216\004\0220\n+SQL"
    "_SUPPORTS_INTEGRITY_ENHANCEMENT_FACILITY"
    "\020\217\004\022\"\n\035SQL_OUTER_JOINS_SUPPORT_LEVEL\020\220\004\022"
    "\024\n\017SQL_SCHEMA_TERM\020\221\004\022\027\n\022SQL_PROCEDURE_T"
    "ERM\020\222\004\022\025\n\020SQL_CATALOG_TERM\020\223\004\022\031\n\024SQL_CAT"
    "ALOG_AT_START\020\224\004\022\"\n\035SQL_SCHEMAS_SUPPORTE"
    "D_ACTIONS\020\225\004\022#\n\036SQL_CATALOGS_SUPPORTED_A"
    "CTIONS\020\226\004\022&\n!SQL_SUPPORTED_POSITIONED_CO"
    "MMANDS\020\227\004\022$\n\037SQL_SELECT_FOR_UPDATE_SUPPO"
    "RTED\020\230\004\022$\n\037SQL_STORED_PROCEDURES_SUPPORT"
    "ED\020\231\004\022\035\n\030SQL_SUPPORTED_SUBQUERIES\020\232\004\022(\n#"
    "SQL_CORRELATED_SUBQUERIES_SUPPORTED\020\233\004\022\031"
    "\n\024SQL_SUPPORTED_UNIONS\020\234\004\022\"\n\035SQL_MAX_BIN"
    "ARY_LITERAL_LENGTH\020\235\004\022 \n\033SQL_MAX_CHAR_LI"
    "TERAL_LENGTH\020\236\004\022\037\n\032SQL_MAX_COLUMN_NAME_L"
    "ENGTH\020\237\004\022 \n\033SQL_MAX_COLUMNS_IN_GROUP_BY\020"
    "\240\004\022\035\n\030SQL_MAX_COLUMNS_IN_INDEX\020\241\004\022 \n\033SQL"
    "_MAX_COLUMNS_IN_ORDER_BY\020\242\004\022\036\n\031SQL_MAX_C"
    "OLUMNS_IN_SELECT\020\243\004\022\035\n\030SQL_MAX_COLUMNS_I"
    "N_TABLE\020\244\004\022\030\n\023SQL_MAX_CONNECTIONS\020\245\004\022\037\n\032"
    "SQL_MAX_CURSOR_NAME_LENGTH\020\246\004\022\031\n\024SQL_MAX"
    "_INDEX_LENGTH\020\247\004\022\036\n\031SQL_DB_SCHEMA_NAME_L"
    "ENGTH\020\250\004\022\"\n\035SQL_MAX_PROCEDURE_NAME_LENGT"
    "H\020\251\004\022 \n\033SQL_MAX_CATALOG_NAME_LENGTH\020\252\004\022\025"
    "\n\020SQL_MAX_ROW_SIZE\020\253\004\022$\n\037SQL_MAX_ROW_SIZ"
    "E_INCLUDES_BLOBS\020\254\004\022\035\n\030SQL_MAX_STATEMENT"
    "_LENGTH\020\255\004\022\027\n\022SQL_MAX_STATEMENTS\020\256\004\022\036\n\031S"
    "QL_MAX_TABLE_NAME_LENGTH\020\257\004\022\035\n\030SQL_MAX_T"
    "ABLES_IN_SELECT\020\260\004\022\034\n\027SQL_MAX_USERNAME_L"
    "ENGTH\020\261\004\022&\n!SQL_DEFAULT_TRANSACTION_ISOL"
    "ATION\020\262\004\022\037\n\032SQL_TRANSACTIONS_SUPPORTED\020\263"
    "\004\0220\n+SQL_SUPPORTED_TRANSACTIONS_ISOLATIO"
    "N_LEVELS\020\264\004\0222\n-SQL_DATA_DEFINITION_CAUSE"
    "S_TRANSACTION_COMMIT\020\265\004\0221\n,SQL_DATA_DEFI"
    "NITIONS_IN_TRANSACTIONS_IGNORED\020\266\004\022#\n\036SQ"
    "L_SUPPORTED_RESULT_SET_TYPES\020\267\004\022;\n6SQL_S"
    "UPPORTED_CONCURRENCIES_FOR_RESULT_SET_UN"
    "SPECIFIED\020\270\004\022<\n7SQL_SUPPORTED_CONCURRENC"
    "IES_FOR_RESULT_SET_FORWARD_ONLY\020\271\004\022@\n;SQ"
    "L_SUPPORTED_CONCURRENCIES_FOR_RESULT_SET"
    "_SCROLL_SENSITIVE\020\272\004\022B\n=SQL_SUPPORTED_CO"
    "NCURRENCIES_FOR_RESULT_SET_SCROLL_INSENS"
    "ITIVE\020\273\004\022 \n\033SQL_BATCH_UPDATES_SUPPORTED\020"
    "\274\004\022\035\n\030SQL_SAVEPOINTS_SUPPORTED\020\275\004\022#\n\036SQL"
    "_NAMED_PARAMETERS_SUPPORTED\020\276\004\022\035\n\030SQL_LO"
    "CATORS_UPDATE_COPY\020\277\004\0225\n0SQL_STORED_FUNC"
    "TIONS_USING_CALL_SYNTAX_SUPPORTED\020\300\004*\221\001\n"
    "\027SqlSupportedTransaction\022\"\n\036SQL_SUPPORTE"
    "D_TRANSACTION_NONE\020\000\022)\n%SQL_SUPPORTED_TR"
    "ANSACTION_TRANSACTION\020\001\022\'\n#SQL_SUPPORTED"
    "_TRANSACTION_SAVEPOINT\020\002*\262\001\n\033SqlSupporte"
    "dCaseSensitivity\022 \n\034SQL_CASE_SENSITIVITY"
    "_UNKNOWN\020\000\022)\n%SQL_CASE_SENSITIVITY_CASE_"
    "INSENSITIVE\020\001\022\"\n\036SQL_CASE_SENSITIVITY_UP"
    "PERCASE\020\002\022\"\n\036SQL_CASE_SENSITIVITY_LOWERC"
    "ASE\020\003*\202\001\n\017SqlNullOrdering\022\031\n\025SQL_NULLS_S"
    "ORTED_HIGH\020\000\022\030\n\024SQL_NULLS_SORTED_LOW\020\001\022\035"
    "\n\031SQL_NULLS_SORTED_AT_START\020\002\022\033\n\027SQL_NUL"
    "LS_SORTED_AT_END\020\003*^\n\023SupportedSqlGramma"
    "r\022\027\n\023SQL_MINIMUM_GRAMMAR\020\000\022\024\n\020SQL_CORE_G"
    "RAMMAR\020\001\022\030\n\024SQL_EXTENDED_GRAMMAR\020\002*h\n\036Su"
    "pportedAnsi92SqlGrammarLevel\022\024\n\020ANSI92_E"
    "NTRY_SQL\020\000\022\033\n\027ANSI92_INTERMEDIATE_SQL\020\001\022"
    "\023\n\017ANSI92_FULL_SQL\020\002*m\n\031SqlOuterJoinsSup"
    "portLevel\022\031\n\025SQL_JOINS_UNSUPPORTED\020\000\022\033\n\027"
    "SQL_LIMITED_OUTER_JOINS\020\001\022\030\n\024SQL_FULL_OU"
    "TER_JOINS\020\002*Q\n\023SqlSupportedGroupBy\022\032\n\026SQ"
    "L_GROUP_BY_UNRELATED\020\000\022\036\n\032SQL_GROUP_BY_B"
    "EYOND_SELECT\020\001*\220\001\n\032SqlSupportedElementAc"
    "tions\022\"\n\036SQL_ELEMENT_IN_PROCEDURE_CALLS\020"
    "\000\022$\n SQL_ELEMENT_IN_INDEX_DEFINITIONS\020\001\022"
    "(\n$SQL_ELEMENT_IN_PRIVILEGE_DEFINITIONS\020"
    "\002*V\n\036SqlSupportedPositionedCommands\022\031\n\025S"
    "QL_POSITIONED_DELETE\020\000\022\031\n\025SQL_POSITIONED"
    "_UPDATE\020\001*\227\001\n\026SqlSupportedSubqueries\022!\n\035"
    "SQL_SUBQUERIES_IN_COMPARISONS\020\000\022\034\n\030SQL_S"
    "UBQUERIES_IN_EXISTS\020\001\022\031\n\025SQL_SUBQUERIES_"
    "IN_INS\020\002\022!\n\035SQL_SUBQUERIES_IN_QUANTIFIED"
    "S\020\003*6\n\022SqlSupportedUnions\022\r\n\tSQL_UNION\020\000"
    "\022\021\n\rSQL_UNION_ALL\020\001*\311\001\n\034SqlTransactionIs"
    "olationLevel\022\030\n\024SQL_TRANSACTION_NONE\020\000\022$"
    "\n SQL_TRANSACTION_READ_UNCOMMITTED\020\001\022\"\n\036"
    "SQL_TRANSACTION_READ_COMMITTED\020\002\022#\n\037SQL_"
    "TRANSACTION_REPEATABLE_READ\020\003\022 \n\034SQL_TRA"
    "NSACTION_SERIALIZABLE\020\004*\211\001\n\030SqlSupported"
    "Transactions\022\037\n\033SQL_TRANSACTION_UNSPECIF"
    "IED\020\000\022$\n SQL_DATA_DEFINITION_TRANSACTION"
    "S\020\001\022&\n\"SQL_DATA_MANIPULATION_TRANSACTION"
    "S\020\002*\274\001\n\031SqlSupportedResultSetType\022#\n\037SQL"
    "_RESULT_SET_TYPE_UNSPECIFIED\020\000\022$\n SQL_RE"
    "SULT_SET_TYPE_FORWARD_ONLY\020\001\022*\n&SQL_RESU"
    "LT_SET_TYPE_SCROLL_INSENSITIVE\020\002\022(\n$SQL_"
    "RESULT_SET_TYPE_SCROLL_SENSITIVE\020\003*\242\001\n S"
    "qlSupportedResultSetConcurrency\022*\n&SQL_R"
    "ESULT_SET_CONCURRENCY_UNSPECIFIED\020\000\022(\n$S"
    "QL_RESULT_SET_CONCURRENCY_READ_ONLY\020\001\022(\n"
    "$SQL_RESULT_SET_CONCURRENCY_UPDATABLE\020\002*"
    "\231\004\n\022SqlSupportsConvert\022\026\n\022SQL_CONVERT_BI"
    "GINT\020\000\022\026\n\022SQL_CONVERT_BINARY\020\001\022\023\n\017SQL_CO"
    "NVERT_BIT\020\002\022\024\n\020SQL_CONVERT_CHAR\020\003\022\024\n\020SQL"
    "_CONVERT_DATE\020\004\022\027\n\023SQL_CONVERT_DECIMAL\020\005"
    "\022\025\n\021SQL_CONVERT_FLOAT\020\006\022\027\n\023SQL_CONVERT_I"
    "NTEGER\020\007\022!\n\035SQL_CONVERT_INTERVAL_DAY_TIM"
    "E\020\010\022#\n\037SQL_CONVERT_INTERVAL_YEAR_MONTH\020\t"
    "\022\035\n\031SQL_CONVERT_LONGVARBINARY\020\n\022\033\n\027SQL_C"
    "ONVERT_LONGVARCHAR\020\013\022\027\n\023SQL_CONVERT_NUME"
    "RIC\020\014\022\024\n\020SQL_CONVERT_REAL\020\r\022\030\n\024SQL_CONVE"
    "RT_SMALLINT\020\016\022\024\n\020SQL_CONVERT_TIME\020\017\022\031\n\025S"
    "QL_CONVERT_TIMESTAMP\020\020\022\027\n\023SQL_CONVERT_TI"
    "NYINT\020\021\022\031\n\025SQL_CONVERT_VARBINARY\020\022\022\027\n\023SQ"
    "L_CONVERT_VARCHAR\020\023*\217\004\n\014XdbcDataType\022\025\n\021"
    "XDBC_UNKNOWN_TYPE\020\000\022\r\n\tXDBC_CHAR\020\001\022\020\n\014XD"
    "BC_NUMERIC\020\002\022\020\n\014XDBC_DECIMAL\020\003\022\020\n\014XDBC_I"
    "NTEGER\020\004\022\021\n\rXDBC_SMALLINT\020\005\022\016\n\nXDBC_FLOA"
    "T\020\006\022\r\n\tXDBC_REAL\020\007\022\017\n\013XDBC_DOUBLE\020\010\022\021\n\rX"
    "DBC_DATETIME\020\t\022\021\n\rXDBC_INTERVAL\020\n\022\020\n\014XDB"
    "C_VARCHAR\020\014\022\r\n\tXDBC_DATE\020[\022\r\n\tXDBC_TIME\020"
    "\\\022\022\n\016XDBC_TIMESTAMP\020]\022\035\n\020XDBC_LONGVARCHA"
    "R\020\377\377\377\377\377\377\377\377\377\001\022\030\n\013XDBC_BINARY\020\376\377\377\377\377\377\377\377\377\001\022\033"
    "\n\016XDBC_VARBINARY\020\375\377\377\377\377\377\377\377\377\001\022\037\n\022XDBC_LONG"
    "VARBINARY\020\374\377\377\377\377\377\377\377\377\001\022\030\n\013XDBC_BIGINT\020\373\377\377\377"
    "\377\377\377\377\377\001\022\031\n\014XDBC_TINYINT\020\372\377\377\377\377\377\377\377\377\001\022\025\n\010XDB"
    "C_BIT\020\371\377\377\377\377\377\377\377\377\001\022\027\n\nXDBC_WCHAR\020\370\377\377\377\377\377\377\377\377"
    "\001\022\032\n\rXDBC_WVARCHAR\020\367\377\377\377\377\377\377\377\377\001*\243\010\n\023XdbcDa"
    "tetimeSubcode\022\030\n\024XDBC_SUBCODE_UNKNOWN\020\000\022"
    "\025\n\021XDBC_SUBCODE_YEAR\020\001\022\025\n\021XDBC_SUBCODE_D"
    "ATE\020\001\022\025\n\021XDBC_SUBCODE_TIME\020\002\022\026\n\022XDBC_SUB"
    "CODE_MONTH\020\002\022\032\n\026XDBC_SUBCODE_TIMESTAMP\020\003"
    "\022\024\n\020XDBC_SUBCODE_DAY\020\003\022#\n\037XDBC_SUBCODE_T"
    "IME_WITH_TIMEZONE\020\004\022\025\n\021XDBC_SUBCODE_HOUR"
    "\020\004\022(\n$XDBC_SUBCODE_TIMESTAMP_WITH_TIMEZO"
    "NE\020\005\022\027\n\023XDBC_SUBCODE_MINUTE\020\005\022\027\n\023XDBC_SU"
    "BCODE_SECOND\020\006\022\036\n\032XDBC_SUBCODE_YEAR_TO_M"
    "ONTH\020\007\022\034\n\030XDBC_SUBCODE_DAY_TO_HOUR\020\010\022\036\n\032"
    "XDBC_SUBCODE_DAY_TO_MINUTE\020\t\022\036\n\032XDBC_SUB"
    "CODE_DAY_TO_SECOND\020\n\022\037\n\033XDBC_SUBCODE_HOU"
    "R_TO_MINUTE\020\013\022\037\n\033XDBC_SUBCODE_HOUR_TO_SE"
    "COND\020\014\022!\n\035XDBC_SUBCODE_MINUTE_TO_SECOND\020"
    "\r\022\036\n\032XDBC_SUBCODE_INTERVAL_YEAR\020e\022\037\n\033XDB"
    "C_SUBCODE_INTERVAL_MONTH\020f\022\035\n\031XDBC_SUBCO"
    "DE_INTERVAL_DAY\020g\022\036\n\032XDBC_SUBCODE_INTERV"
    "AL_HOUR\020h\022 \n\034XDBC_SUBCODE_INTERVAL_MINUT"
    "E\020i\022 \n\034XDBC_SUBCODE_INTERVAL_SECOND\020j\022\'\n"
    "#XDBC_SUBCODE_INTERVAL_YEAR_TO_MONTH\020k\022%"
    "\n!XDBC_SUBCODE_INTERVAL_DAY_TO_HOUR\020l\022\'\n"
    "#XDBC_SUBCODE_INTERVAL_DAY_TO_MINUTE\020m\022\'"
    "\n#XDBC_SUBCODE_INTERVAL_DAY_TO_SECOND\020n\022"
    "(\n$XDBC_SUBCODE_INTERVAL_HOUR_TO_MINUTE\020"
    "o\022(\n$XDBC_SUBCODE_INTERVAL_HOUR_TO_SECON"
    "D\020p\022*\n&XDBC_SUBCODE_INTERVAL_MINUTE_TO_S"
    "ECOND\020q\032\002\020\001*W\n\010Nullable\022\030\n\024NULLABILITY_N"
    "O_NULLS\020\000\022\030\n\024NULLABILITY_NULLABLE\020\001\022\027\n\023N"
    "ULLABILITY_UNKNOWN\020\002*a\n\nSearchable\022\023\n\017SE"
    "ARCHABLE_NONE\020\000\022\023\n\017SEARCHABLE_CHAR\020\001\022\024\n\020"
    "SEARCHABLE_BASIC\020\002\022\023\n\017SEARCHABLE_FULL\020\003*"
    "\\\n\021UpdateDeleteRules\022\013\n\007CASCADE\020\000\022\014\n\010RES"
    "TRICT\020\001\022\014\n\010SET_NULL\020\002\022\r\n\tNO_ACTION\020\003\022\017\n\013"
    "SET_DEFAULT\020\004:6\n\014experimental\022\037.google.p"
    "rotobuf.MessageOptions\030\350\007 \001(\010BV\n org.apa"
    "che.arrow.flight.sql.implZ2github.com/ap"
    "ache/arrow-go/arrow/flight/gen/flightb\006p"
    "roto3"
};
static const ::_pbi::DescriptorTable* PROTOBUF_NONNULL const
    descriptor_table_FlightSql_2eproto_deps[1] = {
        &::descriptor_table_google_2fprotobuf_2fdescriptor_2eproto,
};
static ::absl::once_flag descriptor_table_FlightSql_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_FlightSql_2eproto = {
    false,
    false,
    12045,
    descriptor_table_protodef_FlightSql_2eproto,
    "FlightSql.proto",
    &descriptor_table_FlightSql_2eproto_once,
    descriptor_table_FlightSql_2eproto_deps,
    1,
    34,
    schemas,
    file_message_globals,
    TableStruct_FlightSql_2eproto::offsets,
    file_level_enum_descriptors_FlightSql_2eproto,
    file_level_service_descriptors_FlightSql_2eproto,
};
namespace arrow {
namespace flight {
namespace protocol {
namespace sql {
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
ActionEndTransactionRequest_EndTransaction_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[0];
}
PROTOBUF_CONSTINIT const uint32_t ActionEndTransactionRequest_EndTransaction_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
ActionEndSavepointRequest_EndSavepoint_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[1];
}
PROTOBUF_CONSTINIT const uint32_t ActionEndSavepointRequest_EndSavepoint_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
CommandStatementIngest_TableDefinitionOptions_TableNotExistOption_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[2];
}
PROTOBUF_CONSTINIT const uint32_t CommandStatementIngest_TableDefinitionOptions_TableNotExistOption_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
CommandStatementIngest_TableDefinitionOptions_TableExistsOption_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[3];
}
PROTOBUF_CONSTINIT const uint32_t CommandStatementIngest_TableDefinitionOptions_TableExistsOption_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
ActionCancelQueryResult_CancelResult_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[4];
}
PROTOBUF_CONSTINIT const uint32_t ActionCancelQueryResult_CancelResult_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlInfo_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[5];
}
PROTOBUF_CONSTINIT const uint32_t SqlInfo_internal_data_[] = {
    786432u, 576u, 0u, 0u, 50331648u, 0u, 0u, 0u, 0u, 0u, 0u, 0u, 0u, 0u, 0u, 0u, 0u, 4294967040u, 4294967295u, 2097151u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedTransaction_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[6];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedTransaction_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedCaseSensitivity_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[7];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedCaseSensitivity_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlNullOrdering_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[8];
}
PROTOBUF_CONSTINIT const uint32_t SqlNullOrdering_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SupportedSqlGrammar_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[9];
}
PROTOBUF_CONSTINIT const uint32_t SupportedSqlGrammar_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SupportedAnsi92SqlGrammarLevel_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[10];
}
PROTOBUF_CONSTINIT const uint32_t SupportedAnsi92SqlGrammarLevel_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlOuterJoinsSupportLevel_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[11];
}
PROTOBUF_CONSTINIT const uint32_t SqlOuterJoinsSupportLevel_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedGroupBy_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[12];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedGroupBy_internal_data_[] = {
    131072u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedElementActions_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[13];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedElementActions_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedPositionedCommands_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[14];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedPositionedCommands_internal_data_[] = {
    131072u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedSubqueries_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[15];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedSubqueries_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedUnions_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[16];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedUnions_internal_data_[] = {
    131072u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlTransactionIsolationLevel_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[17];
}
PROTOBUF_CONSTINIT const uint32_t SqlTransactionIsolationLevel_internal_data_[] = {
    327680u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedTransactions_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[18];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedTransactions_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedResultSetType_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[19];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedResultSetType_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportedResultSetConcurrency_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[20];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportedResultSetConcurrency_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
SqlSupportsConvert_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[21];
}
PROTOBUF_CONSTINIT const uint32_t SqlSupportsConvert_internal_data_[] = {
    1310720u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
XdbcDataType_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[22];
}
PROTOBUF_CONSTINIT const uint32_t XdbcDataType_internal_data_[] = {
    1376247u, 96u, 2u, 0u, 458752u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
XdbcDatetimeSubcode_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[23];
}
PROTOBUF_CONSTINIT const uint32_t XdbcDatetimeSubcode_internal_data_[] = {
    917504u, 128u, 0u, 0u, 4286578688u, 15u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
Nullable_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[24];
}
PROTOBUF_CONSTINIT const uint32_t Nullable_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
Searchable_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[25];
}
PROTOBUF_CONSTINIT const uint32_t Searchable_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
UpdateDeleteRules_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_FlightSql_2eproto);
  return file_level_enum_descriptors_FlightSql_2eproto[26];
}
PROTOBUF_CONSTINIT const uint32_t UpdateDeleteRules_internal_data_[] = {
    327680u, 0u, };
// ===================================================================

CommandGetSqlInfo::CommandGetSqlInfo(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetSqlInfo_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetSqlInfo)
}
PROTOBUF_NDEBUG_INLINE CommandGetSqlInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandGetSqlInfo& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        info_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::arrow::flight::protocol::sql::CommandGetSqlInfo,
              PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetSqlInfo, _impl_.info_)>()
          , from.info_
        }
        ,
        _info_cached_byte_size_{0} {}

CommandGetSqlInfo::CommandGetSqlInfo(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandGetSqlInfo& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetSqlInfo_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandGetSqlInfo* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandGetSqlInfo)
}
PROTOBUF_NDEBUG_INLINE CommandGetSqlInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        info_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::sql::CommandGetSqlInfo,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetSqlInfo, _impl_.info_)>()
         }
        ,
        _info_cached_byte_size_{0} {}

inline void CommandGetSqlInfo::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandGetSqlInfo::~CommandGetSqlInfo() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandGetSqlInfo)
  SharedDtor(*this);
}
inline void CommandGetSqlInfo::SharedDtor(MessageLite& self) {
  CommandGetSqlInfo& this_ = static_cast<CommandGetSqlInfo&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetSqlInfo_class_data_ =
        CommandGetSqlInfo::InternalGenerateClassData_(CommandGetSqlInfo_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetSqlInfo::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetSqlInfo_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetSqlInfo_class_data_.tc_table);
  return CommandGetSqlInfo_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetSqlInfo::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetSqlInfo_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetSqlInfo_globals_));
  return CommandGetSqlInfo_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetSqlInfo::ParseTableT_
    CommandGetSqlInfo::_table_ =
        CommandGetSqlInfo::InternalGenerateParseTable_(CommandGetSqlInfo_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandGetSqlInfo::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandGetSqlInfo)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.info_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandGetSqlInfo::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandGetSqlInfo& this_ = static_cast<const CommandGetSqlInfo&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandGetSqlInfo::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandGetSqlInfo& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandGetSqlInfo)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated uint32 info = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    {
      int byte_size = this_._impl_._info_cached_byte_size_.Get();
      if (byte_size > 0) {
        target = stream->WriteUInt32Packed(
            1, this_._internal_info(), byte_size, target);
      }
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandGetSqlInfo)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandGetSqlInfo::ByteSizeLong(const MessageLite& base) {
  const CommandGetSqlInfo& this_ = static_cast<const CommandGetSqlInfo&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandGetSqlInfo::ByteSizeLong() const {
  const CommandGetSqlInfo& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandGetSqlInfo)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated uint32 info = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          ::_pbi::WireFormatLite::UInt32SizeWithPackedTagSize(
              this_._internal_info(), 1,
              this_._impl_._info_cached_byte_size_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandGetSqlInfo::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandGetSqlInfo*>(&to_msg);
  auto& from = static_cast<const CommandGetSqlInfo&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandGetSqlInfo)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_info()->MergeFrom(from._internal_info());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandGetSqlInfo::CopyFrom(const CommandGetSqlInfo& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandGetSqlInfo)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandGetSqlInfo::InternalSwap(CommandGetSqlInfo* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.info_.InternalSwap(&other->_impl_.info_);
}

::google::protobuf::Metadata CommandGetSqlInfo::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandGetXdbcTypeInfo::CommandGetXdbcTypeInfo(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetXdbcTypeInfo_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetXdbcTypeInfo)
}
CommandGetXdbcTypeInfo::CommandGetXdbcTypeInfo(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const CommandGetXdbcTypeInfo& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetXdbcTypeInfo_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE CommandGetXdbcTypeInfo::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void CommandGetXdbcTypeInfo::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.data_type_ = {};
}
CommandGetXdbcTypeInfo::~CommandGetXdbcTypeInfo() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandGetXdbcTypeInfo)
  SharedDtor(*this);
}
inline void CommandGetXdbcTypeInfo::SharedDtor(MessageLite& self) {
  CommandGetXdbcTypeInfo& this_ = static_cast<CommandGetXdbcTypeInfo&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetXdbcTypeInfo_class_data_ =
        CommandGetXdbcTypeInfo::InternalGenerateClassData_(CommandGetXdbcTypeInfo_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetXdbcTypeInfo::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetXdbcTypeInfo_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetXdbcTypeInfo_class_data_.tc_table);
  return CommandGetXdbcTypeInfo_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetXdbcTypeInfo::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetXdbcTypeInfo_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetXdbcTypeInfo_globals_));
  return CommandGetXdbcTypeInfo_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetXdbcTypeInfo::ParseTableT_
    CommandGetXdbcTypeInfo::_table_ =
        CommandGetXdbcTypeInfo::InternalGenerateParseTable_(CommandGetXdbcTypeInfo_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandGetXdbcTypeInfo::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandGetXdbcTypeInfo)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.data_type_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandGetXdbcTypeInfo::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandGetXdbcTypeInfo& this_ = static_cast<const CommandGetXdbcTypeInfo&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandGetXdbcTypeInfo::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandGetXdbcTypeInfo& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandGetXdbcTypeInfo)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional int32 data_type = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target =
        ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<1>(
            stream, this_._internal_data_type(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandGetXdbcTypeInfo)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandGetXdbcTypeInfo::ByteSizeLong(const MessageLite& base) {
  const CommandGetXdbcTypeInfo& this_ = static_cast<const CommandGetXdbcTypeInfo&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandGetXdbcTypeInfo::ByteSizeLong() const {
  const CommandGetXdbcTypeInfo& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandGetXdbcTypeInfo)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // optional int32 data_type = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
          this_._internal_data_type());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandGetXdbcTypeInfo::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandGetXdbcTypeInfo*>(&to_msg);
  auto& from = static_cast<const CommandGetXdbcTypeInfo&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandGetXdbcTypeInfo)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_impl_.data_type_ = from._impl_.data_type_;
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandGetXdbcTypeInfo::CopyFrom(const CommandGetXdbcTypeInfo& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandGetXdbcTypeInfo)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandGetXdbcTypeInfo::InternalSwap(CommandGetXdbcTypeInfo* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.data_type_, other->_impl_.data_type_);
}

::google::protobuf::Metadata CommandGetXdbcTypeInfo::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandGetCatalogs::CommandGetCatalogs(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, CommandGetCatalogs_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetCatalogs)
}
CommandGetCatalogs::CommandGetCatalogs(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandGetCatalogs& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, CommandGetCatalogs_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandGetCatalogs* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandGetCatalogs)
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetCatalogs_class_data_ =
        CommandGetCatalogs::InternalGenerateClassData_(CommandGetCatalogs_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetCatalogs::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetCatalogs_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetCatalogs_class_data_.tc_table);
  return CommandGetCatalogs_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetCatalogs::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetCatalogs_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetCatalogs_globals_));
  return CommandGetCatalogs_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetCatalogs::ParseTableT_
    CommandGetCatalogs::_table_ =
        CommandGetCatalogs::InternalGenerateParseTable_(CommandGetCatalogs_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS







::google::protobuf::Metadata CommandGetCatalogs::GetMetadata() const {
  return ::google::protobuf::internal::ZeroFieldsBase::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandGetDbSchemas::CommandGetDbSchemas(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetDbSchemas_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetDbSchemas)
}
PROTOBUF_NDEBUG_INLINE CommandGetDbSchemas::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandGetDbSchemas& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        catalog_(arena, from.catalog_),
        db_schema_filter_pattern_(arena, from.db_schema_filter_pattern_) {}

CommandGetDbSchemas::CommandGetDbSchemas(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandGetDbSchemas& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetDbSchemas_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandGetDbSchemas* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandGetDbSchemas)
}
PROTOBUF_NDEBUG_INLINE CommandGetDbSchemas::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        catalog_(arena),
        db_schema_filter_pattern_(arena) {}

inline void CommandGetDbSchemas::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandGetDbSchemas::~CommandGetDbSchemas() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandGetDbSchemas)
  SharedDtor(*this);
}
inline void CommandGetDbSchemas::SharedDtor(MessageLite& self) {
  CommandGetDbSchemas& this_ = static_cast<CommandGetDbSchemas&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.catalog_.Destroy();
  this_._impl_.db_schema_filter_pattern_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetDbSchemas_class_data_ =
        CommandGetDbSchemas::InternalGenerateClassData_(CommandGetDbSchemas_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetDbSchemas::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetDbSchemas_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetDbSchemas_class_data_.tc_table);
  return CommandGetDbSchemas_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetDbSchemas::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetDbSchemas_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetDbSchemas_globals_));
  return CommandGetDbSchemas_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetDbSchemas::ParseTableT_
    CommandGetDbSchemas::_table_ =
        CommandGetDbSchemas::InternalGenerateParseTable_(CommandGetDbSchemas_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandGetDbSchemas::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandGetDbSchemas)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.catalog_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.db_schema_filter_pattern_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandGetDbSchemas::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandGetDbSchemas& this_ = static_cast<const CommandGetDbSchemas&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandGetDbSchemas::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandGetDbSchemas& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandGetDbSchemas)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string catalog = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_catalog();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetDbSchemas.catalog");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string db_schema_filter_pattern = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_db_schema_filter_pattern();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetDbSchemas.db_schema_filter_pattern");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandGetDbSchemas)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandGetDbSchemas::ByteSizeLong(const MessageLite& base) {
  const CommandGetDbSchemas& this_ = static_cast<const CommandGetDbSchemas&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandGetDbSchemas::ByteSizeLong() const {
  const CommandGetDbSchemas& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandGetDbSchemas)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // optional string catalog = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_catalog());
    }
    // optional string db_schema_filter_pattern = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_db_schema_filter_pattern());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandGetDbSchemas::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandGetDbSchemas*>(&to_msg);
  auto& from = static_cast<const CommandGetDbSchemas&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandGetDbSchemas)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_catalog(from._internal_catalog());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_db_schema_filter_pattern(from._internal_db_schema_filter_pattern());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandGetDbSchemas::CopyFrom(const CommandGetDbSchemas& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandGetDbSchemas)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandGetDbSchemas::InternalSwap(CommandGetDbSchemas* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.catalog_, &other->_impl_.catalog_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.db_schema_filter_pattern_, &other->_impl_.db_schema_filter_pattern_, arena);
}

::google::protobuf::Metadata CommandGetDbSchemas::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandGetTables::CommandGetTables(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetTables_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetTables)
}
PROTOBUF_NDEBUG_INLINE CommandGetTables::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandGetTables& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        table_types_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::arrow::flight::protocol::sql::CommandGetTables,
              PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetTables, _impl_.table_types_)>()
          , from.table_types_
        }
        ,
        catalog_(arena, from.catalog_),
        db_schema_filter_pattern_(arena, from.db_schema_filter_pattern_),
        table_name_filter_pattern_(arena, from.table_name_filter_pattern_) {}

CommandGetTables::CommandGetTables(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandGetTables& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetTables_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandGetTables* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.include_schema_ = from._impl_.include_schema_;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandGetTables)
}
PROTOBUF_NDEBUG_INLINE CommandGetTables::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        table_types_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::sql::CommandGetTables,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandGetTables, _impl_.table_types_)>()
         }
        ,
        catalog_(arena),
        db_schema_filter_pattern_(arena),
        table_name_filter_pattern_(arena) {}

inline void CommandGetTables::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.include_schema_ = {};
}
CommandGetTables::~CommandGetTables() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandGetTables)
  SharedDtor(*this);
}
inline void CommandGetTables::SharedDtor(MessageLite& self) {
  CommandGetTables& this_ = static_cast<CommandGetTables&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.catalog_.Destroy();
  this_._impl_.db_schema_filter_pattern_.Destroy();
  this_._impl_.table_name_filter_pattern_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetTables_class_data_ =
        CommandGetTables::InternalGenerateClassData_(CommandGetTables_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetTables::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetTables_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetTables_class_data_.tc_table);
  return CommandGetTables_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetTables::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetTables_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetTables_globals_));
  return CommandGetTables_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetTables::ParseTableT_
    CommandGetTables::_table_ =
        CommandGetTables::InternalGenerateParseTable_(CommandGetTables_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandGetTables::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandGetTables)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.table_types_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.catalog_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.db_schema_filter_pattern_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.table_name_filter_pattern_.ClearNonDefaultToEmpty();
    }
  }
  _impl_.include_schema_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandGetTables::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandGetTables& this_ = static_cast<const CommandGetTables&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandGetTables::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandGetTables& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandGetTables)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string catalog = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_catalog();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetTables.catalog");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string db_schema_filter_pattern = 2;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    const ::std::string& _s = this_._internal_db_schema_filter_pattern();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetTables.db_schema_filter_pattern");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // optional string table_name_filter_pattern = 3;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    const ::std::string& _s = this_._internal_table_name_filter_pattern();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetTables.table_name_filter_pattern");
    target = stream->WriteStringMaybeAliased(3, _s, target);
  }

  // repeated string table_types = 4;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_table_types_size(); i < n; ++i) {
      const auto& s = this_._internal_table_types().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetTables.table_types");
      target = stream->WriteString(4, s, target);
    }
  }

  // bool include_schema = 5;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (this_._internal_include_schema() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          5, this_._internal_include_schema(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandGetTables)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandGetTables::ByteSizeLong(const MessageLite& base) {
  const CommandGetTables& this_ = static_cast<const CommandGetTables&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandGetTables::ByteSizeLong() const {
  const CommandGetTables& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandGetTables)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    // repeated string table_types = 4;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_table_types().size());
      for (int i = 0, n = this_._internal_table_types().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_table_types().Get(i));
      }
    }
    // optional string catalog = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_catalog());
    }
    // optional string db_schema_filter_pattern = 2;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_db_schema_filter_pattern());
    }
    // optional string table_name_filter_pattern = 3;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_table_name_filter_pattern());
    }
    // bool include_schema = 5;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (this_._internal_include_schema() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandGetTables::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandGetTables*>(&to_msg);
  auto& from = static_cast<const CommandGetTables&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandGetTables)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_table_types()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_table_types());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_catalog(from._internal_catalog());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_set_db_schema_filter_pattern(from._internal_db_schema_filter_pattern());
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_internal_set_table_name_filter_pattern(from._internal_table_name_filter_pattern());
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (from._internal_include_schema() != 0) {
        _this->_impl_.include_schema_ = from._impl_.include_schema_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandGetTables::CopyFrom(const CommandGetTables& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandGetTables)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandGetTables::InternalSwap(CommandGetTables* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.table_types_.InternalSwap(&other->_impl_.table_types_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.catalog_, &other->_impl_.catalog_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.db_schema_filter_pattern_, &other->_impl_.db_schema_filter_pattern_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.table_name_filter_pattern_, &other->_impl_.table_name_filter_pattern_, arena);
  swap(_impl_.include_schema_, other->_impl_.include_schema_);
}

::google::protobuf::Metadata CommandGetTables::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandGetTableTypes::CommandGetTableTypes(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, CommandGetTableTypes_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetTableTypes)
}
CommandGetTableTypes::CommandGetTableTypes(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandGetTableTypes& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, CommandGetTableTypes_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandGetTableTypes* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandGetTableTypes)
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetTableTypes_class_data_ =
        CommandGetTableTypes::InternalGenerateClassData_(CommandGetTableTypes_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetTableTypes::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetTableTypes_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetTableTypes_class_data_.tc_table);
  return CommandGetTableTypes_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetTableTypes::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetTableTypes_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetTableTypes_globals_));
  return CommandGetTableTypes_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetTableTypes::ParseTableT_
    CommandGetTableTypes::_table_ =
        CommandGetTableTypes::InternalGenerateParseTable_(CommandGetTableTypes_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS







::google::protobuf::Metadata CommandGetTableTypes::GetMetadata() const {
  return ::google::protobuf::internal::ZeroFieldsBase::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandGetPrimaryKeys::CommandGetPrimaryKeys(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetPrimaryKeys_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetPrimaryKeys)
}
PROTOBUF_NDEBUG_INLINE CommandGetPrimaryKeys::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandGetPrimaryKeys& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        catalog_(arena, from.catalog_),
        db_schema_(arena, from.db_schema_),
        table_(arena, from.table_) {}

CommandGetPrimaryKeys::CommandGetPrimaryKeys(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandGetPrimaryKeys& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetPrimaryKeys_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandGetPrimaryKeys* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandGetPrimaryKeys)
}
PROTOBUF_NDEBUG_INLINE CommandGetPrimaryKeys::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        catalog_(arena),
        db_schema_(arena),
        table_(arena) {}

inline void CommandGetPrimaryKeys::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandGetPrimaryKeys::~CommandGetPrimaryKeys() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandGetPrimaryKeys)
  SharedDtor(*this);
}
inline void CommandGetPrimaryKeys::SharedDtor(MessageLite& self) {
  CommandGetPrimaryKeys& this_ = static_cast<CommandGetPrimaryKeys&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.catalog_.Destroy();
  this_._impl_.db_schema_.Destroy();
  this_._impl_.table_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetPrimaryKeys_class_data_ =
        CommandGetPrimaryKeys::InternalGenerateClassData_(CommandGetPrimaryKeys_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetPrimaryKeys::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetPrimaryKeys_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetPrimaryKeys_class_data_.tc_table);
  return CommandGetPrimaryKeys_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetPrimaryKeys::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetPrimaryKeys_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetPrimaryKeys_globals_));
  return CommandGetPrimaryKeys_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetPrimaryKeys::ParseTableT_
    CommandGetPrimaryKeys::_table_ =
        CommandGetPrimaryKeys::InternalGenerateParseTable_(CommandGetPrimaryKeys_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandGetPrimaryKeys::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandGetPrimaryKeys)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.catalog_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.db_schema_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.table_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandGetPrimaryKeys::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandGetPrimaryKeys& this_ = static_cast<const CommandGetPrimaryKeys&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandGetPrimaryKeys::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandGetPrimaryKeys& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandGetPrimaryKeys)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string catalog = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_catalog();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetPrimaryKeys.catalog");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string db_schema = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_db_schema();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetPrimaryKeys.db_schema");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // string table = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_table().empty()) {
      const ::std::string& _s = this_._internal_table();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetPrimaryKeys.table");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandGetPrimaryKeys)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandGetPrimaryKeys::ByteSizeLong(const MessageLite& base) {
  const CommandGetPrimaryKeys& this_ = static_cast<const CommandGetPrimaryKeys&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandGetPrimaryKeys::ByteSizeLong() const {
  const CommandGetPrimaryKeys& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandGetPrimaryKeys)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // optional string catalog = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_catalog());
    }
    // optional string db_schema = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_db_schema());
    }
    // string table = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_table().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_table());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandGetPrimaryKeys::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandGetPrimaryKeys*>(&to_msg);
  auto& from = static_cast<const CommandGetPrimaryKeys&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandGetPrimaryKeys)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_catalog(from._internal_catalog());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_db_schema(from._internal_db_schema());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_table().empty()) {
        _this->_internal_set_table(from._internal_table());
      } else {
        if (_this->_impl_.table_.IsDefault()) {
          _this->_internal_set_table("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandGetPrimaryKeys::CopyFrom(const CommandGetPrimaryKeys& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandGetPrimaryKeys)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandGetPrimaryKeys::InternalSwap(CommandGetPrimaryKeys* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.catalog_, &other->_impl_.catalog_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.db_schema_, &other->_impl_.db_schema_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.table_, &other->_impl_.table_, arena);
}

::google::protobuf::Metadata CommandGetPrimaryKeys::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandGetExportedKeys::CommandGetExportedKeys(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetExportedKeys_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetExportedKeys)
}
PROTOBUF_NDEBUG_INLINE CommandGetExportedKeys::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandGetExportedKeys& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        catalog_(arena, from.catalog_),
        db_schema_(arena, from.db_schema_),
        table_(arena, from.table_) {}

CommandGetExportedKeys::CommandGetExportedKeys(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandGetExportedKeys& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetExportedKeys_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandGetExportedKeys* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandGetExportedKeys)
}
PROTOBUF_NDEBUG_INLINE CommandGetExportedKeys::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        catalog_(arena),
        db_schema_(arena),
        table_(arena) {}

inline void CommandGetExportedKeys::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandGetExportedKeys::~CommandGetExportedKeys() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandGetExportedKeys)
  SharedDtor(*this);
}
inline void CommandGetExportedKeys::SharedDtor(MessageLite& self) {
  CommandGetExportedKeys& this_ = static_cast<CommandGetExportedKeys&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.catalog_.Destroy();
  this_._impl_.db_schema_.Destroy();
  this_._impl_.table_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetExportedKeys_class_data_ =
        CommandGetExportedKeys::InternalGenerateClassData_(CommandGetExportedKeys_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetExportedKeys::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetExportedKeys_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetExportedKeys_class_data_.tc_table);
  return CommandGetExportedKeys_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetExportedKeys::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetExportedKeys_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetExportedKeys_globals_));
  return CommandGetExportedKeys_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetExportedKeys::ParseTableT_
    CommandGetExportedKeys::_table_ =
        CommandGetExportedKeys::InternalGenerateParseTable_(CommandGetExportedKeys_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandGetExportedKeys::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandGetExportedKeys)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.catalog_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.db_schema_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.table_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandGetExportedKeys::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandGetExportedKeys& this_ = static_cast<const CommandGetExportedKeys&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandGetExportedKeys::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandGetExportedKeys& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandGetExportedKeys)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string catalog = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_catalog();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetExportedKeys.catalog");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string db_schema = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_db_schema();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetExportedKeys.db_schema");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // string table = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_table().empty()) {
      const ::std::string& _s = this_._internal_table();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetExportedKeys.table");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandGetExportedKeys)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandGetExportedKeys::ByteSizeLong(const MessageLite& base) {
  const CommandGetExportedKeys& this_ = static_cast<const CommandGetExportedKeys&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandGetExportedKeys::ByteSizeLong() const {
  const CommandGetExportedKeys& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandGetExportedKeys)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // optional string catalog = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_catalog());
    }
    // optional string db_schema = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_db_schema());
    }
    // string table = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_table().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_table());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandGetExportedKeys::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandGetExportedKeys*>(&to_msg);
  auto& from = static_cast<const CommandGetExportedKeys&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandGetExportedKeys)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_catalog(from._internal_catalog());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_db_schema(from._internal_db_schema());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_table().empty()) {
        _this->_internal_set_table(from._internal_table());
      } else {
        if (_this->_impl_.table_.IsDefault()) {
          _this->_internal_set_table("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandGetExportedKeys::CopyFrom(const CommandGetExportedKeys& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandGetExportedKeys)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandGetExportedKeys::InternalSwap(CommandGetExportedKeys* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.catalog_, &other->_impl_.catalog_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.db_schema_, &other->_impl_.db_schema_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.table_, &other->_impl_.table_, arena);
}

::google::protobuf::Metadata CommandGetExportedKeys::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandGetImportedKeys::CommandGetImportedKeys(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetImportedKeys_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetImportedKeys)
}
PROTOBUF_NDEBUG_INLINE CommandGetImportedKeys::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandGetImportedKeys& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        catalog_(arena, from.catalog_),
        db_schema_(arena, from.db_schema_),
        table_(arena, from.table_) {}

CommandGetImportedKeys::CommandGetImportedKeys(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandGetImportedKeys& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetImportedKeys_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandGetImportedKeys* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandGetImportedKeys)
}
PROTOBUF_NDEBUG_INLINE CommandGetImportedKeys::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        catalog_(arena),
        db_schema_(arena),
        table_(arena) {}

inline void CommandGetImportedKeys::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandGetImportedKeys::~CommandGetImportedKeys() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandGetImportedKeys)
  SharedDtor(*this);
}
inline void CommandGetImportedKeys::SharedDtor(MessageLite& self) {
  CommandGetImportedKeys& this_ = static_cast<CommandGetImportedKeys&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.catalog_.Destroy();
  this_._impl_.db_schema_.Destroy();
  this_._impl_.table_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetImportedKeys_class_data_ =
        CommandGetImportedKeys::InternalGenerateClassData_(CommandGetImportedKeys_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetImportedKeys::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetImportedKeys_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetImportedKeys_class_data_.tc_table);
  return CommandGetImportedKeys_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetImportedKeys::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetImportedKeys_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetImportedKeys_globals_));
  return CommandGetImportedKeys_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetImportedKeys::ParseTableT_
    CommandGetImportedKeys::_table_ =
        CommandGetImportedKeys::InternalGenerateParseTable_(CommandGetImportedKeys_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandGetImportedKeys::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandGetImportedKeys)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.catalog_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.db_schema_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.table_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandGetImportedKeys::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandGetImportedKeys& this_ = static_cast<const CommandGetImportedKeys&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandGetImportedKeys::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandGetImportedKeys& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandGetImportedKeys)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string catalog = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_catalog();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetImportedKeys.catalog");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string db_schema = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_db_schema();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetImportedKeys.db_schema");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // string table = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_table().empty()) {
      const ::std::string& _s = this_._internal_table();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetImportedKeys.table");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandGetImportedKeys)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandGetImportedKeys::ByteSizeLong(const MessageLite& base) {
  const CommandGetImportedKeys& this_ = static_cast<const CommandGetImportedKeys&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandGetImportedKeys::ByteSizeLong() const {
  const CommandGetImportedKeys& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandGetImportedKeys)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // optional string catalog = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_catalog());
    }
    // optional string db_schema = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_db_schema());
    }
    // string table = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_table().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_table());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandGetImportedKeys::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandGetImportedKeys*>(&to_msg);
  auto& from = static_cast<const CommandGetImportedKeys&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandGetImportedKeys)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_catalog(from._internal_catalog());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_db_schema(from._internal_db_schema());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_table().empty()) {
        _this->_internal_set_table(from._internal_table());
      } else {
        if (_this->_impl_.table_.IsDefault()) {
          _this->_internal_set_table("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandGetImportedKeys::CopyFrom(const CommandGetImportedKeys& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandGetImportedKeys)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandGetImportedKeys::InternalSwap(CommandGetImportedKeys* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.catalog_, &other->_impl_.catalog_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.db_schema_, &other->_impl_.db_schema_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.table_, &other->_impl_.table_, arena);
}

::google::protobuf::Metadata CommandGetImportedKeys::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandGetCrossReference::CommandGetCrossReference(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetCrossReference_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandGetCrossReference)
}
PROTOBUF_NDEBUG_INLINE CommandGetCrossReference::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandGetCrossReference& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        pk_catalog_(arena, from.pk_catalog_),
        pk_db_schema_(arena, from.pk_db_schema_),
        pk_table_(arena, from.pk_table_),
        fk_catalog_(arena, from.fk_catalog_),
        fk_db_schema_(arena, from.fk_db_schema_),
        fk_table_(arena, from.fk_table_) {}

CommandGetCrossReference::CommandGetCrossReference(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandGetCrossReference& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandGetCrossReference_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandGetCrossReference* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandGetCrossReference)
}
PROTOBUF_NDEBUG_INLINE CommandGetCrossReference::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        pk_catalog_(arena),
        pk_db_schema_(arena),
        pk_table_(arena),
        fk_catalog_(arena),
        fk_db_schema_(arena),
        fk_table_(arena) {}

inline void CommandGetCrossReference::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandGetCrossReference::~CommandGetCrossReference() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandGetCrossReference)
  SharedDtor(*this);
}
inline void CommandGetCrossReference::SharedDtor(MessageLite& self) {
  CommandGetCrossReference& this_ = static_cast<CommandGetCrossReference&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.pk_catalog_.Destroy();
  this_._impl_.pk_db_schema_.Destroy();
  this_._impl_.pk_table_.Destroy();
  this_._impl_.fk_catalog_.Destroy();
  this_._impl_.fk_db_schema_.Destroy();
  this_._impl_.fk_table_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandGetCrossReference_class_data_ =
        CommandGetCrossReference::InternalGenerateClassData_(CommandGetCrossReference_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetCrossReference::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetCrossReference_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandGetCrossReference_class_data_.tc_table);
  return CommandGetCrossReference_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandGetCrossReference::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandGetCrossReference_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandGetCrossReference_globals_));
  return CommandGetCrossReference_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandGetCrossReference::ParseTableT_
    CommandGetCrossReference::_table_ =
        CommandGetCrossReference::InternalGenerateParseTable_(CommandGetCrossReference_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandGetCrossReference::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandGetCrossReference)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.pk_catalog_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.pk_db_schema_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.pk_table_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.fk_catalog_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _impl_.fk_db_schema_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      _impl_.fk_table_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandGetCrossReference::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandGetCrossReference& this_ = static_cast<const CommandGetCrossReference&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandGetCrossReference::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandGetCrossReference& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandGetCrossReference)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string pk_catalog = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_pk_catalog();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetCrossReference.pk_catalog");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string pk_db_schema = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_pk_db_schema();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetCrossReference.pk_db_schema");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // string pk_table = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_pk_table().empty()) {
      const ::std::string& _s = this_._internal_pk_table();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetCrossReference.pk_table");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  // optional string fk_catalog = 4;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    const ::std::string& _s = this_._internal_fk_catalog();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetCrossReference.fk_catalog");
    target = stream->WriteStringMaybeAliased(4, _s, target);
  }

  // optional string fk_db_schema = 5;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    const ::std::string& _s = this_._internal_fk_db_schema();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetCrossReference.fk_db_schema");
    target = stream->WriteStringMaybeAliased(5, _s, target);
  }

  // string fk_table = 6;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (!this_._internal_fk_table().empty()) {
      const ::std::string& _s = this_._internal_fk_table();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandGetCrossReference.fk_table");
      target = stream->WriteStringMaybeAliased(6, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandGetCrossReference)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandGetCrossReference::ByteSizeLong(const MessageLite& base) {
  const CommandGetCrossReference& this_ = static_cast<const CommandGetCrossReference&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandGetCrossReference::ByteSizeLong() const {
  const CommandGetCrossReference& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandGetCrossReference)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    // optional string pk_catalog = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_pk_catalog());
    }
    // optional string pk_db_schema = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_pk_db_schema());
    }
    // string pk_table = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_pk_table().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_pk_table());
      }
    }
    // optional string fk_catalog = 4;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_fk_catalog());
    }
    // optional string fk_db_schema = 5;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_fk_db_schema());
    }
    // string fk_table = 6;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (!this_._internal_fk_table().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_fk_table());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandGetCrossReference::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandGetCrossReference*>(&to_msg);
  auto& from = static_cast<const CommandGetCrossReference&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandGetCrossReference)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_pk_catalog(from._internal_pk_catalog());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_pk_db_schema(from._internal_pk_db_schema());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_pk_table().empty()) {
        _this->_internal_set_pk_table(from._internal_pk_table());
      } else {
        if (_this->_impl_.pk_table_.IsDefault()) {
          _this->_internal_set_pk_table("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_internal_set_fk_catalog(from._internal_fk_catalog());
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _this->_internal_set_fk_db_schema(from._internal_fk_db_schema());
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (!from._internal_fk_table().empty()) {
        _this->_internal_set_fk_table(from._internal_fk_table());
      } else {
        if (_this->_impl_.fk_table_.IsDefault()) {
          _this->_internal_set_fk_table("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandGetCrossReference::CopyFrom(const CommandGetCrossReference& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandGetCrossReference)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandGetCrossReference::InternalSwap(CommandGetCrossReference* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.pk_catalog_, &other->_impl_.pk_catalog_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.pk_db_schema_, &other->_impl_.pk_db_schema_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.pk_table_, &other->_impl_.pk_table_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.fk_catalog_, &other->_impl_.fk_catalog_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.fk_db_schema_, &other->_impl_.fk_db_schema_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.fk_table_, &other->_impl_.fk_table_, arena);
}

::google::protobuf::Metadata CommandGetCrossReference::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionCreatePreparedStatementRequest::ActionCreatePreparedStatementRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCreatePreparedStatementRequest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest)
}
PROTOBUF_NDEBUG_INLINE ActionCreatePreparedStatementRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionCreatePreparedStatementRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        query_(arena, from.query_),
        transaction_id_(arena, from.transaction_id_) {}

ActionCreatePreparedStatementRequest::ActionCreatePreparedStatementRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionCreatePreparedStatementRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCreatePreparedStatementRequest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionCreatePreparedStatementRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest)
}
PROTOBUF_NDEBUG_INLINE ActionCreatePreparedStatementRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        query_(arena),
        transaction_id_(arena) {}

inline void ActionCreatePreparedStatementRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
ActionCreatePreparedStatementRequest::~ActionCreatePreparedStatementRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest)
  SharedDtor(*this);
}
inline void ActionCreatePreparedStatementRequest::SharedDtor(MessageLite& self) {
  ActionCreatePreparedStatementRequest& this_ = static_cast<ActionCreatePreparedStatementRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.query_.Destroy();
  this_._impl_.transaction_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionCreatePreparedStatementRequest_class_data_ =
        ActionCreatePreparedStatementRequest::InternalGenerateClassData_(ActionCreatePreparedStatementRequest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCreatePreparedStatementRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCreatePreparedStatementRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionCreatePreparedStatementRequest_class_data_.tc_table);
  return ActionCreatePreparedStatementRequest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCreatePreparedStatementRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCreatePreparedStatementRequest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionCreatePreparedStatementRequest_globals_));
  return ActionCreatePreparedStatementRequest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionCreatePreparedStatementRequest::ParseTableT_
    ActionCreatePreparedStatementRequest::_table_ =
        ActionCreatePreparedStatementRequest::InternalGenerateParseTable_(ActionCreatePreparedStatementRequest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionCreatePreparedStatementRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.query_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.transaction_id_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionCreatePreparedStatementRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionCreatePreparedStatementRequest& this_ = static_cast<const ActionCreatePreparedStatementRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionCreatePreparedStatementRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionCreatePreparedStatementRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string query = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_query().empty()) {
      const ::std::string& _s = this_._internal_query();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest.query");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // optional bytes transaction_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_transaction_id();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionCreatePreparedStatementRequest::ByteSizeLong(const MessageLite& base) {
  const ActionCreatePreparedStatementRequest& this_ = static_cast<const ActionCreatePreparedStatementRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionCreatePreparedStatementRequest::ByteSizeLong() const {
  const ActionCreatePreparedStatementRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string query = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_query().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_query());
      }
    }
    // optional bytes transaction_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_transaction_id());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionCreatePreparedStatementRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionCreatePreparedStatementRequest*>(&to_msg);
  auto& from = static_cast<const ActionCreatePreparedStatementRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_query().empty()) {
        _this->_internal_set_query(from._internal_query());
      } else {
        if (_this->_impl_.query_.IsDefault()) {
          _this->_internal_set_query("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_transaction_id(from._internal_transaction_id());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionCreatePreparedStatementRequest::CopyFrom(const ActionCreatePreparedStatementRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionCreatePreparedStatementRequest::InternalSwap(ActionCreatePreparedStatementRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.query_, &other->_impl_.query_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.transaction_id_, &other->_impl_.transaction_id_, arena);
}

::google::protobuf::Metadata ActionCreatePreparedStatementRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

SubstraitPlan::SubstraitPlan(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SubstraitPlan_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.SubstraitPlan)
}
PROTOBUF_NDEBUG_INLINE SubstraitPlan::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::SubstraitPlan& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        plan_(arena, from.plan_),
        version_(arena, from.version_) {}

SubstraitPlan::SubstraitPlan(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const SubstraitPlan& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, SubstraitPlan_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SubstraitPlan* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.SubstraitPlan)
}
PROTOBUF_NDEBUG_INLINE SubstraitPlan::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        plan_(arena),
        version_(arena) {}

inline void SubstraitPlan::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
SubstraitPlan::~SubstraitPlan() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.SubstraitPlan)
  SharedDtor(*this);
}
inline void SubstraitPlan::SharedDtor(MessageLite& self) {
  SubstraitPlan& this_ = static_cast<SubstraitPlan&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.plan_.Destroy();
  this_._impl_.version_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull SubstraitPlan_class_data_ =
        SubstraitPlan::InternalGenerateClassData_(SubstraitPlan_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SubstraitPlan::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SubstraitPlan_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(SubstraitPlan_class_data_.tc_table);
  return SubstraitPlan_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
SubstraitPlan::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&SubstraitPlan_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&SubstraitPlan_globals_));
  return SubstraitPlan_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const SubstraitPlan::ParseTableT_
    SubstraitPlan::_table_ =
        SubstraitPlan::InternalGenerateParseTable_(SubstraitPlan_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void SubstraitPlan::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.SubstraitPlan)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.plan_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.version_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL SubstraitPlan::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const SubstraitPlan& this_ = static_cast<const SubstraitPlan&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL SubstraitPlan::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const SubstraitPlan& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.SubstraitPlan)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes plan = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_plan().empty()) {
      const ::std::string& _s = this_._internal_plan();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  // string version = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_version().empty()) {
      const ::std::string& _s = this_._internal_version();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.SubstraitPlan.version");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.SubstraitPlan)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t SubstraitPlan::ByteSizeLong(const MessageLite& base) {
  const SubstraitPlan& this_ = static_cast<const SubstraitPlan&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t SubstraitPlan::ByteSizeLong() const {
  const SubstraitPlan& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.SubstraitPlan)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // bytes plan = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_plan().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_plan());
      }
    }
    // string version = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_version().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_version());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void SubstraitPlan::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<SubstraitPlan*>(&to_msg);
  auto& from = static_cast<const SubstraitPlan&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.SubstraitPlan)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_plan().empty()) {
        _this->_internal_set_plan(from._internal_plan());
      } else {
        if (_this->_impl_.plan_.IsDefault()) {
          _this->_internal_set_plan("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_version().empty()) {
        _this->_internal_set_version(from._internal_version());
      } else {
        if (_this->_impl_.version_.IsDefault()) {
          _this->_internal_set_version("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void SubstraitPlan::CopyFrom(const SubstraitPlan& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.SubstraitPlan)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void SubstraitPlan::InternalSwap(SubstraitPlan* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.plan_, &other->_impl_.plan_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.version_, &other->_impl_.version_, arena);
}

::google::protobuf::Metadata SubstraitPlan::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionCreatePreparedSubstraitPlanRequest::ActionCreatePreparedSubstraitPlanRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCreatePreparedSubstraitPlanRequest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionCreatePreparedSubstraitPlanRequest)
}
PROTOBUF_NDEBUG_INLINE ActionCreatePreparedSubstraitPlanRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionCreatePreparedSubstraitPlanRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        transaction_id_(arena, from.transaction_id_) {}

ActionCreatePreparedSubstraitPlanRequest::ActionCreatePreparedSubstraitPlanRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionCreatePreparedSubstraitPlanRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCreatePreparedSubstraitPlanRequest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionCreatePreparedSubstraitPlanRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.plan_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.plan_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionCreatePreparedSubstraitPlanRequest)
}
PROTOBUF_NDEBUG_INLINE ActionCreatePreparedSubstraitPlanRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        transaction_id_(arena) {}

inline void ActionCreatePreparedSubstraitPlanRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.plan_ = {};
}
ActionCreatePreparedSubstraitPlanRequest::~ActionCreatePreparedSubstraitPlanRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionCreatePreparedSubstraitPlanRequest)
  SharedDtor(*this);
}
inline void ActionCreatePreparedSubstraitPlanRequest::SharedDtor(MessageLite& self) {
  ActionCreatePreparedSubstraitPlanRequest& this_ = static_cast<ActionCreatePreparedSubstraitPlanRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.transaction_id_.Destroy();
  delete this_._impl_.plan_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionCreatePreparedSubstraitPlanRequest_class_data_ =
        ActionCreatePreparedSubstraitPlanRequest::InternalGenerateClassData_(ActionCreatePreparedSubstraitPlanRequest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCreatePreparedSubstraitPlanRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCreatePreparedSubstraitPlanRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionCreatePreparedSubstraitPlanRequest_class_data_.tc_table);
  return ActionCreatePreparedSubstraitPlanRequest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCreatePreparedSubstraitPlanRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCreatePreparedSubstraitPlanRequest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionCreatePreparedSubstraitPlanRequest_globals_));
  return ActionCreatePreparedSubstraitPlanRequest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionCreatePreparedSubstraitPlanRequest::ParseTableT_
    ActionCreatePreparedSubstraitPlanRequest::_table_ =
        ActionCreatePreparedSubstraitPlanRequest::InternalGenerateParseTable_(ActionCreatePreparedSubstraitPlanRequest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionCreatePreparedSubstraitPlanRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionCreatePreparedSubstraitPlanRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.transaction_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.plan_ != nullptr);
      _impl_.plan_->Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionCreatePreparedSubstraitPlanRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionCreatePreparedSubstraitPlanRequest& this_ = static_cast<const ActionCreatePreparedSubstraitPlanRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionCreatePreparedSubstraitPlanRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionCreatePreparedSubstraitPlanRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionCreatePreparedSubstraitPlanRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.sql.SubstraitPlan plan = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.plan_, this_._impl_.plan_->GetCachedSize(), target,
        stream);
  }

  // optional bytes transaction_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_transaction_id();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionCreatePreparedSubstraitPlanRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionCreatePreparedSubstraitPlanRequest::ByteSizeLong(const MessageLite& base) {
  const ActionCreatePreparedSubstraitPlanRequest& this_ = static_cast<const ActionCreatePreparedSubstraitPlanRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionCreatePreparedSubstraitPlanRequest::ByteSizeLong() const {
  const ActionCreatePreparedSubstraitPlanRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionCreatePreparedSubstraitPlanRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // optional bytes transaction_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_transaction_id());
    }
    // .arrow.flight.protocol.sql.SubstraitPlan plan = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.plan_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionCreatePreparedSubstraitPlanRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionCreatePreparedSubstraitPlanRequest*>(&to_msg);
  auto& from = static_cast<const ActionCreatePreparedSubstraitPlanRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionCreatePreparedSubstraitPlanRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_transaction_id(from._internal_transaction_id());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.plan_ != nullptr);
      if (_this->_impl_.plan_ == nullptr) {
        _this->_impl_.plan_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.plan_);
      } else {
        _this->_impl_.plan_->MergeFrom(*from._impl_.plan_);
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionCreatePreparedSubstraitPlanRequest::CopyFrom(const ActionCreatePreparedSubstraitPlanRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionCreatePreparedSubstraitPlanRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionCreatePreparedSubstraitPlanRequest::InternalSwap(ActionCreatePreparedSubstraitPlanRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.transaction_id_, &other->_impl_.transaction_id_, arena);
  swap(_impl_.plan_, other->_impl_.plan_);
}

::google::protobuf::Metadata ActionCreatePreparedSubstraitPlanRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionCreatePreparedStatementResult::ActionCreatePreparedStatementResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCreatePreparedStatementResult_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionCreatePreparedStatementResult)
}
PROTOBUF_NDEBUG_INLINE ActionCreatePreparedStatementResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionCreatePreparedStatementResult& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        prepared_statement_handle_(arena, from.prepared_statement_handle_),
        dataset_schema_(arena, from.dataset_schema_),
        parameter_schema_(arena, from.parameter_schema_) {}

ActionCreatePreparedStatementResult::ActionCreatePreparedStatementResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionCreatePreparedStatementResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCreatePreparedStatementResult_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionCreatePreparedStatementResult* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionCreatePreparedStatementResult)
}
PROTOBUF_NDEBUG_INLINE ActionCreatePreparedStatementResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        prepared_statement_handle_(arena),
        dataset_schema_(arena),
        parameter_schema_(arena) {}

inline void ActionCreatePreparedStatementResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
ActionCreatePreparedStatementResult::~ActionCreatePreparedStatementResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionCreatePreparedStatementResult)
  SharedDtor(*this);
}
inline void ActionCreatePreparedStatementResult::SharedDtor(MessageLite& self) {
  ActionCreatePreparedStatementResult& this_ = static_cast<ActionCreatePreparedStatementResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.prepared_statement_handle_.Destroy();
  this_._impl_.dataset_schema_.Destroy();
  this_._impl_.parameter_schema_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionCreatePreparedStatementResult_class_data_ =
        ActionCreatePreparedStatementResult::InternalGenerateClassData_(ActionCreatePreparedStatementResult_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCreatePreparedStatementResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCreatePreparedStatementResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionCreatePreparedStatementResult_class_data_.tc_table);
  return ActionCreatePreparedStatementResult_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCreatePreparedStatementResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCreatePreparedStatementResult_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionCreatePreparedStatementResult_globals_));
  return ActionCreatePreparedStatementResult_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionCreatePreparedStatementResult::ParseTableT_
    ActionCreatePreparedStatementResult::_table_ =
        ActionCreatePreparedStatementResult::InternalGenerateParseTable_(ActionCreatePreparedStatementResult_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionCreatePreparedStatementResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.prepared_statement_handle_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.dataset_schema_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.parameter_schema_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionCreatePreparedStatementResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionCreatePreparedStatementResult& this_ = static_cast<const ActionCreatePreparedStatementResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionCreatePreparedStatementResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionCreatePreparedStatementResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes prepared_statement_handle = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_prepared_statement_handle().empty()) {
      const ::std::string& _s = this_._internal_prepared_statement_handle();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  // bytes dataset_schema = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_dataset_schema().empty()) {
      const ::std::string& _s = this_._internal_dataset_schema();
      target = stream->WriteBytesMaybeAliased(2, _s, target);
    }
  }

  // bytes parameter_schema = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_parameter_schema().empty()) {
      const ::std::string& _s = this_._internal_parameter_schema();
      target = stream->WriteBytesMaybeAliased(3, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionCreatePreparedStatementResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionCreatePreparedStatementResult::ByteSizeLong(const MessageLite& base) {
  const ActionCreatePreparedStatementResult& this_ = static_cast<const ActionCreatePreparedStatementResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionCreatePreparedStatementResult::ByteSizeLong() const {
  const ActionCreatePreparedStatementResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // bytes prepared_statement_handle = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_prepared_statement_handle().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_prepared_statement_handle());
      }
    }
    // bytes dataset_schema = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_dataset_schema().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_dataset_schema());
      }
    }
    // bytes parameter_schema = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_parameter_schema().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_parameter_schema());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionCreatePreparedStatementResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionCreatePreparedStatementResult*>(&to_msg);
  auto& from = static_cast<const ActionCreatePreparedStatementResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_prepared_statement_handle().empty()) {
        _this->_internal_set_prepared_statement_handle(from._internal_prepared_statement_handle());
      } else {
        if (_this->_impl_.prepared_statement_handle_.IsDefault()) {
          _this->_internal_set_prepared_statement_handle("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_dataset_schema().empty()) {
        _this->_internal_set_dataset_schema(from._internal_dataset_schema());
      } else {
        if (_this->_impl_.dataset_schema_.IsDefault()) {
          _this->_internal_set_dataset_schema("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_parameter_schema().empty()) {
        _this->_internal_set_parameter_schema(from._internal_parameter_schema());
      } else {
        if (_this->_impl_.parameter_schema_.IsDefault()) {
          _this->_internal_set_parameter_schema("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionCreatePreparedStatementResult::CopyFrom(const ActionCreatePreparedStatementResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionCreatePreparedStatementResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionCreatePreparedStatementResult::InternalSwap(ActionCreatePreparedStatementResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.prepared_statement_handle_, &other->_impl_.prepared_statement_handle_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.dataset_schema_, &other->_impl_.dataset_schema_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.parameter_schema_, &other->_impl_.parameter_schema_, arena);
}

::google::protobuf::Metadata ActionCreatePreparedStatementResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionClosePreparedStatementRequest::ActionClosePreparedStatementRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionClosePreparedStatementRequest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionClosePreparedStatementRequest)
}
PROTOBUF_NDEBUG_INLINE ActionClosePreparedStatementRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionClosePreparedStatementRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        prepared_statement_handle_(arena, from.prepared_statement_handle_) {}

ActionClosePreparedStatementRequest::ActionClosePreparedStatementRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionClosePreparedStatementRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionClosePreparedStatementRequest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionClosePreparedStatementRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionClosePreparedStatementRequest)
}
PROTOBUF_NDEBUG_INLINE ActionClosePreparedStatementRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        prepared_statement_handle_(arena) {}

inline void ActionClosePreparedStatementRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
ActionClosePreparedStatementRequest::~ActionClosePreparedStatementRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionClosePreparedStatementRequest)
  SharedDtor(*this);
}
inline void ActionClosePreparedStatementRequest::SharedDtor(MessageLite& self) {
  ActionClosePreparedStatementRequest& this_ = static_cast<ActionClosePreparedStatementRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.prepared_statement_handle_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionClosePreparedStatementRequest_class_data_ =
        ActionClosePreparedStatementRequest::InternalGenerateClassData_(ActionClosePreparedStatementRequest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionClosePreparedStatementRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionClosePreparedStatementRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionClosePreparedStatementRequest_class_data_.tc_table);
  return ActionClosePreparedStatementRequest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionClosePreparedStatementRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionClosePreparedStatementRequest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionClosePreparedStatementRequest_globals_));
  return ActionClosePreparedStatementRequest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionClosePreparedStatementRequest::ParseTableT_
    ActionClosePreparedStatementRequest::_table_ =
        ActionClosePreparedStatementRequest::InternalGenerateParseTable_(ActionClosePreparedStatementRequest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionClosePreparedStatementRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionClosePreparedStatementRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.prepared_statement_handle_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionClosePreparedStatementRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionClosePreparedStatementRequest& this_ = static_cast<const ActionClosePreparedStatementRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionClosePreparedStatementRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionClosePreparedStatementRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionClosePreparedStatementRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes prepared_statement_handle = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_prepared_statement_handle().empty()) {
      const ::std::string& _s = this_._internal_prepared_statement_handle();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionClosePreparedStatementRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionClosePreparedStatementRequest::ByteSizeLong(const MessageLite& base) {
  const ActionClosePreparedStatementRequest& this_ = static_cast<const ActionClosePreparedStatementRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionClosePreparedStatementRequest::ByteSizeLong() const {
  const ActionClosePreparedStatementRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionClosePreparedStatementRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes prepared_statement_handle = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_prepared_statement_handle().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_prepared_statement_handle());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionClosePreparedStatementRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionClosePreparedStatementRequest*>(&to_msg);
  auto& from = static_cast<const ActionClosePreparedStatementRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionClosePreparedStatementRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_prepared_statement_handle().empty()) {
      _this->_internal_set_prepared_statement_handle(from._internal_prepared_statement_handle());
    } else {
      if (_this->_impl_.prepared_statement_handle_.IsDefault()) {
        _this->_internal_set_prepared_statement_handle("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionClosePreparedStatementRequest::CopyFrom(const ActionClosePreparedStatementRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionClosePreparedStatementRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionClosePreparedStatementRequest::InternalSwap(ActionClosePreparedStatementRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.prepared_statement_handle_, &other->_impl_.prepared_statement_handle_, arena);
}

::google::protobuf::Metadata ActionClosePreparedStatementRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionBeginTransactionRequest::ActionBeginTransactionRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, ActionBeginTransactionRequest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionBeginTransactionRequest)
}
ActionBeginTransactionRequest::ActionBeginTransactionRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionBeginTransactionRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::internal::ZeroFieldsBase(arena, ActionBeginTransactionRequest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::internal::ZeroFieldsBase(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionBeginTransactionRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionBeginTransactionRequest)
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionBeginTransactionRequest_class_data_ =
        ActionBeginTransactionRequest::InternalGenerateClassData_(ActionBeginTransactionRequest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionBeginTransactionRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionBeginTransactionRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionBeginTransactionRequest_class_data_.tc_table);
  return ActionBeginTransactionRequest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionBeginTransactionRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionBeginTransactionRequest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionBeginTransactionRequest_globals_));
  return ActionBeginTransactionRequest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionBeginTransactionRequest::ParseTableT_
    ActionBeginTransactionRequest::_table_ =
        ActionBeginTransactionRequest::InternalGenerateParseTable_(ActionBeginTransactionRequest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS







::google::protobuf::Metadata ActionBeginTransactionRequest::GetMetadata() const {
  return ::google::protobuf::internal::ZeroFieldsBase::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionBeginSavepointRequest::ActionBeginSavepointRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionBeginSavepointRequest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionBeginSavepointRequest)
}
PROTOBUF_NDEBUG_INLINE ActionBeginSavepointRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionBeginSavepointRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        transaction_id_(arena, from.transaction_id_),
        name_(arena, from.name_) {}

ActionBeginSavepointRequest::ActionBeginSavepointRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionBeginSavepointRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionBeginSavepointRequest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionBeginSavepointRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionBeginSavepointRequest)
}
PROTOBUF_NDEBUG_INLINE ActionBeginSavepointRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        transaction_id_(arena),
        name_(arena) {}

inline void ActionBeginSavepointRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
ActionBeginSavepointRequest::~ActionBeginSavepointRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionBeginSavepointRequest)
  SharedDtor(*this);
}
inline void ActionBeginSavepointRequest::SharedDtor(MessageLite& self) {
  ActionBeginSavepointRequest& this_ = static_cast<ActionBeginSavepointRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.transaction_id_.Destroy();
  this_._impl_.name_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionBeginSavepointRequest_class_data_ =
        ActionBeginSavepointRequest::InternalGenerateClassData_(ActionBeginSavepointRequest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionBeginSavepointRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionBeginSavepointRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionBeginSavepointRequest_class_data_.tc_table);
  return ActionBeginSavepointRequest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionBeginSavepointRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionBeginSavepointRequest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionBeginSavepointRequest_globals_));
  return ActionBeginSavepointRequest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionBeginSavepointRequest::ParseTableT_
    ActionBeginSavepointRequest::_table_ =
        ActionBeginSavepointRequest::InternalGenerateParseTable_(ActionBeginSavepointRequest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionBeginSavepointRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionBeginSavepointRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.transaction_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.name_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionBeginSavepointRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionBeginSavepointRequest& this_ = static_cast<const ActionBeginSavepointRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionBeginSavepointRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionBeginSavepointRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionBeginSavepointRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes transaction_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_transaction_id().empty()) {
      const ::std::string& _s = this_._internal_transaction_id();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  // string name = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_name().empty()) {
      const ::std::string& _s = this_._internal_name();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.ActionBeginSavepointRequest.name");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionBeginSavepointRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionBeginSavepointRequest::ByteSizeLong(const MessageLite& base) {
  const ActionBeginSavepointRequest& this_ = static_cast<const ActionBeginSavepointRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionBeginSavepointRequest::ByteSizeLong() const {
  const ActionBeginSavepointRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionBeginSavepointRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // bytes transaction_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_transaction_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_transaction_id());
      }
    }
    // string name = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_name().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_name());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionBeginSavepointRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionBeginSavepointRequest*>(&to_msg);
  auto& from = static_cast<const ActionBeginSavepointRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionBeginSavepointRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_transaction_id().empty()) {
        _this->_internal_set_transaction_id(from._internal_transaction_id());
      } else {
        if (_this->_impl_.transaction_id_.IsDefault()) {
          _this->_internal_set_transaction_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_name().empty()) {
        _this->_internal_set_name(from._internal_name());
      } else {
        if (_this->_impl_.name_.IsDefault()) {
          _this->_internal_set_name("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionBeginSavepointRequest::CopyFrom(const ActionBeginSavepointRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionBeginSavepointRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionBeginSavepointRequest::InternalSwap(ActionBeginSavepointRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.transaction_id_, &other->_impl_.transaction_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.name_, &other->_impl_.name_, arena);
}

::google::protobuf::Metadata ActionBeginSavepointRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionBeginTransactionResult::ActionBeginTransactionResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionBeginTransactionResult_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionBeginTransactionResult)
}
PROTOBUF_NDEBUG_INLINE ActionBeginTransactionResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionBeginTransactionResult& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        transaction_id_(arena, from.transaction_id_) {}

ActionBeginTransactionResult::ActionBeginTransactionResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionBeginTransactionResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionBeginTransactionResult_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionBeginTransactionResult* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionBeginTransactionResult)
}
PROTOBUF_NDEBUG_INLINE ActionBeginTransactionResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        transaction_id_(arena) {}

inline void ActionBeginTransactionResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
ActionBeginTransactionResult::~ActionBeginTransactionResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionBeginTransactionResult)
  SharedDtor(*this);
}
inline void ActionBeginTransactionResult::SharedDtor(MessageLite& self) {
  ActionBeginTransactionResult& this_ = static_cast<ActionBeginTransactionResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.transaction_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionBeginTransactionResult_class_data_ =
        ActionBeginTransactionResult::InternalGenerateClassData_(ActionBeginTransactionResult_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionBeginTransactionResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionBeginTransactionResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionBeginTransactionResult_class_data_.tc_table);
  return ActionBeginTransactionResult_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionBeginTransactionResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionBeginTransactionResult_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionBeginTransactionResult_globals_));
  return ActionBeginTransactionResult_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionBeginTransactionResult::ParseTableT_
    ActionBeginTransactionResult::_table_ =
        ActionBeginTransactionResult::InternalGenerateParseTable_(ActionBeginTransactionResult_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionBeginTransactionResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionBeginTransactionResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.transaction_id_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionBeginTransactionResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionBeginTransactionResult& this_ = static_cast<const ActionBeginTransactionResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionBeginTransactionResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionBeginTransactionResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionBeginTransactionResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes transaction_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_transaction_id().empty()) {
      const ::std::string& _s = this_._internal_transaction_id();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionBeginTransactionResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionBeginTransactionResult::ByteSizeLong(const MessageLite& base) {
  const ActionBeginTransactionResult& this_ = static_cast<const ActionBeginTransactionResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionBeginTransactionResult::ByteSizeLong() const {
  const ActionBeginTransactionResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionBeginTransactionResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes transaction_id = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_transaction_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_transaction_id());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionBeginTransactionResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionBeginTransactionResult*>(&to_msg);
  auto& from = static_cast<const ActionBeginTransactionResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionBeginTransactionResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_transaction_id().empty()) {
      _this->_internal_set_transaction_id(from._internal_transaction_id());
    } else {
      if (_this->_impl_.transaction_id_.IsDefault()) {
        _this->_internal_set_transaction_id("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionBeginTransactionResult::CopyFrom(const ActionBeginTransactionResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionBeginTransactionResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionBeginTransactionResult::InternalSwap(ActionBeginTransactionResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.transaction_id_, &other->_impl_.transaction_id_, arena);
}

::google::protobuf::Metadata ActionBeginTransactionResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionBeginSavepointResult::ActionBeginSavepointResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionBeginSavepointResult_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionBeginSavepointResult)
}
PROTOBUF_NDEBUG_INLINE ActionBeginSavepointResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionBeginSavepointResult& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        savepoint_id_(arena, from.savepoint_id_) {}

ActionBeginSavepointResult::ActionBeginSavepointResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionBeginSavepointResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionBeginSavepointResult_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionBeginSavepointResult* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionBeginSavepointResult)
}
PROTOBUF_NDEBUG_INLINE ActionBeginSavepointResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        savepoint_id_(arena) {}

inline void ActionBeginSavepointResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
ActionBeginSavepointResult::~ActionBeginSavepointResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionBeginSavepointResult)
  SharedDtor(*this);
}
inline void ActionBeginSavepointResult::SharedDtor(MessageLite& self) {
  ActionBeginSavepointResult& this_ = static_cast<ActionBeginSavepointResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.savepoint_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionBeginSavepointResult_class_data_ =
        ActionBeginSavepointResult::InternalGenerateClassData_(ActionBeginSavepointResult_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionBeginSavepointResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionBeginSavepointResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionBeginSavepointResult_class_data_.tc_table);
  return ActionBeginSavepointResult_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionBeginSavepointResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionBeginSavepointResult_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionBeginSavepointResult_globals_));
  return ActionBeginSavepointResult_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionBeginSavepointResult::ParseTableT_
    ActionBeginSavepointResult::_table_ =
        ActionBeginSavepointResult::InternalGenerateParseTable_(ActionBeginSavepointResult_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionBeginSavepointResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionBeginSavepointResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.savepoint_id_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionBeginSavepointResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionBeginSavepointResult& this_ = static_cast<const ActionBeginSavepointResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionBeginSavepointResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionBeginSavepointResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionBeginSavepointResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes savepoint_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_savepoint_id().empty()) {
      const ::std::string& _s = this_._internal_savepoint_id();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionBeginSavepointResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionBeginSavepointResult::ByteSizeLong(const MessageLite& base) {
  const ActionBeginSavepointResult& this_ = static_cast<const ActionBeginSavepointResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionBeginSavepointResult::ByteSizeLong() const {
  const ActionBeginSavepointResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionBeginSavepointResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes savepoint_id = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_savepoint_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_savepoint_id());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionBeginSavepointResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionBeginSavepointResult*>(&to_msg);
  auto& from = static_cast<const ActionBeginSavepointResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionBeginSavepointResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_savepoint_id().empty()) {
      _this->_internal_set_savepoint_id(from._internal_savepoint_id());
    } else {
      if (_this->_impl_.savepoint_id_.IsDefault()) {
        _this->_internal_set_savepoint_id("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionBeginSavepointResult::CopyFrom(const ActionBeginSavepointResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionBeginSavepointResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionBeginSavepointResult::InternalSwap(ActionBeginSavepointResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.savepoint_id_, &other->_impl_.savepoint_id_, arena);
}

::google::protobuf::Metadata ActionBeginSavepointResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionEndTransactionRequest::ActionEndTransactionRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionEndTransactionRequest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionEndTransactionRequest)
}
PROTOBUF_NDEBUG_INLINE ActionEndTransactionRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionEndTransactionRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        transaction_id_(arena, from.transaction_id_) {}

ActionEndTransactionRequest::ActionEndTransactionRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionEndTransactionRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionEndTransactionRequest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionEndTransactionRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.action_ = from._impl_.action_;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionEndTransactionRequest)
}
PROTOBUF_NDEBUG_INLINE ActionEndTransactionRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        transaction_id_(arena) {}

inline void ActionEndTransactionRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.action_ = {};
}
ActionEndTransactionRequest::~ActionEndTransactionRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionEndTransactionRequest)
  SharedDtor(*this);
}
inline void ActionEndTransactionRequest::SharedDtor(MessageLite& self) {
  ActionEndTransactionRequest& this_ = static_cast<ActionEndTransactionRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.transaction_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionEndTransactionRequest_class_data_ =
        ActionEndTransactionRequest::InternalGenerateClassData_(ActionEndTransactionRequest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionEndTransactionRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionEndTransactionRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionEndTransactionRequest_class_data_.tc_table);
  return ActionEndTransactionRequest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionEndTransactionRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionEndTransactionRequest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionEndTransactionRequest_globals_));
  return ActionEndTransactionRequest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionEndTransactionRequest::ParseTableT_
    ActionEndTransactionRequest::_table_ =
        ActionEndTransactionRequest::InternalGenerateParseTable_(ActionEndTransactionRequest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionEndTransactionRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionEndTransactionRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.transaction_id_.ClearNonDefaultToEmpty();
  }
  _impl_.action_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionEndTransactionRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionEndTransactionRequest& this_ = static_cast<const ActionEndTransactionRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionEndTransactionRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionEndTransactionRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionEndTransactionRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes transaction_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_transaction_id().empty()) {
      const ::std::string& _s = this_._internal_transaction_id();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  // .arrow.flight.protocol.sql.ActionEndTransactionRequest.EndTransaction action = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_action() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          2, this_._internal_action(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionEndTransactionRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionEndTransactionRequest::ByteSizeLong(const MessageLite& base) {
  const ActionEndTransactionRequest& this_ = static_cast<const ActionEndTransactionRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionEndTransactionRequest::ByteSizeLong() const {
  const ActionEndTransactionRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionEndTransactionRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // bytes transaction_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_transaction_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_transaction_id());
      }
    }
    // .arrow.flight.protocol.sql.ActionEndTransactionRequest.EndTransaction action = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_action() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_action());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionEndTransactionRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionEndTransactionRequest*>(&to_msg);
  auto& from = static_cast<const ActionEndTransactionRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionEndTransactionRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_transaction_id().empty()) {
        _this->_internal_set_transaction_id(from._internal_transaction_id());
      } else {
        if (_this->_impl_.transaction_id_.IsDefault()) {
          _this->_internal_set_transaction_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_action() != 0) {
        _this->_impl_.action_ = from._impl_.action_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionEndTransactionRequest::CopyFrom(const ActionEndTransactionRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionEndTransactionRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionEndTransactionRequest::InternalSwap(ActionEndTransactionRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.transaction_id_, &other->_impl_.transaction_id_, arena);
  swap(_impl_.action_, other->_impl_.action_);
}

::google::protobuf::Metadata ActionEndTransactionRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionEndSavepointRequest::ActionEndSavepointRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionEndSavepointRequest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionEndSavepointRequest)
}
PROTOBUF_NDEBUG_INLINE ActionEndSavepointRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionEndSavepointRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        savepoint_id_(arena, from.savepoint_id_) {}

ActionEndSavepointRequest::ActionEndSavepointRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionEndSavepointRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionEndSavepointRequest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionEndSavepointRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.action_ = from._impl_.action_;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionEndSavepointRequest)
}
PROTOBUF_NDEBUG_INLINE ActionEndSavepointRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        savepoint_id_(arena) {}

inline void ActionEndSavepointRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.action_ = {};
}
ActionEndSavepointRequest::~ActionEndSavepointRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionEndSavepointRequest)
  SharedDtor(*this);
}
inline void ActionEndSavepointRequest::SharedDtor(MessageLite& self) {
  ActionEndSavepointRequest& this_ = static_cast<ActionEndSavepointRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.savepoint_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionEndSavepointRequest_class_data_ =
        ActionEndSavepointRequest::InternalGenerateClassData_(ActionEndSavepointRequest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionEndSavepointRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionEndSavepointRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionEndSavepointRequest_class_data_.tc_table);
  return ActionEndSavepointRequest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionEndSavepointRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionEndSavepointRequest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionEndSavepointRequest_globals_));
  return ActionEndSavepointRequest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionEndSavepointRequest::ParseTableT_
    ActionEndSavepointRequest::_table_ =
        ActionEndSavepointRequest::InternalGenerateParseTable_(ActionEndSavepointRequest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionEndSavepointRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionEndSavepointRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.savepoint_id_.ClearNonDefaultToEmpty();
  }
  _impl_.action_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionEndSavepointRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionEndSavepointRequest& this_ = static_cast<const ActionEndSavepointRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionEndSavepointRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionEndSavepointRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionEndSavepointRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes savepoint_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_savepoint_id().empty()) {
      const ::std::string& _s = this_._internal_savepoint_id();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  // .arrow.flight.protocol.sql.ActionEndSavepointRequest.EndSavepoint action = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_action() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          2, this_._internal_action(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionEndSavepointRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionEndSavepointRequest::ByteSizeLong(const MessageLite& base) {
  const ActionEndSavepointRequest& this_ = static_cast<const ActionEndSavepointRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionEndSavepointRequest::ByteSizeLong() const {
  const ActionEndSavepointRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionEndSavepointRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // bytes savepoint_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_savepoint_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_savepoint_id());
      }
    }
    // .arrow.flight.protocol.sql.ActionEndSavepointRequest.EndSavepoint action = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_action() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_action());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionEndSavepointRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionEndSavepointRequest*>(&to_msg);
  auto& from = static_cast<const ActionEndSavepointRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionEndSavepointRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_savepoint_id().empty()) {
        _this->_internal_set_savepoint_id(from._internal_savepoint_id());
      } else {
        if (_this->_impl_.savepoint_id_.IsDefault()) {
          _this->_internal_set_savepoint_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_action() != 0) {
        _this->_impl_.action_ = from._impl_.action_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionEndSavepointRequest::CopyFrom(const ActionEndSavepointRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionEndSavepointRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionEndSavepointRequest::InternalSwap(ActionEndSavepointRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.savepoint_id_, &other->_impl_.savepoint_id_, arena);
  swap(_impl_.action_, other->_impl_.action_);
}

::google::protobuf::Metadata ActionEndSavepointRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandStatementQuery::CommandStatementQuery(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementQuery_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandStatementQuery)
}
PROTOBUF_NDEBUG_INLINE CommandStatementQuery::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandStatementQuery& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        query_(arena, from.query_),
        transaction_id_(arena, from.transaction_id_) {}

CommandStatementQuery::CommandStatementQuery(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandStatementQuery& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementQuery_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandStatementQuery* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandStatementQuery)
}
PROTOBUF_NDEBUG_INLINE CommandStatementQuery::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        query_(arena),
        transaction_id_(arena) {}

inline void CommandStatementQuery::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandStatementQuery::~CommandStatementQuery() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandStatementQuery)
  SharedDtor(*this);
}
inline void CommandStatementQuery::SharedDtor(MessageLite& self) {
  CommandStatementQuery& this_ = static_cast<CommandStatementQuery&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.query_.Destroy();
  this_._impl_.transaction_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandStatementQuery_class_data_ =
        CommandStatementQuery::InternalGenerateClassData_(CommandStatementQuery_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementQuery::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementQuery_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandStatementQuery_class_data_.tc_table);
  return CommandStatementQuery_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementQuery::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementQuery_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandStatementQuery_globals_));
  return CommandStatementQuery_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandStatementQuery::ParseTableT_
    CommandStatementQuery::_table_ =
        CommandStatementQuery::InternalGenerateParseTable_(CommandStatementQuery_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandStatementQuery::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandStatementQuery)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.query_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.transaction_id_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandStatementQuery::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandStatementQuery& this_ = static_cast<const CommandStatementQuery&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandStatementQuery::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandStatementQuery& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandStatementQuery)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string query = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_query().empty()) {
      const ::std::string& _s = this_._internal_query();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandStatementQuery.query");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // optional bytes transaction_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_transaction_id();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandStatementQuery)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandStatementQuery::ByteSizeLong(const MessageLite& base) {
  const CommandStatementQuery& this_ = static_cast<const CommandStatementQuery&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandStatementQuery::ByteSizeLong() const {
  const CommandStatementQuery& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandStatementQuery)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string query = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_query().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_query());
      }
    }
    // optional bytes transaction_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_transaction_id());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandStatementQuery::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandStatementQuery*>(&to_msg);
  auto& from = static_cast<const CommandStatementQuery&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandStatementQuery)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_query().empty()) {
        _this->_internal_set_query(from._internal_query());
      } else {
        if (_this->_impl_.query_.IsDefault()) {
          _this->_internal_set_query("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_transaction_id(from._internal_transaction_id());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandStatementQuery::CopyFrom(const CommandStatementQuery& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandStatementQuery)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandStatementQuery::InternalSwap(CommandStatementQuery* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.query_, &other->_impl_.query_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.transaction_id_, &other->_impl_.transaction_id_, arena);
}

::google::protobuf::Metadata CommandStatementQuery::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandStatementSubstraitPlan::CommandStatementSubstraitPlan(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementSubstraitPlan_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandStatementSubstraitPlan)
}
PROTOBUF_NDEBUG_INLINE CommandStatementSubstraitPlan::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandStatementSubstraitPlan& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        transaction_id_(arena, from.transaction_id_) {}

CommandStatementSubstraitPlan::CommandStatementSubstraitPlan(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandStatementSubstraitPlan& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementSubstraitPlan_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandStatementSubstraitPlan* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.plan_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.plan_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandStatementSubstraitPlan)
}
PROTOBUF_NDEBUG_INLINE CommandStatementSubstraitPlan::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        transaction_id_(arena) {}

inline void CommandStatementSubstraitPlan::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.plan_ = {};
}
CommandStatementSubstraitPlan::~CommandStatementSubstraitPlan() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandStatementSubstraitPlan)
  SharedDtor(*this);
}
inline void CommandStatementSubstraitPlan::SharedDtor(MessageLite& self) {
  CommandStatementSubstraitPlan& this_ = static_cast<CommandStatementSubstraitPlan&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.transaction_id_.Destroy();
  delete this_._impl_.plan_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandStatementSubstraitPlan_class_data_ =
        CommandStatementSubstraitPlan::InternalGenerateClassData_(CommandStatementSubstraitPlan_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementSubstraitPlan::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementSubstraitPlan_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandStatementSubstraitPlan_class_data_.tc_table);
  return CommandStatementSubstraitPlan_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementSubstraitPlan::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementSubstraitPlan_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandStatementSubstraitPlan_globals_));
  return CommandStatementSubstraitPlan_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandStatementSubstraitPlan::ParseTableT_
    CommandStatementSubstraitPlan::_table_ =
        CommandStatementSubstraitPlan::InternalGenerateParseTable_(CommandStatementSubstraitPlan_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandStatementSubstraitPlan::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandStatementSubstraitPlan)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.transaction_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.plan_ != nullptr);
      _impl_.plan_->Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandStatementSubstraitPlan::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandStatementSubstraitPlan& this_ = static_cast<const CommandStatementSubstraitPlan&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandStatementSubstraitPlan::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandStatementSubstraitPlan& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandStatementSubstraitPlan)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.sql.SubstraitPlan plan = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.plan_, this_._impl_.plan_->GetCachedSize(), target,
        stream);
  }

  // optional bytes transaction_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_transaction_id();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandStatementSubstraitPlan)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandStatementSubstraitPlan::ByteSizeLong(const MessageLite& base) {
  const CommandStatementSubstraitPlan& this_ = static_cast<const CommandStatementSubstraitPlan&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandStatementSubstraitPlan::ByteSizeLong() const {
  const CommandStatementSubstraitPlan& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandStatementSubstraitPlan)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // optional bytes transaction_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_transaction_id());
    }
    // .arrow.flight.protocol.sql.SubstraitPlan plan = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.plan_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandStatementSubstraitPlan::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandStatementSubstraitPlan*>(&to_msg);
  auto& from = static_cast<const CommandStatementSubstraitPlan&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandStatementSubstraitPlan)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_transaction_id(from._internal_transaction_id());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.plan_ != nullptr);
      if (_this->_impl_.plan_ == nullptr) {
        _this->_impl_.plan_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.plan_);
      } else {
        _this->_impl_.plan_->MergeFrom(*from._impl_.plan_);
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandStatementSubstraitPlan::CopyFrom(const CommandStatementSubstraitPlan& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandStatementSubstraitPlan)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandStatementSubstraitPlan::InternalSwap(CommandStatementSubstraitPlan* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.transaction_id_, &other->_impl_.transaction_id_, arena);
  swap(_impl_.plan_, other->_impl_.plan_);
}

::google::protobuf::Metadata CommandStatementSubstraitPlan::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

TicketStatementQuery::TicketStatementQuery(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, TicketStatementQuery_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.TicketStatementQuery)
}
PROTOBUF_NDEBUG_INLINE TicketStatementQuery::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::TicketStatementQuery& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        statement_handle_(arena, from.statement_handle_) {}

TicketStatementQuery::TicketStatementQuery(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const TicketStatementQuery& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, TicketStatementQuery_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  TicketStatementQuery* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.TicketStatementQuery)
}
PROTOBUF_NDEBUG_INLINE TicketStatementQuery::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        statement_handle_(arena) {}

inline void TicketStatementQuery::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
TicketStatementQuery::~TicketStatementQuery() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.TicketStatementQuery)
  SharedDtor(*this);
}
inline void TicketStatementQuery::SharedDtor(MessageLite& self) {
  TicketStatementQuery& this_ = static_cast<TicketStatementQuery&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.statement_handle_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull TicketStatementQuery_class_data_ =
        TicketStatementQuery::InternalGenerateClassData_(TicketStatementQuery_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
TicketStatementQuery::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&TicketStatementQuery_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(TicketStatementQuery_class_data_.tc_table);
  return TicketStatementQuery_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
TicketStatementQuery::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&TicketStatementQuery_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&TicketStatementQuery_globals_));
  return TicketStatementQuery_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const TicketStatementQuery::ParseTableT_
    TicketStatementQuery::_table_ =
        TicketStatementQuery::InternalGenerateParseTable_(TicketStatementQuery_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void TicketStatementQuery::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.TicketStatementQuery)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.statement_handle_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL TicketStatementQuery::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const TicketStatementQuery& this_ = static_cast<const TicketStatementQuery&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL TicketStatementQuery::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const TicketStatementQuery& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.TicketStatementQuery)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes statement_handle = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_statement_handle().empty()) {
      const ::std::string& _s = this_._internal_statement_handle();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.TicketStatementQuery)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t TicketStatementQuery::ByteSizeLong(const MessageLite& base) {
  const TicketStatementQuery& this_ = static_cast<const TicketStatementQuery&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t TicketStatementQuery::ByteSizeLong() const {
  const TicketStatementQuery& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.TicketStatementQuery)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes statement_handle = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_statement_handle().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_statement_handle());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void TicketStatementQuery::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<TicketStatementQuery*>(&to_msg);
  auto& from = static_cast<const TicketStatementQuery&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.TicketStatementQuery)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_statement_handle().empty()) {
      _this->_internal_set_statement_handle(from._internal_statement_handle());
    } else {
      if (_this->_impl_.statement_handle_.IsDefault()) {
        _this->_internal_set_statement_handle("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void TicketStatementQuery::CopyFrom(const TicketStatementQuery& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.TicketStatementQuery)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void TicketStatementQuery::InternalSwap(TicketStatementQuery* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.statement_handle_, &other->_impl_.statement_handle_, arena);
}

::google::protobuf::Metadata TicketStatementQuery::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandPreparedStatementQuery::CommandPreparedStatementQuery(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandPreparedStatementQuery_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandPreparedStatementQuery)
}
PROTOBUF_NDEBUG_INLINE CommandPreparedStatementQuery::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandPreparedStatementQuery& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        prepared_statement_handle_(arena, from.prepared_statement_handle_) {}

CommandPreparedStatementQuery::CommandPreparedStatementQuery(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandPreparedStatementQuery& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandPreparedStatementQuery_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandPreparedStatementQuery* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandPreparedStatementQuery)
}
PROTOBUF_NDEBUG_INLINE CommandPreparedStatementQuery::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        prepared_statement_handle_(arena) {}

inline void CommandPreparedStatementQuery::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandPreparedStatementQuery::~CommandPreparedStatementQuery() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandPreparedStatementQuery)
  SharedDtor(*this);
}
inline void CommandPreparedStatementQuery::SharedDtor(MessageLite& self) {
  CommandPreparedStatementQuery& this_ = static_cast<CommandPreparedStatementQuery&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.prepared_statement_handle_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandPreparedStatementQuery_class_data_ =
        CommandPreparedStatementQuery::InternalGenerateClassData_(CommandPreparedStatementQuery_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandPreparedStatementQuery::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandPreparedStatementQuery_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandPreparedStatementQuery_class_data_.tc_table);
  return CommandPreparedStatementQuery_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandPreparedStatementQuery::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandPreparedStatementQuery_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandPreparedStatementQuery_globals_));
  return CommandPreparedStatementQuery_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandPreparedStatementQuery::ParseTableT_
    CommandPreparedStatementQuery::_table_ =
        CommandPreparedStatementQuery::InternalGenerateParseTable_(CommandPreparedStatementQuery_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandPreparedStatementQuery::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandPreparedStatementQuery)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.prepared_statement_handle_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandPreparedStatementQuery::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandPreparedStatementQuery& this_ = static_cast<const CommandPreparedStatementQuery&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandPreparedStatementQuery::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandPreparedStatementQuery& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandPreparedStatementQuery)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes prepared_statement_handle = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_prepared_statement_handle().empty()) {
      const ::std::string& _s = this_._internal_prepared_statement_handle();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandPreparedStatementQuery)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandPreparedStatementQuery::ByteSizeLong(const MessageLite& base) {
  const CommandPreparedStatementQuery& this_ = static_cast<const CommandPreparedStatementQuery&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandPreparedStatementQuery::ByteSizeLong() const {
  const CommandPreparedStatementQuery& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandPreparedStatementQuery)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes prepared_statement_handle = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_prepared_statement_handle().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_prepared_statement_handle());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandPreparedStatementQuery::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandPreparedStatementQuery*>(&to_msg);
  auto& from = static_cast<const CommandPreparedStatementQuery&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandPreparedStatementQuery)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_prepared_statement_handle().empty()) {
      _this->_internal_set_prepared_statement_handle(from._internal_prepared_statement_handle());
    } else {
      if (_this->_impl_.prepared_statement_handle_.IsDefault()) {
        _this->_internal_set_prepared_statement_handle("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandPreparedStatementQuery::CopyFrom(const CommandPreparedStatementQuery& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandPreparedStatementQuery)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandPreparedStatementQuery::InternalSwap(CommandPreparedStatementQuery* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.prepared_statement_handle_, &other->_impl_.prepared_statement_handle_, arena);
}

::google::protobuf::Metadata CommandPreparedStatementQuery::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandStatementUpdate::CommandStatementUpdate(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementUpdate_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandStatementUpdate)
}
PROTOBUF_NDEBUG_INLINE CommandStatementUpdate::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandStatementUpdate& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        query_(arena, from.query_),
        transaction_id_(arena, from.transaction_id_) {}

CommandStatementUpdate::CommandStatementUpdate(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandStatementUpdate& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementUpdate_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandStatementUpdate* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandStatementUpdate)
}
PROTOBUF_NDEBUG_INLINE CommandStatementUpdate::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        query_(arena),
        transaction_id_(arena) {}

inline void CommandStatementUpdate::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandStatementUpdate::~CommandStatementUpdate() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandStatementUpdate)
  SharedDtor(*this);
}
inline void CommandStatementUpdate::SharedDtor(MessageLite& self) {
  CommandStatementUpdate& this_ = static_cast<CommandStatementUpdate&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.query_.Destroy();
  this_._impl_.transaction_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandStatementUpdate_class_data_ =
        CommandStatementUpdate::InternalGenerateClassData_(CommandStatementUpdate_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementUpdate::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementUpdate_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandStatementUpdate_class_data_.tc_table);
  return CommandStatementUpdate_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementUpdate::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementUpdate_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandStatementUpdate_globals_));
  return CommandStatementUpdate_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandStatementUpdate::ParseTableT_
    CommandStatementUpdate::_table_ =
        CommandStatementUpdate::InternalGenerateParseTable_(CommandStatementUpdate_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandStatementUpdate::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandStatementUpdate)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.query_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.transaction_id_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandStatementUpdate::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandStatementUpdate& this_ = static_cast<const CommandStatementUpdate&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandStatementUpdate::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandStatementUpdate& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandStatementUpdate)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string query = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_query().empty()) {
      const ::std::string& _s = this_._internal_query();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandStatementUpdate.query");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // optional bytes transaction_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_transaction_id();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandStatementUpdate)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandStatementUpdate::ByteSizeLong(const MessageLite& base) {
  const CommandStatementUpdate& this_ = static_cast<const CommandStatementUpdate&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandStatementUpdate::ByteSizeLong() const {
  const CommandStatementUpdate& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandStatementUpdate)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string query = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_query().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_query());
      }
    }
    // optional bytes transaction_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_transaction_id());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandStatementUpdate::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandStatementUpdate*>(&to_msg);
  auto& from = static_cast<const CommandStatementUpdate&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandStatementUpdate)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_query().empty()) {
        _this->_internal_set_query(from._internal_query());
      } else {
        if (_this->_impl_.query_.IsDefault()) {
          _this->_internal_set_query("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_transaction_id(from._internal_transaction_id());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandStatementUpdate::CopyFrom(const CommandStatementUpdate& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandStatementUpdate)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandStatementUpdate::InternalSwap(CommandStatementUpdate* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.query_, &other->_impl_.query_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.transaction_id_, &other->_impl_.transaction_id_, arena);
}

::google::protobuf::Metadata CommandStatementUpdate::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandPreparedStatementUpdate::CommandPreparedStatementUpdate(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandPreparedStatementUpdate_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandPreparedStatementUpdate)
}
PROTOBUF_NDEBUG_INLINE CommandPreparedStatementUpdate::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandPreparedStatementUpdate& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        prepared_statement_handle_(arena, from.prepared_statement_handle_) {}

CommandPreparedStatementUpdate::CommandPreparedStatementUpdate(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandPreparedStatementUpdate& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandPreparedStatementUpdate_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandPreparedStatementUpdate* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandPreparedStatementUpdate)
}
PROTOBUF_NDEBUG_INLINE CommandPreparedStatementUpdate::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        prepared_statement_handle_(arena) {}

inline void CommandPreparedStatementUpdate::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CommandPreparedStatementUpdate::~CommandPreparedStatementUpdate() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandPreparedStatementUpdate)
  SharedDtor(*this);
}
inline void CommandPreparedStatementUpdate::SharedDtor(MessageLite& self) {
  CommandPreparedStatementUpdate& this_ = static_cast<CommandPreparedStatementUpdate&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.prepared_statement_handle_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandPreparedStatementUpdate_class_data_ =
        CommandPreparedStatementUpdate::InternalGenerateClassData_(CommandPreparedStatementUpdate_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandPreparedStatementUpdate::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandPreparedStatementUpdate_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandPreparedStatementUpdate_class_data_.tc_table);
  return CommandPreparedStatementUpdate_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandPreparedStatementUpdate::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandPreparedStatementUpdate_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandPreparedStatementUpdate_globals_));
  return CommandPreparedStatementUpdate_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandPreparedStatementUpdate::ParseTableT_
    CommandPreparedStatementUpdate::_table_ =
        CommandPreparedStatementUpdate::InternalGenerateParseTable_(CommandPreparedStatementUpdate_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandPreparedStatementUpdate::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandPreparedStatementUpdate)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.prepared_statement_handle_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandPreparedStatementUpdate::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandPreparedStatementUpdate& this_ = static_cast<const CommandPreparedStatementUpdate&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandPreparedStatementUpdate::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandPreparedStatementUpdate& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandPreparedStatementUpdate)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes prepared_statement_handle = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_prepared_statement_handle().empty()) {
      const ::std::string& _s = this_._internal_prepared_statement_handle();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandPreparedStatementUpdate)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandPreparedStatementUpdate::ByteSizeLong(const MessageLite& base) {
  const CommandPreparedStatementUpdate& this_ = static_cast<const CommandPreparedStatementUpdate&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandPreparedStatementUpdate::ByteSizeLong() const {
  const CommandPreparedStatementUpdate& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandPreparedStatementUpdate)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes prepared_statement_handle = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_prepared_statement_handle().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_prepared_statement_handle());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandPreparedStatementUpdate::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandPreparedStatementUpdate*>(&to_msg);
  auto& from = static_cast<const CommandPreparedStatementUpdate&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandPreparedStatementUpdate)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_prepared_statement_handle().empty()) {
      _this->_internal_set_prepared_statement_handle(from._internal_prepared_statement_handle());
    } else {
      if (_this->_impl_.prepared_statement_handle_.IsDefault()) {
        _this->_internal_set_prepared_statement_handle("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandPreparedStatementUpdate::CopyFrom(const CommandPreparedStatementUpdate& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandPreparedStatementUpdate)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandPreparedStatementUpdate::InternalSwap(CommandPreparedStatementUpdate* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.prepared_statement_handle_, &other->_impl_.prepared_statement_handle_, arena);
}

::google::protobuf::Metadata CommandPreparedStatementUpdate::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CommandStatementIngest_TableDefinitionOptions::CommandStatementIngest_TableDefinitionOptions(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementIngest_TableDefinitionOptions_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions)
}
CommandStatementIngest_TableDefinitionOptions::CommandStatementIngest_TableDefinitionOptions(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const CommandStatementIngest_TableDefinitionOptions& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementIngest_TableDefinitionOptions_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE CommandStatementIngest_TableDefinitionOptions::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void CommandStatementIngest_TableDefinitionOptions::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, if_not_exist_),
           0,
           offsetof(Impl_, if_exists_) -
               offsetof(Impl_, if_not_exist_) +
               sizeof(Impl_::if_exists_));
}
CommandStatementIngest_TableDefinitionOptions::~CommandStatementIngest_TableDefinitionOptions() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions)
  SharedDtor(*this);
}
inline void CommandStatementIngest_TableDefinitionOptions::SharedDtor(MessageLite& self) {
  CommandStatementIngest_TableDefinitionOptions& this_ = static_cast<CommandStatementIngest_TableDefinitionOptions&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandStatementIngest_TableDefinitionOptions_class_data_ =
        CommandStatementIngest_TableDefinitionOptions::InternalGenerateClassData_(CommandStatementIngest_TableDefinitionOptions_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementIngest_TableDefinitionOptions::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementIngest_TableDefinitionOptions_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandStatementIngest_TableDefinitionOptions_class_data_.tc_table);
  return CommandStatementIngest_TableDefinitionOptions_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementIngest_TableDefinitionOptions::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementIngest_TableDefinitionOptions_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandStatementIngest_TableDefinitionOptions_globals_));
  return CommandStatementIngest_TableDefinitionOptions_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandStatementIngest_TableDefinitionOptions::ParseTableT_
    CommandStatementIngest_TableDefinitionOptions::_table_ =
        CommandStatementIngest_TableDefinitionOptions::InternalGenerateParseTable_(CommandStatementIngest_TableDefinitionOptions_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandStatementIngest_TableDefinitionOptions::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    ::memset(&_impl_.if_not_exist_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.if_exists_) -
        reinterpret_cast<char*>(&_impl_.if_not_exist_)) + sizeof(_impl_.if_exists_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandStatementIngest_TableDefinitionOptions::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandStatementIngest_TableDefinitionOptions& this_ = static_cast<const CommandStatementIngest_TableDefinitionOptions&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandStatementIngest_TableDefinitionOptions::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandStatementIngest_TableDefinitionOptions& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions.TableNotExistOption if_not_exist = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (this_._internal_if_not_exist() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          1, this_._internal_if_not_exist(), target);
    }
  }

  // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions.TableExistsOption if_exists = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_if_exists() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          2, this_._internal_if_exists(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandStatementIngest_TableDefinitionOptions::ByteSizeLong(const MessageLite& base) {
  const CommandStatementIngest_TableDefinitionOptions& this_ = static_cast<const CommandStatementIngest_TableDefinitionOptions&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandStatementIngest_TableDefinitionOptions::ByteSizeLong() const {
  const CommandStatementIngest_TableDefinitionOptions& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions.TableNotExistOption if_not_exist = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (this_._internal_if_not_exist() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_if_not_exist());
      }
    }
    // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions.TableExistsOption if_exists = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_if_exists() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_if_exists());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandStatementIngest_TableDefinitionOptions::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandStatementIngest_TableDefinitionOptions*>(&to_msg);
  auto& from = static_cast<const CommandStatementIngest_TableDefinitionOptions&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (from._internal_if_not_exist() != 0) {
        _this->_impl_.if_not_exist_ = from._impl_.if_not_exist_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_if_exists() != 0) {
        _this->_impl_.if_exists_ = from._impl_.if_exists_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandStatementIngest_TableDefinitionOptions::CopyFrom(const CommandStatementIngest_TableDefinitionOptions& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandStatementIngest_TableDefinitionOptions::InternalSwap(CommandStatementIngest_TableDefinitionOptions* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptions, _impl_.if_exists_)
      + sizeof(CommandStatementIngest_TableDefinitionOptions::_impl_.if_exists_)
      - PROTOBUF_FIELD_OFFSET(CommandStatementIngest_TableDefinitionOptions, _impl_.if_not_exist_)>(
          reinterpret_cast<char*>(&_impl_.if_not_exist_),
          reinterpret_cast<char*>(&other->_impl_.if_not_exist_));
}

::google::protobuf::Metadata CommandStatementIngest_TableDefinitionOptions::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

#if defined(PROTOBUF_CUSTOM_VTABLE)
CommandStatementIngest_OptionsEntry_DoNotUse::CommandStatementIngest_OptionsEntry_DoNotUse() : SuperType(CommandStatementIngest_OptionsEntry_DoNotUse_get_class_data()) {}
CommandStatementIngest_OptionsEntry_DoNotUse::CommandStatementIngest_OptionsEntry_DoNotUse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
    : SuperType(arena, CommandStatementIngest_OptionsEntry_DoNotUse_get_class_data()) {}
#else   // PROTOBUF_CUSTOM_VTABLE
CommandStatementIngest_OptionsEntry_DoNotUse::CommandStatementIngest_OptionsEntry_DoNotUse() : SuperType() {}
CommandStatementIngest_OptionsEntry_DoNotUse::CommandStatementIngest_OptionsEntry_DoNotUse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena) : SuperType(arena) {}
#endif  // PROTOBUF_CUSTOM_VTABLE
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandStatementIngest_OptionsEntry_DoNotUse_class_data_ =
        CommandStatementIngest_OptionsEntry_DoNotUse::InternalGenerateClassData_(CommandStatementIngest_OptionsEntry_DoNotUse_globals_._default);

// including ~ all the vtables and everything they use become part
// of the same ~ SCC.
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementIngest_OptionsEntry_DoNotUse::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementIngest_OptionsEntry_DoNotUse_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandStatementIngest_OptionsEntry_DoNotUse_class_data_.tc_table);
  return CommandStatementIngest_OptionsEntry_DoNotUse_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementIngest_OptionsEntry_DoNotUse::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementIngest_OptionsEntry_DoNotUse_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandStatementIngest_OptionsEntry_DoNotUse_globals_));
  return CommandStatementIngest_OptionsEntry_DoNotUse_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandStatementIngest_OptionsEntry_DoNotUse::ParseTableT_
    CommandStatementIngest_OptionsEntry_DoNotUse::_table_ =
        CommandStatementIngest_OptionsEntry_DoNotUse::InternalGenerateParseTable_(CommandStatementIngest_OptionsEntry_DoNotUse_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
// ===================================================================

CommandStatementIngest::CommandStatementIngest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementIngest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.CommandStatementIngest)
}
PROTOBUF_NDEBUG_INLINE CommandStatementIngest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::CommandStatementIngest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        table_(arena, from.table_),
        schema_(arena, from.schema_),
        catalog_(arena, from.catalog_),
        transaction_id_(arena, from.transaction_id_),
        options_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::arrow::flight::protocol::sql::CommandStatementIngest,
              PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.options_)>()
          , from.options_
        }
     {}

CommandStatementIngest::CommandStatementIngest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CommandStatementIngest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CommandStatementIngest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CommandStatementIngest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.table_definition_options_ = (CheckHasBit(cached_has_bits, 0x00000010U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.table_definition_options_)
                : nullptr;
  _impl_.temporary_ = from._impl_.temporary_;

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.CommandStatementIngest)
}
PROTOBUF_NDEBUG_INLINE CommandStatementIngest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        table_(arena),
        schema_(arena),
        catalog_(arena),
        transaction_id_(arena),
        options_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::arrow::flight::protocol::sql::CommandStatementIngest,
            PROTOBUF_FIELD_OFFSET(::arrow::flight::protocol::sql::CommandStatementIngest, _impl_.options_)>()
         }
     {}

inline void CommandStatementIngest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, table_definition_options_),
           0,
           offsetof(Impl_, temporary_) -
               offsetof(Impl_, table_definition_options_) +
               sizeof(Impl_::temporary_));
}
CommandStatementIngest::~CommandStatementIngest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.CommandStatementIngest)
  SharedDtor(*this);
}
inline void CommandStatementIngest::SharedDtor(MessageLite& self) {
  CommandStatementIngest& this_ = static_cast<CommandStatementIngest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.table_.Destroy();
  this_._impl_.schema_.Destroy();
  this_._impl_.catalog_.Destroy();
  this_._impl_.transaction_id_.Destroy();
  delete this_._impl_.table_definition_options_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CommandStatementIngest_class_data_ =
        CommandStatementIngest::InternalGenerateClassData_(CommandStatementIngest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementIngest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementIngest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CommandStatementIngest_class_data_.tc_table);
  return CommandStatementIngest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CommandStatementIngest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CommandStatementIngest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CommandStatementIngest_globals_));
  return CommandStatementIngest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CommandStatementIngest::ParseTableT_
    CommandStatementIngest::_table_ =
        CommandStatementIngest::InternalGenerateParseTable_(CommandStatementIngest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CommandStatementIngest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.CommandStatementIngest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.table_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.schema_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.catalog_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.transaction_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      ABSL_DCHECK(_impl_.table_definition_options_ != nullptr);
      _impl_.table_definition_options_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000060U)) {
    _impl_.temporary_ = false;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      _impl_.options_.Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CommandStatementIngest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CommandStatementIngest& this_ = static_cast<const CommandStatementIngest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CommandStatementIngest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CommandStatementIngest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.CommandStatementIngest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions table_definition_options = 1;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.table_definition_options_, this_._impl_.table_definition_options_->GetCachedSize(), target,
        stream);
  }

  // string table = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_table().empty()) {
      const ::std::string& _s = this_._internal_table();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandStatementIngest.table");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // optional string schema = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_schema();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandStatementIngest.schema");
    target = stream->WriteStringMaybeAliased(3, _s, target);
  }

  // optional string catalog = 4;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    const ::std::string& _s = this_._internal_catalog();
    ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
        _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandStatementIngest.catalog");
    target = stream->WriteStringMaybeAliased(4, _s, target);
  }

  // bool temporary = 5;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (this_._internal_temporary() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          5, this_._internal_temporary(), target);
    }
  }

  // optional bytes transaction_id = 6;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    const ::std::string& _s = this_._internal_transaction_id();
    target = stream->WriteBytesMaybeAliased(6, _s, target);
  }

  // map<string, string> options = 1000;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    if (!this_._internal_options().empty()) {
      using MapType = ::google::protobuf::Map<::std::string, ::std::string>;
      using WireHelper = _pbi::MapEntryFuncs<::std::string, ::std::string,
                                     _pbi::WireFormatLite::TYPE_STRING,
                                     _pbi::WireFormatLite::TYPE_STRING>;
      const auto& field = this_._internal_options();

      if (stream->IsSerializationDeterministic() && field.size() > 1) {
        for (const auto& entry : ::google::protobuf::internal::MapSorterPtr<MapType>(field)) {
          target = WireHelper::InternalSerialize(
              1000, entry.first, entry.second, target, stream);
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.first.data(), static_cast<int>(entry.first.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandStatementIngest.options");
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.second.data(), static_cast<int>(entry.second.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandStatementIngest.options");
        }
      } else {
        for (const auto& entry : field) {
          target = WireHelper::InternalSerialize(
              1000, entry.first, entry.second, target, stream);
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.first.data(), static_cast<int>(entry.first.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandStatementIngest.options");
          ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
              entry.second.data(), static_cast<int>(entry.second.length()),
 ::google::protobuf::internal::WireFormatLite::SERIALIZE, "arrow.flight.protocol.sql.CommandStatementIngest.options");
        }
      }
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.CommandStatementIngest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CommandStatementIngest::ByteSizeLong(const MessageLite& base) {
  const CommandStatementIngest& this_ = static_cast<const CommandStatementIngest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CommandStatementIngest::ByteSizeLong() const {
  const CommandStatementIngest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.CommandStatementIngest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000007fU)) {
    // string table = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_table().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_table());
      }
    }
    // optional string schema = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_schema());
    }
    // optional string catalog = 4;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_catalog());
    }
    // optional bytes transaction_id = 6;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_transaction_id());
    }
    // .arrow.flight.protocol.sql.CommandStatementIngest.TableDefinitionOptions table_definition_options = 1;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.table_definition_options_);
    }
    // bool temporary = 5;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (this_._internal_temporary() != 0) {
        total_size += 2;
      }
    }
    // map<string, string> options = 1000;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      total_size +=
          2 * ::google::protobuf::internal::FromIntSize(this_._internal_options_size());
      for (const auto& entry : this_._internal_options()) {
        total_size += _pbi::MapEntryFuncs<::std::string, ::std::string,
                                       _pbi::WireFormatLite::TYPE_STRING,
                                       _pbi::WireFormatLite::TYPE_STRING>::ByteSizeLong(entry.first, entry.second);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CommandStatementIngest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CommandStatementIngest*>(&to_msg);
  auto& from = static_cast<const CommandStatementIngest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.CommandStatementIngest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000007fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_table().empty()) {
        _this->_internal_set_table(from._internal_table());
      } else {
        if (_this->_impl_.table_.IsDefault()) {
          _this->_internal_set_table("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_schema(from._internal_schema());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_set_catalog(from._internal_catalog());
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_internal_set_transaction_id(from._internal_transaction_id());
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      ABSL_DCHECK(from._impl_.table_definition_options_ != nullptr);
      if (_this->_impl_.table_definition_options_ == nullptr) {
        _this->_impl_.table_definition_options_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.table_definition_options_);
      } else {
        _this->_impl_.table_definition_options_->MergeFrom(*from._impl_.table_definition_options_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (from._internal_temporary() != 0) {
        _this->_impl_.temporary_ = from._impl_.temporary_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      _this->_impl_.options_.MergeFrom(from._impl_.options_);
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CommandStatementIngest::CopyFrom(const CommandStatementIngest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.CommandStatementIngest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CommandStatementIngest::InternalSwap(CommandStatementIngest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.table_, &other->_impl_.table_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.schema_, &other->_impl_.schema_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.catalog_, &other->_impl_.catalog_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.transaction_id_, &other->_impl_.transaction_id_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.temporary_)
      + sizeof(CommandStatementIngest::_impl_.temporary_)
      - PROTOBUF_FIELD_OFFSET(CommandStatementIngest, _impl_.table_definition_options_)>(
          reinterpret_cast<char*>(&_impl_.table_definition_options_),
          reinterpret_cast<char*>(&other->_impl_.table_definition_options_));
  _impl_.options_.InternalSwap(&other->_impl_.options_);
}

::google::protobuf::Metadata CommandStatementIngest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

DoPutUpdateResult::DoPutUpdateResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DoPutUpdateResult_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.DoPutUpdateResult)
}
DoPutUpdateResult::DoPutUpdateResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const DoPutUpdateResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DoPutUpdateResult_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE DoPutUpdateResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void DoPutUpdateResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.record_count_ = {};
}
DoPutUpdateResult::~DoPutUpdateResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.DoPutUpdateResult)
  SharedDtor(*this);
}
inline void DoPutUpdateResult::SharedDtor(MessageLite& self) {
  DoPutUpdateResult& this_ = static_cast<DoPutUpdateResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull DoPutUpdateResult_class_data_ =
        DoPutUpdateResult::InternalGenerateClassData_(DoPutUpdateResult_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DoPutUpdateResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DoPutUpdateResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(DoPutUpdateResult_class_data_.tc_table);
  return DoPutUpdateResult_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DoPutUpdateResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DoPutUpdateResult_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&DoPutUpdateResult_globals_));
  return DoPutUpdateResult_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const DoPutUpdateResult::ParseTableT_
    DoPutUpdateResult::_table_ =
        DoPutUpdateResult::InternalGenerateParseTable_(DoPutUpdateResult_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void DoPutUpdateResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.DoPutUpdateResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.record_count_ = ::int64_t{0};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL DoPutUpdateResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const DoPutUpdateResult& this_ = static_cast<const DoPutUpdateResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL DoPutUpdateResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const DoPutUpdateResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.DoPutUpdateResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // int64 record_count = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (this_._internal_record_count() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt64ToArrayWithField<1>(
              stream, this_._internal_record_count(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.DoPutUpdateResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t DoPutUpdateResult::ByteSizeLong(const MessageLite& base) {
  const DoPutUpdateResult& this_ = static_cast<const DoPutUpdateResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t DoPutUpdateResult::ByteSizeLong() const {
  const DoPutUpdateResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.DoPutUpdateResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // int64 record_count = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (this_._internal_record_count() != 0) {
        total_size += ::_pbi::WireFormatLite::Int64SizePlusOne(
            this_._internal_record_count());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void DoPutUpdateResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<DoPutUpdateResult*>(&to_msg);
  auto& from = static_cast<const DoPutUpdateResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.DoPutUpdateResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (from._internal_record_count() != 0) {
      _this->_impl_.record_count_ = from._impl_.record_count_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void DoPutUpdateResult::CopyFrom(const DoPutUpdateResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.DoPutUpdateResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void DoPutUpdateResult::InternalSwap(DoPutUpdateResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.record_count_, other->_impl_.record_count_);
}

::google::protobuf::Metadata DoPutUpdateResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

DoPutPreparedStatementResult::DoPutPreparedStatementResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DoPutPreparedStatementResult_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.DoPutPreparedStatementResult)
}
PROTOBUF_NDEBUG_INLINE DoPutPreparedStatementResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::DoPutPreparedStatementResult& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        prepared_statement_handle_(arena, from.prepared_statement_handle_) {}

DoPutPreparedStatementResult::DoPutPreparedStatementResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const DoPutPreparedStatementResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DoPutPreparedStatementResult_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  DoPutPreparedStatementResult* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.DoPutPreparedStatementResult)
}
PROTOBUF_NDEBUG_INLINE DoPutPreparedStatementResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        prepared_statement_handle_(arena) {}

inline void DoPutPreparedStatementResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
DoPutPreparedStatementResult::~DoPutPreparedStatementResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.DoPutPreparedStatementResult)
  SharedDtor(*this);
}
inline void DoPutPreparedStatementResult::SharedDtor(MessageLite& self) {
  DoPutPreparedStatementResult& this_ = static_cast<DoPutPreparedStatementResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.prepared_statement_handle_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull DoPutPreparedStatementResult_class_data_ =
        DoPutPreparedStatementResult::InternalGenerateClassData_(DoPutPreparedStatementResult_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DoPutPreparedStatementResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DoPutPreparedStatementResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(DoPutPreparedStatementResult_class_data_.tc_table);
  return DoPutPreparedStatementResult_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DoPutPreparedStatementResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DoPutPreparedStatementResult_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&DoPutPreparedStatementResult_globals_));
  return DoPutPreparedStatementResult_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const DoPutPreparedStatementResult::ParseTableT_
    DoPutPreparedStatementResult::_table_ =
        DoPutPreparedStatementResult::InternalGenerateParseTable_(DoPutPreparedStatementResult_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void DoPutPreparedStatementResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.DoPutPreparedStatementResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.prepared_statement_handle_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL DoPutPreparedStatementResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const DoPutPreparedStatementResult& this_ = static_cast<const DoPutPreparedStatementResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL DoPutPreparedStatementResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const DoPutPreparedStatementResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.DoPutPreparedStatementResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional bytes prepared_statement_handle = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_prepared_statement_handle();
    target = stream->WriteBytesMaybeAliased(1, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.DoPutPreparedStatementResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t DoPutPreparedStatementResult::ByteSizeLong(const MessageLite& base) {
  const DoPutPreparedStatementResult& this_ = static_cast<const DoPutPreparedStatementResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t DoPutPreparedStatementResult::ByteSizeLong() const {
  const DoPutPreparedStatementResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.DoPutPreparedStatementResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // optional bytes prepared_statement_handle = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_prepared_statement_handle());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void DoPutPreparedStatementResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<DoPutPreparedStatementResult*>(&to_msg);
  auto& from = static_cast<const DoPutPreparedStatementResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.DoPutPreparedStatementResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_set_prepared_statement_handle(from._internal_prepared_statement_handle());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void DoPutPreparedStatementResult::CopyFrom(const DoPutPreparedStatementResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.DoPutPreparedStatementResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void DoPutPreparedStatementResult::InternalSwap(DoPutPreparedStatementResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.prepared_statement_handle_, &other->_impl_.prepared_statement_handle_, arena);
}

::google::protobuf::Metadata DoPutPreparedStatementResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionCancelQueryRequest::ActionCancelQueryRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCancelQueryRequest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionCancelQueryRequest)
}
PROTOBUF_NDEBUG_INLINE ActionCancelQueryRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::arrow::flight::protocol::sql::ActionCancelQueryRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        info_(arena, from.info_) {}

ActionCancelQueryRequest::ActionCancelQueryRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ActionCancelQueryRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCancelQueryRequest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ActionCancelQueryRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:arrow.flight.protocol.sql.ActionCancelQueryRequest)
}
PROTOBUF_NDEBUG_INLINE ActionCancelQueryRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        info_(arena) {}

inline void ActionCancelQueryRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
ActionCancelQueryRequest::~ActionCancelQueryRequest() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionCancelQueryRequest)
  SharedDtor(*this);
}
inline void ActionCancelQueryRequest::SharedDtor(MessageLite& self) {
  ActionCancelQueryRequest& this_ = static_cast<ActionCancelQueryRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.info_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionCancelQueryRequest_class_data_ =
        ActionCancelQueryRequest::InternalGenerateClassData_(ActionCancelQueryRequest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCancelQueryRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCancelQueryRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionCancelQueryRequest_class_data_.tc_table);
  return ActionCancelQueryRequest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCancelQueryRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCancelQueryRequest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionCancelQueryRequest_globals_));
  return ActionCancelQueryRequest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionCancelQueryRequest::ParseTableT_
    ActionCancelQueryRequest::_table_ =
        ActionCancelQueryRequest::InternalGenerateParseTable_(ActionCancelQueryRequest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionCancelQueryRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionCancelQueryRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.info_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionCancelQueryRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionCancelQueryRequest& this_ = static_cast<const ActionCancelQueryRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionCancelQueryRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionCancelQueryRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionCancelQueryRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // bytes info = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_info().empty()) {
      const ::std::string& _s = this_._internal_info();
      target = stream->WriteBytesMaybeAliased(1, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionCancelQueryRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionCancelQueryRequest::ByteSizeLong(const MessageLite& base) {
  const ActionCancelQueryRequest& this_ = static_cast<const ActionCancelQueryRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionCancelQueryRequest::ByteSizeLong() const {
  const ActionCancelQueryRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionCancelQueryRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // bytes info = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_info().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_info());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionCancelQueryRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionCancelQueryRequest*>(&to_msg);
  auto& from = static_cast<const ActionCancelQueryRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionCancelQueryRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!from._internal_info().empty()) {
      _this->_internal_set_info(from._internal_info());
    } else {
      if (_this->_impl_.info_.IsDefault()) {
        _this->_internal_set_info("");
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionCancelQueryRequest::CopyFrom(const ActionCancelQueryRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionCancelQueryRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionCancelQueryRequest::InternalSwap(ActionCancelQueryRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.info_, &other->_impl_.info_, arena);
}

::google::protobuf::Metadata ActionCancelQueryRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ActionCancelQueryResult::ActionCancelQueryResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCancelQueryResult_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:arrow.flight.protocol.sql.ActionCancelQueryResult)
}
ActionCancelQueryResult::ActionCancelQueryResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ActionCancelQueryResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ActionCancelQueryResult_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE ActionCancelQueryResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void ActionCancelQueryResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.result_ = {};
}
ActionCancelQueryResult::~ActionCancelQueryResult() {
  // @@protoc_insertion_point(destructor:arrow.flight.protocol.sql.ActionCancelQueryResult)
  SharedDtor(*this);
}
inline void ActionCancelQueryResult::SharedDtor(MessageLite& self) {
  ActionCancelQueryResult& this_ = static_cast<ActionCancelQueryResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ActionCancelQueryResult_class_data_ =
        ActionCancelQueryResult::InternalGenerateClassData_(ActionCancelQueryResult_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCancelQueryResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCancelQueryResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ActionCancelQueryResult_class_data_.tc_table);
  return ActionCancelQueryResult_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ActionCancelQueryResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ActionCancelQueryResult_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ActionCancelQueryResult_globals_));
  return ActionCancelQueryResult_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ActionCancelQueryResult::ParseTableT_
    ActionCancelQueryResult::_table_ =
        ActionCancelQueryResult::InternalGenerateParseTable_(ActionCancelQueryResult_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ActionCancelQueryResult::Clear() {
// @@protoc_insertion_point(message_clear_start:arrow.flight.protocol.sql.ActionCancelQueryResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.result_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ActionCancelQueryResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ActionCancelQueryResult& this_ = static_cast<const ActionCancelQueryResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ActionCancelQueryResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ActionCancelQueryResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:arrow.flight.protocol.sql.ActionCancelQueryResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // .arrow.flight.protocol.sql.ActionCancelQueryResult.CancelResult result = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (this_._internal_result() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          1, this_._internal_result(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:arrow.flight.protocol.sql.ActionCancelQueryResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ActionCancelQueryResult::ByteSizeLong(const MessageLite& base) {
  const ActionCancelQueryResult& this_ = static_cast<const ActionCancelQueryResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ActionCancelQueryResult::ByteSizeLong() const {
  const ActionCancelQueryResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:arrow.flight.protocol.sql.ActionCancelQueryResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // .arrow.flight.protocol.sql.ActionCancelQueryResult.CancelResult result = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (this_._internal_result() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_result());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ActionCancelQueryResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ActionCancelQueryResult*>(&to_msg);
  auto& from = static_cast<const ActionCancelQueryResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:arrow.flight.protocol.sql.ActionCancelQueryResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (from._internal_result() != 0) {
      _this->_impl_.result_ = from._impl_.result_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ActionCancelQueryResult::CopyFrom(const ActionCancelQueryResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:arrow.flight.protocol.sql.ActionCancelQueryResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ActionCancelQueryResult::InternalSwap(ActionCancelQueryResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.result_, other->_impl_.result_);
}

::google::protobuf::Metadata ActionCancelQueryResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
PROTOBUF_CONSTINIT ARROW_FLIGHT_SQL_EXPORT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 ::_pbi::ExtensionIdentifier<
        ::google::protobuf::MessageOptions, ::_pbi::PrimitiveTypeTraits< bool >, 8, false>
        experimental(kExperimentalFieldNumber, false);
// @@protoc_insertion_point(namespace_scope)
}  // namespace sql
}  // namespace protocol
}  // namespace flight
}  // namespace arrow
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_FlightSql_2eproto),
        ::_pbi::ExtensionSet::RegisterExtension(
            &::google::protobuf::MessageOptions::default_instance(), 1000, 8,
            false, false, /*is_utf8=*/false),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
